/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  
 *  Modifications :
 *  2006 Creation P.Delrieu
 * 
 */
package org.openmobileis.module.core;

import java.io.IOException;
import java.io.Reader;
import java.util.StringTokenizer;

import org.kxml2.io.KXmlParser;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.common.util.log.LogServices;
import org.openmobileis.module.profiles.terminal.ProfileRubricSynchroListener;
import org.xmlpull.v1.XmlPullParserException;

/**
 * Title: OpenMobileIS project source <BR>
 * Description: Profile file parser. Extract module and Rubric from the file.
 * 
 * @author Philippe Delrieu
 * @since JDK 1.1
 * @version 1.0.
 */
public class XmlModuleParser {

  private final static String TAG_XML = "xml";

  private final static String TAG_RUBRIC = "rubric";

  private final static String TAG_PROPERTY = "property";

  private final static String TAG_RUBRIC_LOADER = "loader";

  private final static String TAG_RUBRIC_SYNCHROLISTENER = "synchrolistener";

  private final static String TAG_MODULE = "module";

  private final static String TAG_MODDEPEND = "moddepend";

  private Reader reader;

  private ProfileModuleRubric currentRubric = null;

  private ProfileModule moduletoload = null;

  String currentTag;

  public XmlModuleParser(Reader reader) {
    this.reader = reader;
  }

  public void parse() throws XmlPullParserException, IOException {
    try {
      KXmlParser parser = new KXmlParser();
      parser.setInput(this.reader);
  
      int eventType = parser.getEventType();
      do {
        if (eventType == KXmlParser.START_DOCUMENT) {
        } else if (eventType == KXmlParser.END_DOCUMENT) {
        } else if (eventType == KXmlParser.START_TAG) {
          this.processStartElement(parser);
        } else if (eventType == KXmlParser.END_TAG) {
          this.processEndElement(parser);
        } else if (eventType == KXmlParser.TEXT) {
          this.processText(parser);
        }
        eventType = parser.next();
      } while (eventType != KXmlParser.END_DOCUMENT);
    } finally   {
      reader.close();
    }
  }

  private void processStartElement(KXmlParser parser) {
    String name = parser.getName();

    if (name.equals(XmlModuleParser.TAG_XML)) {
      // Do nothing
    } else if (name.equals(XmlModuleParser.TAG_MODULE)) {
      this.openModuleTag(parser);
   } else if (name.equals(XmlModuleParser.TAG_RUBRIC)) {
      this.openRubricTag(parser);
    } else if (name.equals(XmlModuleParser.TAG_PROPERTY)) {
      this.openPropertyTag(parser);
    } else if (name.equals(XmlModuleParser.TAG_RUBRIC_LOADER)) {
      this.openLoaderTag(parser);
    } else if (name.equals(XmlModuleParser.TAG_RUBRIC_SYNCHROLISTENER)) {
      this.openSynchroListenerTag(parser);
    } else if (name.equals(XmlModuleParser.TAG_MODDEPEND)) {
      this.openModDependTag(parser);
    }
  }

  private void openModuleTag(KXmlParser parser) {
    if (currentRubric != null) {
      currentRubric = null;
    }
    moduletoload = new ProfileModule();
    int count = parser.getAttributeCount();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("name")) {
        moduletoload.setName(parser.getAttributeValue(i));
      } else if (parser.getAttributeName(i).equals("version")) {
        moduletoload.setVersion(parser.getAttributeValue(i));
      } else if (parser.getAttributeName(i).equals("serverjar")) {
        moduletoload.setServerjar(parser.getAttributeValue(i));
      } else if (parser.getAttributeName(i).equals("terminaljar")) {
        moduletoload.setTerminaljar(parser.getAttributeValue(i));
      } else if (parser.getAttributeName(i).equals("terminaltype")) {
        moduletoload.setTerminalType(parser.getAttributeValue(i));
      } else {
        LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted module file, unrecognized attribute " + parser.getAttributeName(i) + " for modrubric");
      }
    }
  }

  private void openRubricTag(KXmlParser parser) {
    if (moduletoload == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted module file, starting a new rubric while module isn't opened.");
      return;
    }
    currentRubric = new ProfileModuleRubric();
    int count = parser.getAttributeCount();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("name")) {
        currentRubric.setName(parser.getAttributeValue(i));
      } else if (parser.getAttributeName(i).equals("depends")) {
        String dep = parser.getAttributeValue(i).trim();
        Array deplist = new Array();
        StringTokenizer token = new StringTokenizer(dep, ",");
        while (token.hasMoreElements()) {
          deplist.add(token.nextElement());
        }
        String[] strdeplist = new String[deplist.size()];
        deplist.toArray(strdeplist);
        currentRubric.setDependsList(strdeplist);
      } 
    }
  }

  private void closeRubricTag() {
    if (currentRubric == null) {
      LogManager.traceError(LogServices.SYNCHROSERVICE, "Corrupted module file, found rubric closing tag while no rubric is open");
      return;
    }
    if (moduletoload == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted module file, end rubrictag while module isn't opened.");
      return;
    }
    moduletoload.addModuleRubric(currentRubric);
    currentRubric = null;
  }

  private void openPropertyTag(KXmlParser parser) {
    String propertyName = null;
    String propertyValue = null;

    int count = parser.getAttributeCount();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("name")) {
        propertyName = parser.getAttributeValue(i);
      } else if (parser.getAttributeName(i).equals("value")) {
        propertyValue = parser.getAttributeValue(i);
      }
    }
    if (currentRubric != null) {
      currentRubric.addProperty(propertyName, propertyValue);
    } else if (moduletoload != null) {
      moduletoload.addProperty(propertyName, propertyValue);
    }
  }

  private void openLoaderTag(KXmlParser parser) {
    if (currentRubric == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted module file, loader tag found outside a rubric");
      return;
    }
   int count = parser.getAttributeCount();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("class")) {
        String classname = parser.getAttributeValue(i).trim();
        currentRubric.setLoaderClass(classname) ;
      }
    }
  }

  private void openSynchroListenerTag(KXmlParser parser) {
    if (currentRubric == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, synchrolistener tag found outside a rubric");
      return;
    }
    ProfileRubricSynchroListener synclistener = new ProfileRubricSynchroListener();
    int count = parser.getAttributeCount();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("name")) {
        synclistener.name = parser.getAttributeValue(i).trim();
      } else if (parser.getAttributeName(i).equals("depends")) {
        String dep = parser.getAttributeValue(i).trim();
        Array deplist = new Array();
        StringTokenizer token = new StringTokenizer(dep, ",");
        while (token.hasMoreElements()) {
          deplist.add(token.nextElement());
        }
        String[] strdeplist = new String[deplist.size()];
        deplist.toArray(strdeplist);
        synclistener.dependsList = strdeplist;
      }
    }
    currentRubric.addSynchroListener(synclistener);
  }

  private void openModDependTag(KXmlParser parser) {
    if (moduletoload == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted module file, moddepend tag found outside a module");
      return;
    }
    int count = parser.getAttributeCount();
    DependProfileModule depend = new DependProfileModule();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("modulename")) {
        depend.setModuleName(parser.getAttributeValue(i).trim());
      } else if (parser.getAttributeName(i).equals("version")) {
        depend.setModuleVersion(parser.getAttributeValue(i).trim());
      }
    }
    moduletoload.addDependProfilModule(depend);
  }
  
  private void processEndElement(KXmlParser parser) {
    String name = parser.getName();

    if (name.equals(XmlModuleParser.TAG_XML)) {
      // Do nothing
    } else if (name.equals(XmlModuleParser.TAG_RUBRIC)) {
      this.closeRubricTag();
    } 
  }

  private void processText(KXmlParser parser) throws XmlPullParserException {
    // Do nothing
  }

  public ProfileModule getModuletoload() {
    return moduletoload;
  }
}
