/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */

package org.openmobileis.module.profiles.terminal;

import org.openmobileis.common.user.profile.Profile;
import org.openmobileis.common.util.codec.GeneralCoder;
import org.openmobileis.common.util.log.*;
import org.openmobileis.module.terminal.ModuleManager;
import org.openmobileis.synchro.client.SynchroDescriptor;
import org.openmobileis.synchro.journal.JournalEntry;
import org.openmobileis.synchro.journal.JournalManager;
import org.openmobileis.synchro.openmsp.OpenMSPException;
import org.openmobileis.synchro.openmsp.client.DefaultOpenMSPSyncListener;
import org.openmobileis.synchro.openmsp.client.OpenMSPDBSyncListener;
import org.openmobileis.synchro.openmsp.client.core.NumSyncManagerDB;
import org.openmobileis.synchro.openmsp.protocol.AbstractCommand;
import org.openmobileis.synchro.openmsp.protocol.Command;
import org.openmobileis.synchro.openmsp.protocol.ContainerMessage;
import org.openmobileis.synchro.openmsp.protocol.DataItem;
import org.openmobileis.synchro.openmsp.protocol.Element;
import org.openmobileis.synchro.openmsp.protocol.Message;
import org.openmobileis.synchro.openmsp.protocol.RequestCommand;
import org.openmobileis.synchro.openmsp.protocol.Result;
import org.openmobileis.synchro.openmsp.protocol.Status;
import org.openmobileis.synchro.security.auth.Credential;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description:  Listener for ProfilService (manage profil synchronisation)
 * Manager priorities for OpenMSP messages (sequence)
 *
 * registered by EmbeddedProfilChannelManager
 *
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public class OpenMSPProfileSyncListener extends DefaultOpenMSPSyncListener {

  public static final String profilSynchroSourceName = "ProfilSyncListener";

  public static final String profilSynchroTargetName = "ProfilSynchroTarget";

  private boolean firstSync = false;

  //  private int syncCMDID = 0;

  public OpenMSPProfileSyncListener() {
  }

  public void receiveResultCommand(ContainerMessage resultContainer, ContainerMessage initialCommand) throws OpenMSPException {
    Result result = (Result) resultContainer.getElement();
    if (result.getTargetRef().equals(profilSynchroSourceName)) {
      // set new profil
      try {
        String encodeddata = ((DataItem) (resultContainer.nextMessage().getElement())).getData();
        String data = new String(GeneralCoder.decodeBase64(encodeddata.getBytes()));
        Profile oldproffile = TerminalProfileManager.getManager().getProfil();
        TerminalProfileManager.getManager().saveProfilFile(data);
        
        //If the profil has changed application must be reloaded and resynchronized if needed.
        Profile newproffile = TerminalProfileManager.getManager().getProfil();
        if (!newproffile.equals(oldproffile))   {
          //reinit ModuleManager with new profil and new profil modules.
          ModuleManager.getManager().initModule();
        }
        
      } catch (Exception e) {
        LogManager.traceError(LogServices.WEBSERVICE, "Wrong result format for profile, impossible to read profile data");
        LogManager.traceError(LogServices.WEBSERVICE, e);
      }
    }
  }

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener#receiveStatusCommand(org.openmobileis.synchro.openmsp.protocol.Status, long)
   */
  public void receiveStatusCommand(Status statusCommande, ContainerMessage initialContenaire) throws OpenMSPException {
    try {
      //      int cmlRef = statusCommande.getCmdRef();
      int status = statusCommande.getStatus();
      if (initialContenaire != null) {
        AbstractCommand command = (AbstractCommand) initialContenaire.getElement();
        this.manageReturnStatusLog(status, command.getElementType(), initialContenaire);
        if (status != Status.STATUS_OK) {
          //set sync in error
          this.notifySynchroFailure();
        }
      }
    } catch (Exception ex) {
      this.notifySynchroFailure();
      LogManager.traceError(0, ex);
    }
  }

  /*
   * Event sent when the sync manager sends a openML message to the server
   * Build a sequence with two or three message :
   * Sync message if the user has updated his profil
   * Get message to get the updated profile from the server
   * Sync message to get synchronized channels (could be all channels in the profile or specific channels)
   * we don't use the context since we don't manage status
   */
  public void sendData(Message message) {
    if (firstSync) {
      try {
        Command getCommand = new Command(Element.GET, this.getSyncName(), profilSynchroTargetName);
        long ns = NumSyncManagerDB.getManager().getSyncNumberForService(this.getSyncName());
        getCommand.setSourceSessionID(ns);
        ContainerMessage getContainer = new ContainerMessage(getCommand);

        // add GET command to make synchro of profil
         DataItem newItem = new DataItem(Element.ITEM, "", "", null, null);

        getContainer.add(newItem);
        message.add(getContainer);
      } catch (OpenMSPException e) {
        LogManager.traceError(LogServices.WEBSERVICE, "Error during building syncML message for Sync Profil service : " + e.getMessage());
      }
      firstSync = false;
    }
  }

  /*
   * Return the name used to identiy the receiver of the syncML messages (source and target)
   */
  public String getSyncName() {
    return profilSynchroSourceName;
  }

  public void startSync(Credential cred, SynchroDescriptor synchrodescriptor) throws OpenMSPException {
    super.startSync(cred, synchrodescriptor);
    firstSync = true;
  }

  /**
   * notify when receiving a status commnand the status code, the CMLElement type of the send command and the container
   * containing the command.
   */
  protected void manageReturnStatusLog(int statusCode, int cmlType, ContainerMessage container) {
    this.setSynchroStatus(statusCode);
    if (statusCode != Status.STATUS_OK) {
      String message = null;
      short journalEntryType = JournalEntry.ERROR_SYNC_MOBILE;
      if (statusCode == Status.STATUS_UNAUTHORIZED) {
        message = OpenMSPDBSyncListener.SYNC_UNAUTHORIZED_MESSAGE;
      } else {
        message = OpenMSPDBSyncListener.SYNC_ERROR_MESSAGE;
      }
      if (message != null) {
        JournalManager.getManager().saveJournalEntry(

        this.getSyncName()

        , journalEntryType

        , message

        , statusCode

        );
      }
    }

  }

}