/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */

package org.openmobileis.common.user.profile.impl;

import org.xml.sax.helpers.DefaultHandler;
import org.xmlpull.v1.XmlPullParserException;

import java.io.IOException;
import java.io.Reader;

import org.kxml2.io.KXmlParser;
import org.openmobileis.common.user.profile.Profile;
import org.openmobileis.common.user.profile.ProfileDataFactory;
import org.openmobileis.common.user.profile.ProfileNotFoundException;
import org.openmobileis.common.user.profile.ProfileProperty;
import org.openmobileis.common.user.profile.ProfileRubric;
import org.openmobileis.common.util.collection.Array;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Manage profil data in an xml file.

 * @author  Philippe Delrieu
 * @since JDK 1.1
 * @version 1.0.
 */
public class TerminalXmlFileProfilDataFactory implements ProfileDataFactory {
  private Profile profil;

  class TerminalProfilListParser  extends DefaultHandler  {
    private final static String TAG_XML = "xml";

    private final static String TAG_PROFILLIST = "PROFILLIST";

    private final static String TAG_PROFILNAME = "PROFILNAME";

    private final static String TAG_ENTRY = "ENTRY";

    private final static String TAG_PROPERTY = "PROPERTY";
    
    
    private String currentProfilName;
    private Array rubricList;
    private Array propertyList;

    private Reader reader;

    TerminalProfilListParser(Reader reader)  {
      this.reader = reader;
      rubricList = new Array(10);
      propertyList = new Array(10);
    }

    public void parse() throws XmlPullParserException, IOException {
      try   {
        KXmlParser parser = new KXmlParser();
        parser.setInput(this.reader);
  
        int eventType = parser.getEventType();
        do {
          if (eventType == KXmlParser.START_DOCUMENT) {
          } else if (eventType == KXmlParser.END_DOCUMENT) {
          } else if (eventType == KXmlParser.START_TAG) {
            this.processStartElement(parser);
          } else if (eventType == KXmlParser.END_TAG) {
            this.processEndElement(parser);
          } else if (eventType == KXmlParser.TEXT) {
            this.processText(parser);
          }
          eventType = parser.next();
        } while (eventType != KXmlParser.END_DOCUMENT);
      } finally {
        reader.close();
      }
    }

    private void processStartElement(KXmlParser parser) {
      String name = parser.getName();

      if (name.equals(TerminalProfilListParser.TAG_XML)) {
        // Do nothing
      } else if (name.equals(TerminalProfilListParser.TAG_PROFILLIST)) {
       //Do nothing
     } else if (name.equals(TerminalProfilListParser.TAG_PROFILNAME)) {
        this.openProfilTag(parser);
     } else if (name.equals(TerminalProfilListParser.TAG_ENTRY)) {
       this.openEntryTag(parser);
     } else if (name.equals(TerminalProfilListParser.TAG_PROPERTY)) {
       this.openPropertyTag(parser);
      } 
    }

    private void openProfilTag(KXmlParser parser) {
      rubricList.clear();
      propertyList.clear();
      int count = parser.getAttributeCount();
      for (int i = 0; i < count; i++) {
        if (parser.getAttributeName(i).equals("name")) {
          currentProfilName = parser.getAttributeValue(i);
        }
      }
    }

    private void openEntryTag(KXmlParser parser) {
      String module = null;
      String rubric = null;
      String description = null;
      String version = null;
      int count = parser.getAttributeCount();
      for (int i = 0; i < count; i++) {
        if (parser.getAttributeName(i).equals("moduleName")) {
          module = parser.getAttributeValue(i);
        } else if (parser.getAttributeName(i).equals("rubricName")) {
          rubric = parser.getAttributeValue(i);
        } else if (parser.getAttributeName(i).equals("description")) {
          description = parser.getAttributeValue(i);
        } else if (parser.getAttributeName(i).equals("version")) {
          version = parser.getAttributeValue(i);
        }
      }
      ProfileRubric rub = new ProfileRubric(module, rubric, null, new ProfileProperty[0]);
      if (description != null)  {
        rub.description = description;
      }
      if (version != null)  {
        rub.version = version;
      }
      rubricList.add(rub);
   }

    private void openPropertyTag(KXmlParser parser) {
      String key = null;
      String value = null;
      String moduleName = null;
      int count = parser.getAttributeCount();
      for (int i = 0; i < count; i++) {
        if (parser.getAttributeName(i).equals("key")) {
          key = parser.getAttributeValue(i);
        } else if (parser.getAttributeName(i).equals("value")) {
          value = parser.getAttributeValue(i);
        } else if (parser.getAttributeName(i).equals("module")) {
          moduleName = parser.getAttributeValue(i);
        }
      }
      ProfileProperty prop = new ProfileProperty(key, value);
      prop.moduleName = moduleName;
      propertyList.add(prop);     
   }
    
    private void processEndElement(KXmlParser parser) {
      String name = parser.getName();

      if (name.equals(TerminalProfilListParser.TAG_XML)) {
        // Do nothing
      } else if (name.equals(TerminalProfilListParser.TAG_PROFILNAME)) {
        profil = new Profile();
        ProfileRubric[] rubs = new ProfileRubric[rubricList.size()];
        rubricList.toArray(rubs);
        profil.profilRubric = rubs;
        ProfileProperty[] props = new ProfileProperty[propertyList.size()];
        propertyList.toArray(props);
        profil.profilProperties = props;
        profil.profilName = currentProfilName;
      } 
    }

    private void processText(KXmlParser parser) throws XmlPullParserException {
      // Do nothing
    }
  }
  
  public TerminalXmlFileProfilDataFactory(Reader reader) throws IOException, XmlPullParserException   {
    TerminalProfilListParser parser = new TerminalProfilListParser(reader);
    parser.parse();   
  }
    
  public Profile getProfil(String profilName)  throws ProfileNotFoundException {
    if (!profilName.equals(profil.profilName)) throw new ProfileNotFoundException("profil not found :"+profilName);
    return profil;     
  }
  
  public Profile getParsedProfil() {
    return profil;
  }

    
}