/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 * 
 */

package org.openmobileis.common.util;

import java.util.Properties;
import java.io.*;

import org.openmobileis.common.context.ApplicationContextManager;

/**
 * Use to manage application properties.
 * Application properties  files should be stored in a folder called "properties" inside the classpath.
 * Not mandatory but its a project practice. Otherwise the manager looks in the classpath for file properties
 * 
 * Implements singleton pattern.
 * Use getManager method to get access to the application resource manager
 * 
 * Propertie file can contain fields that begin with $.
 * These fields are replaced when loaded by system property first that PropertiesManager property
 * For exemple if you define in the propery file the following property :
 * if system user.dir is /home/application
 * 	myproperty=$user.dir/somepath/file
 * The property loaded will be 	myproperty=/home/application/somepath/file
 *
 * @author Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0
 */
public class PropertiesManager {

	public static PropertiesManager instance = null;

	public Properties globalProperties = new java.util.Properties();

	protected PropertiesManager() {
		super();
	}

	/** 
	 * remove all resource manager properties
	 */
	public void removeAllProperties() {
		globalProperties = new java.util.Properties();
	}

	/** 
	 * main method to get ResourceManager
	 * @return application properties manager
	 */
	public static PropertiesManager getManager() {
		if (instance == null) {
			synchronized (PropertiesManager.class) {
				if (instance == null) {
					instance = new PropertiesManager();
					ApplicationContextManager.getManager().addManager(instance);
				}
			}
		}
		return instance;
	}

	/** 
	 * Use to register another ResourceManager. 
	 *
	 * @param manager : properties manager to register
	 *
	 */
	public static void registerInstance(PropertiesManager manager) {
		instance = manager;
		ApplicationContextManager.getManager().addManager(instance);
	}

	public java.io.InputStream getRessourceAsStream(String name) throws java.io.IOException {
		java.io.InputStream in = this.getClass().getResourceAsStream(name);
		return in;
	}

	/** 
	 * load a property file inside the classpath.
	 * if the path is /org/openmobileis/file.properties
	 * First look in the classpath for  /properties/org/openmobileis/file.properties
	 * then in /org/openmobileis/file.properties.
	 * If not found throw IOException
	 *
	 * @param filename : properties file name.
	 * @exception IOException : throws if the file is not found
	 */
	public void addPropertiesFile(String filename) throws IOException {
      InputStream in = null;
      try   {
        in = this.getRessourceAsStream("/properties" + filename);
        if (in == null) in = this.getRessourceAsStream(filename);
        if (in == null) in = new FileInputStream(filename);
      } catch (IOException ex)  {
        try {
          in = this.getRessourceAsStream(filename);
        } catch (IOException exi)    {
          in = new FileInputStream(filename);
        }
      }
    	try {
    		globalProperties.load(this.convertPropertiesFile(in));
    	} finally {
    		in.close();
    	}

	}

	/** 
	 * load a property file from the file system
	 * If not found throw IOException
	 *
	 * @param filename : properties file name.
	 * @exception IOException : throws if the file is not found
	 */
	public void addPropertiesFileFromFilePath(String filename)
			throws java.io.IOException {
		InputStream in = new java.io.FileInputStream(filename);
		try {
			globalProperties.load(this.convertPropertiesFile(in));
		} finally {
			in.close();
		}
	}

	/** 
	 * read the property file inside the classpath.
	 * Does not load it
	 * if the path is /org/openmobileis/file.properties
	 * First looks in the classpath for  /properties/org/openmobileis/file.properties
	 * then for /org/openmobileis/file.properties.
	 * If not found throw IOException
	 *
	 * @param filename : properties file name.
	 * @exception IOException : throws if the file is not found
	 */
	public java.util.Properties getProperties(String prop_file)
			throws java.io.IOException {
		Properties prop = new Properties();
        InputStream in = null;
        try {
          in = this.getRessourceAsStream("/properties" + prop_file);
          if (in == null) in = this.getRessourceAsStream(prop_file);
          if (in == null) in = new FileInputStream(prop_file);
       } catch (IOException ex)    {
          try   {
            in = this.getRessourceAsStream(prop_file);
          } catch (IOException exi)  {
            in = new FileInputStream(prop_file);
          }
        }
		try {
			prop.load(this.convertPropertiesFile(in));
		} finally {
			in.close();
		}
		return prop;
	}

	/** 
	 * read the property file from the file system
	 * Does not load it
	 *
	 * @param filename : properties file name.
	 * @exception IOException : throws if the file is not found
	 */
	public java.util.Properties getPropertiesFromPath(String prop_file)
			throws java.io.IOException {
		Properties prop = new Properties();
		InputStream in = new java.io.FileInputStream(prop_file);
		try {
			prop.load(this.convertPropertiesFile(new BufferedInputStream(in)));
		} finally {
			in.close();
		}
		return prop;
	}

	/** 
	 * return all the loaded properties
	 *
	 * @return the loaded properties
	 */
	public java.util.Properties getProperties() {
		return globalProperties;
	}

	/** 
	 * return the property with specified key from loaded properties.
	 * return default value if not found.
	 *
	 * @param key : the property key to get
	 * @param defaultValue : default value if key is not found
	 * @return the property with specified key or default value if not found.
	 */
	public String getProperty(String key, String defaultValue) {
		return globalProperties.getProperty(key, defaultValue);
	}

	/** 
	 * return the property with specified key from loaded properties.
	 * return null if not found.
	 *
	 * @param key : the property key to get
	 * @return the property with specified key or null if not found.
	 */
	public String getProperty(String key) {
		return globalProperties.getProperty(key);
	}

	/** 
	 * add the specified key,value to the manager properties 
	 *
	 * @param key : the property key
	 * @param value : value of the property
	 */
	public void addProperty(String key, String value) {
		globalProperties.put(key, value);
	}

	protected InputStream convertPropertiesFile(InputStream inStream)
			throws java.io.IOException {
		int r;
		char c;
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		StringBuffer propertyName = new StringBuffer();

		boolean readPropName = false;
		while ((r = inStream.read()) != -1) {
			c = (char) r;
			if (c == '$') {
				readPropName = true;
				propertyName.setLength(0);
				continue;
			} else if (readPropName) {
				if (((c >= 'a') && (c <= 'z')) || (c == '.')
						|| ((c >= 'A') && (c <= 'Z'))
						|| ((c >= '0') && (c <= '9'))) { // test end of the property. Property use normal caractere and dot (.)
					propertyName.append(c);
					continue;
				} else { // end property name
					readPropName = false;
					String propName = propertyName.toString();
					String prop = System.getProperty(propName);
					if (prop == null) {
						prop = globalProperties.getProperty(propName, "null");
					}

					// test if the property contains anti slash value. They must be doubled because
					// property load remove slash. PB found when using user.home property on windows
					int index = prop.indexOf("\\");
					if (index != -1) {
						byte[] tab = prop.getBytes();
						for (int i = 0; i < tab.length; i++) {
							if (tab[i] == '\\') {
								out.write('\\');
							}
							out.write(tab[i]);
						}
					} else {
						out.write(prop.getBytes());
					}
				}
			}
			out.write(c);
		}
		ByteArrayInputStream retin = new ByteArrayInputStream(out.toByteArray());
		return retin;
	}

}
