/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */
package org.openmobileis.common.util.collection.tree;

import org.openmobileis.common.util.collection.Array;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Manage listener order during different synchro phase.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class DependTree {
  private DependTreePhaseList firstPhase, currentPhase;
  private static final Array emptyArray = new Array();

  /**
   * 
   */
  public DependTree() {
    super();
  }
  
  public Object getObjectByName(String listenerName) {
    DependTreePhaseList phase = firstPhase;
    while (phase != null) {
      DependNode node = this.getNodeFromPhaseList(listenerName);
      if (node != null) {
        return node.nodeObject;
      }
      phase = phase.nextPhase;
    }
    return null;
  }
  
  public void addObject(String name, Object object, String[] dependList) {
    if (firstPhase == null)  {
      firstPhase = new DependTreePhaseList(1);
    }
    Array depends = null;
    if (dependList == null) {
      depends = emptyArray;
    } else  {
     depends = new Array(dependList);
    }
    //first add listener dependance.
    Array depArray = new Array();
    int dependssize = depends.size();
    for (int i=0; i<dependssize; i++)  {
     DependNode depnode = new DependNode();
     depnode.nodeName = (String)depends.get(i);
     this.addDependNodeToList(depArray, firstPhase, depnode); 
    }
    //add new listener
    DependNode newnode = new DependNode();
    newnode.nodeName = name;
    newnode.nodeObject = object;
    
    //process depends will not all treated. Always end because all depends has already been added.
    do  {
      depends = this.addDependNodeToList(depends, firstPhase, newnode);  
    } while (depends.size() != 0);
  }
  
  public void resetPhaseList()  {
    //set all synchro listener node to OK
    DependTreePhaseList phase = firstPhase;
    while (phase != null) {
      int size = phase.nodeList.size();
      for (int i=0; i<size; i++)  {
        DependNode node = (DependNode) phase.nodeList.get(i);
        node.validState = true;
      }
      phase = phase.nextPhase;
    }
    currentPhase = firstPhase;
  }
  
  public Array getNextPhaseObjectList()  {
    if (currentPhase == null) return DependTree.emptyArray;
    Array listenerList = new Array();
    //remove not ok listener
    int size = currentPhase.nodeList.size();
    for (int i=0; i<size; i++)  {
      DependNode node = (DependNode) currentPhase.nodeList.get(i);
      if (node.validState)  {
        listenerList.add(node.nodeObject);
      }
    }
    currentPhase = currentPhase.nextPhase;
    return listenerList;
 }
  
  public void invalidateNode(String listenerName) {
    DependNode listener = this.getNodeFromPhaseList(listenerName);
    if (listener != null) {
      invalidDependNodeChilds(listener);
    }
  }
  
  protected void invalidDependNodeChilds(DependNode listener)  {
    listener.validState = false;
    int size = listener.nexts.size();
    for (int i=0; i<size; i++)  {
      this.invalidDependNodeChilds((DependNode)listener.nexts.get(i));
    }
  }
  
  private DependNode getNodeFromPhaseList(String listenerName)  {
    DependTreePhaseList phase = firstPhase;
    while (phase != null) {
      int size = phase.nodeList.size();
      for (int i=0; i<size; i++)  {
        DependNode node = (DependNode) phase.nodeList.get(i);
        if (node.nodeName.equals(listenerName)) {
          return node;
        }
      }
      phase = phase.nextPhase;
    }
    return null;
  }
  
  private Array addDependNodeToList(Array depends, DependTreePhaseList phase, DependNode newnode) {
          
    int phasesize = phase.nodeList.size();
    DependNode phaseNode = null;
   
   //test if node already present process already added node
    for (int j=0; j<phasesize; j++)  {
      phaseNode = (DependNode)phase.nodeList.get(j);
      if (newnode.nodeName.equals(phaseNode.nodeName)) {
        if (newnode.nodeObject != null) {
          phaseNode.nodeObject = newnode.nodeObject;
        }
        // node already in a phase, use already added node to keep node context data.
         newnode = phaseNode;
      }    
    }
    
    //if no dependence add listener
    if (depends.size() == 0)  {
      //validate if not already added in posterior phase.
      if ((newnode.nodePhase != null) && (newnode.nodePhase.phaseNumber < phase.phaseNumber)) {
        newnode.nodePhase.nodeList.remove(newnode);
        newnode.nodePhase = null;
      }
      //if not in a phase add it. //we suppose that all dependance has been added.
      if (newnode.nodePhase == null)  {
        //only add if not already added. Avoid multiple depend add.
        if(this.getNodeFromPhaseList(newnode.nodeName)== null)  {
          newnode.nodePhase = phase;
          phase.nodeList.add(newnode);
          if (newnode.nexts.size() >0)  {
            //move all dependant listerner to next phase and dependent dependant listener and so on.
            movedependentListenerTonextPhase(newnode);
          }
        }
      }
      return DependTree.emptyArray;
    }   
    
    // validate dependance if depend listener is already registered
   Array newdepends = (Array)depends.clone();
   int dependssize = depends.size();
   
   Array clonePhaseNodeList = (Array) phase.nodeList.clone();
   for (int i=0; i<dependssize; i++)  {
      for (int j=0; j<phasesize; j++)  {
        phaseNode = (DependNode)clonePhaseNodeList.get(j);
        // test if depends are in this phase.
        if (depends.get(i).equals(phaseNode.nodeName)) {
          newdepends.remove(phaseNode.nodeName);
          phaseNode.nexts.add(newnode);
          if (phase.nextPhase == null)  {
            phase.nextPhase = new DependTreePhaseList(phase.phaseNumber+1);
          }
          newdepends = this.addDependNodeToList(newdepends, phase.nextPhase, newnode);
        }
      }
      if ((phase.nextPhase != null) && (depends.size() >0))  { //perhaps depends are in next phase.
        newdepends = this.addDependNodeToList(newdepends, phase.nextPhase, newnode);
      }
    }    
    return newdepends;
  }

  private void movedependentListenerTonextPhase(DependNode node)  {
    Array othernodelist = node.nexts;
    node.nexts = new Array();
    Array newdepends = new Array();
    newdepends.add(node.nodeName);
    for (int i=0; i<othernodelist.size(); i++)  {
      DependNode listener = (DependNode) othernodelist.get(i);
      
      //PB if depend are in next Phase not added.
      this.addDependNodeToList(newdepends, listener.nodePhase, listener);
      this.movedependentListenerTonextPhase(listener);
    }
    
  }
}


