/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 * 
 */

package org.openmobileis.common.util.database;

import java.sql.ResultSet;
import org.openmobileis.common.util.exception.DatabaseException;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;

/**
 * Title: OpenMobileIS project source <BR>
 * Description: This class is ths super class of the queries It allows to execute queries through the db manager. A query is built with : A pattern : SQL query
 * with variables Parameters : Values of the variables defined in the pattern For instance, the query "SELECT * FROM TABLE WHERE ROW1 = val1 AND ROW2 = val2
 * will have the following pattern : "SELECT * FROM TABLE WHERE ROW1 = %0% AND ROW2 = %1% and the following parameters : {"val1", "val2"}. Note that the name of
 * the variables are a sequential number starting by 0 and inserted between two separators : % The Manager variable contains dbManager instance which provides
 * services to be connected / disconnected to the database as well as a service to get a connection from the pool connection object.
 * 
 * @author Philippe Delrieu
 * @since JDK 1.1
 * @version 1.0.
 */

public abstract class AbstractQueryManager {

  // Reference the DB manager and provide a statement to perform queries
  //  static protected dbManager Manager;

  // separator for query variables
  static char Separator = '%';

  private ManagerDB dbManager;

  public AbstractQueryManager() {
  }

  public void registerManagerDB(ManagerDB manager) {
    dbManager = manager;
  }

  /**
   * connectDBManager() allows to access to the DB manager to get a statement
   * 
   * @param ManagerDB
   * @return None
   * @exception None
   */
  /*
   * public static void connectDBManager (ManagerDB db) throws Exception { if (db != null) Manager = db; }
   */

  /**
   * close() allows to close the database
   * 
   * @param None
   * @return None
   * @exception None
   */
  public void close() {
    this.getDbManager().disconnect();
  }

  /**
   * execute insert or update query with Blob data. Query parameters are indicated with %0% tag and blob data with ? tag.
   * Use SQL syntax to perform queries
   * 
   * @param String
   *          representing the query with parameters in tag
   * @param query
   *          parameters.
   * @param blob
   *          data.
   * @return void
   * @exception DatabaseException
   */
  protected void executeBlobUpdate(String queryPattern, String parameters[], byte[][] blobs) throws DatabaseException {
    try {
      String query = this.buildQuery(queryPattern, parameters);
      Connection connection = this.getDbManager().getConnection();
      if (connection != null) {
        try {
          PreparedStatement stat = connection.prepareStatement(query);
          for (int i = 0; i < blobs.length; i++) {
            stat.setBytes(i + parameters.length, blobs[i]);
          }
          stat.execute();
        } finally {
          this.getDbManager().garbageOpenedConnection();
        }
      } else {
        throw new DatabaseException("Error during Query execution. No connection available.");
      }
    } catch (SQLException ex) {
      throw new DatabaseException(ex);
    }

  }

  /**
   * executeQuery() allows to execute a query such as "SELECT * FROM TABLE") use SQL syntax to perform queries Connection is not close to keep the resulting
   * resultset. It must be garbaged by calling ManagerDB.getManager().garbageOpenedConnection();;
   * 
   * @param String
   *          representing the query
   * @param query
   *          parameters.
   * @return ResultSet : don't forget to close the resultSet after using.
   * @exception SQLException
   */
  protected ResultSet executeQuery(String queryPattern, String parameters[]) throws DatabaseException {
    //  LogManager.traceDebug(0, "executeQuery date:"+System.currentTimeMillis());

    String query = "";
    try {
      ResultSet result = null;
      query = this.buildQuery(queryPattern, parameters);
      //   System.out.println("query : "+query);

      //     LogManager.traceWarning(0, "AbstractQueryManager executeQuery query :"+query);

      Connection connection = this.getDbManager().getConnection();
      if (connection != null) {
        Statement stat = connection.createStatement();
        result = stat.executeQuery(query);
      } else {
        throw new DatabaseException("Error during Query execution. No connection available.");
      }
      //	LogManager.traceDebug(0, "executeQuery END date:"+System.currentTimeMillis());
      return result;
    } catch (java.sql.SQLException ex) {
      throw new DatabaseException(ex);
    }
  }

  /**
   * executeQuery() allows to execute a query such as "SELECT * FROM TABLE") use SQL syntax to perform queries Connection is not close to keep the resulting
   * resultset. It must be garbaged by calling ManagerDB.getManager().garbageOpenedConnection();;
   * 
   * @param String
   *          representing the query
   * @return ResultSet : don't forget to close the resultSet after using.
   * @exception SQLException
   */
  protected ResultSet executeDynamicQuery(String queryPattern, java.util.Hashtable variables) throws DatabaseException {
    String query = "";
    try {
      ResultSet result = null;
      query = this.buildDynamicQuery(queryPattern, variables);
      Connection connection = this.getDbManager().getConnection();
      if (connection != null) {
        Statement stat = connection.createStatement();
        result = stat.executeQuery(query);
      } else {
        throw new DatabaseException("Error during Query execution. No connection available.");
      }
      return result;
    } catch (java.sql.SQLException ex) {
      throw new DatabaseException(ex);
    }
  }

  /**
   * executeUpdateQuery() allows to execute a query with INSERT, UPDATE or DELETE statements use SQL syntax to perform queries
   * 
   * Close and garbage the connection and pending result.
   * 
   * @param String
   *          representing the query
   * @return None
   * @exception SQLException
   */
  protected void executeUpdate(String queryPattern, String parameters[]) throws DatabaseException {
    String query = "";
    try {
      query = this.buildQuery(queryPattern, parameters);
      //System.out.println("update query : "+query);
      Connection connection = this.getDbManager().getConnection();
      if (connection != null) {
        try {
          Statement stat = connection.createStatement();
          stat.executeUpdate(query);
          stat.close();

        } finally {
          this.getDbManager().garbageOpenedConnection();
        }
      } else {
        throw new DatabaseException("Error during Query execution. No connection available.");
      }
    } catch (java.sql.SQLException ex) {
      throw new DatabaseException(ex);
    }
  }

  /**
   * executeUpdateQuery() allows to execute a query with INSERT, UPDATE or DELETE statements use SQL syntax to perform queries Connection is not close to keep
   * the resulting resultset. It must be garbaged by calling ManagerDB.getManager().garbageOpenedConnection();;
   * 
   * @param String
   *          representing the query
   * @return None
   * @exception SQLException
   */
  protected void executeDynamicUpdate(String queryPattern, java.util.Hashtable variables) throws DatabaseException {
    String query = "";
    try {
      query = this.buildDynamicQuery(queryPattern, variables);
      Connection connection = this.getDbManager().getConnection();
      if (connection != null) {
        try {
          Statement stat = connection.createStatement();
          stat.executeUpdate(query);
          stat.close();
        } finally {
          this.getDbManager().garbageOpenedConnection();
        }
      } else {
        throw new DatabaseException("Error during Query execution. No connection available.");
      }
    } catch (java.sql.SQLException ex) {
      throw new DatabaseException(ex);
    }
  }

  /*
   * Build dynamically the query from a query pattern. The variables in the pattern are set as ${var} and the hashtable contains an entry var-> value.
   */
  protected String buildDynamicQuery(String pattern, java.util.Hashtable variables) {
    StringBuffer result = new StringBuffer();
    int beginIndex = 0;
    int endIndex = 0;
    while (endIndex != -1) {
      endIndex = pattern.indexOf("${", beginIndex);
      if (endIndex == -1)
        result.append(pattern.substring(beginIndex, pattern.length()));
      else { // find one variable
        result.append(pattern.substring(beginIndex, endIndex));
        beginIndex = endIndex;
        endIndex = pattern.indexOf("}", beginIndex);
        if (endIndex != -1) {
          String varName = pattern.substring(beginIndex + 2, endIndex);
          if (variables.containsKey(varName))
            result.append((String) variables.get(varName));
          beginIndex = endIndex + 1; // skip '}' character
          //} // end if containsKey
        } // end if index != -1
      } // end if find variable
    } // end while
    return result.toString();
  }

  /**
   * buildQuery() get the query pattern stored in queries and insert the parameters.
   * 
   * @param String
   *          QueryPattern representing the query with the variables
   * @param String
   *          parameters representing the variable values or piece of sql codes %x% represents a variable and %code% represents a piece of sql code to replace
   * @return String : definitive query
   * @exception Exception
   */
  protected String buildQuery(String queryPattern, String parameters[]) throws DatabaseException {
    // Case of zero parameters
    if (parameters == null) {
      return queryPattern;
    }

    StringBuffer newQuery = new StringBuffer();
    int currentIndex = 0;
    int startIndex = 0;
    int lastIndex = 0;
    String variable;
    String value;

    for (int i = 0; i < parameters.length; i++) {
      currentIndex = queryPattern.indexOf(Separator, startIndex);
      //remove % tag for SQL like search
      if (queryPattern.charAt(currentIndex + 1) == Separator) {
        currentIndex += 1;
      }
      lastIndex = queryPattern.indexOf(Separator, currentIndex + 1);
      variable = queryPattern.substring(currentIndex + 1, lastIndex);

      if (variable.equals(String.valueOf(i))) {
        // replace %i by the real parameter or NULL
        newQuery.append(queryPattern.substring(startIndex, currentIndex));
        value = parameters[i];
        if ((value == null)) { // || (value.length() == 0)
          newQuery.append("NULL");
        } else {
          //        newQuery.append("\"");
          newQuery.append(parameters[i]);
          //        newQuery.append("\"");
        } // end test null
        startIndex = lastIndex + 1;
        //remove % tag for SQL like search
        if ((startIndex < queryPattern.length()) && (queryPattern.charAt(startIndex) == Separator)) {
          startIndex += 1;
          newQuery.append(Separator);
        }
      } else {
        throw new DatabaseException("Parsing error in query : " + queryPattern);
      }
    } // end for
    // copy the end of the pattern
    newQuery.append(queryPattern.substring(startIndex));
    return newQuery.toString();
  }

  /*
   * build a string in SQL list format : ('item1', 'item2',...) if vector size == 0, then return empty string
   */
  protected String makeSQLList(String[] items) {
    if (items == null)
      return "";
    StringBuffer buffer = new StringBuffer("(");
    for (int i = 0; i < items.length; i++) {
      //     buffer.append("'");
      buffer.append((String) items[i]);
      //     buffer.append("'");
      if (i < items.length - 1)
        buffer.append(",");
    } // end for
    buffer.append(")");
    return buffer.toString();
  }

  /**
   * Returns the dbManager.
   * 
   * @return ManagerDB
   */
  public ManagerDB getDbManager() {
    if (dbManager == null)
      dbManager = ManagerDB.getManager();
    return dbManager;
  }

}
