/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 * 
 */

package org.openmobileis.common.util.database;

import java.sql.*;
import java.util.HashMap;
import com.bitmechanic.sql.*;

import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.common.util.log.*;
import org.openmobileis.common.util.exception.ServiceException;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description:  
      * This class represents the manager for the SGBDR .
      * It allows to connect a database by giving an url and a jdbc driver
      * It establishes the connection, prepare statement and manage the application
      * queries (using pool connections)
      * All the DB parameters are contained in the database.properties file which
      * is read by the InitFactories class.
 * @author  Philippe Delrieu
 * @since JDK 1.1
 * @version 1.0.
 */

public class JdbcPoolManagerDB extends ManagerDB {

   // Store the connection of the database
   private ConnectionPoolManager poolManager;
   private static String DEFAULT_ALIAS = "JDBCPoolManagerDB";
   private String poolAlias;
   private String shutdownQuery = null;
   private HashMap openedConnection;
   private int monitorInterval=0;
   private String driver;
   private String server = null;
   private String login = null;
   private String password = null;
   private int maxConnections = 0;
   private int idleTimeout = 0;
   private int checkoutTimeout = 0;
   private int maxCheckout = 0;


/**   * Connect to the database
  * @param     monitor interval : how many seconds to wait between reaping connections in the pool.
  * @param     String driver : jdbc driver bridge (ex: "com.imaginary.sql.msql.MsqlDriver")
  * @param     String server url : Name of the database "jdbc:msql:databaseName"
  * @param     String login
  * @param     String password
  * @param     int maximum available connections
  * @param     idle timeout : seconds a connection can be idle before it is closed
  * @param     int checkout timeout : seconds a connection can be checked out by a thread before it is returned back to the pool
  * @param     int maxCheckout : number of times a connection can be re-used before connection to database is closed and re-opened
  * @param     String shutdown query called before closing database. For instance, hsql need to query "shutdown compact" explicitly before
  * closing (can be null)
  * @return    None
  * @exception IOException for log file
  * @exception Exception : Impossible to find the database or to find the driver
*/
  public JdbcPoolManagerDB () {
  	this(
  			DEFAULT_ALIAS
  			,PropertiesManager.getManager().getProperty("server.database.dbDriver")
        , PropertiesManager.getManager().getProperty("server.database.dbServer")
        , PropertiesManager.getManager().getProperty("server.database.dbLogin")
        , PropertiesManager.getManager().getProperty("server.database.dbPassword")
        , Integer.parseInt(PropertiesManager.getManager().getProperty("server.database.pool.monitorInterval"))
        , Integer.parseInt(PropertiesManager.getManager().getProperty("server.database.pool.maxConns"))
        , Integer.parseInt(PropertiesManager.getManager().getProperty("server.database.pool.idleTimeout"))
        , Integer.parseInt(PropertiesManager.getManager().getProperty("server.database.pool.checkoutTimeout"))
        , Integer.parseInt(PropertiesManager.getManager().getProperty("server.database.pool.maxCheckout"))
        , PropertiesManager.getManager().getProperty("server.database.shutDownQuery")
     );
  }
  public JdbcPoolManagerDB (
  			String alias
		  	,String srvdriver
		  	, String dbserver
		  	, String srvlogin
		  	, String srvpassword
		  	, int monitorinterval
		  	, int maxconnection
		  	, int idltimeout
		  	, int checkouttimeout
		  	, int maxcheckout
		  	, String shutdownquery
  		) {
  	poolAlias = alias;
  	driver = srvdriver;
   	monitorInterval=monitorinterval;
   	server = dbserver;
   	login = srvlogin;
   	password = srvpassword;
   	maxConnections = maxconnection;
   	idleTimeout = idltimeout;
   	checkoutTimeout = checkouttimeout;
   	maxCheckout = maxcheckout;
    this.shutdownQuery = shutdownquery;
    try {

      if ( driver == null) {// no DB
          LogManager.traceInfo(LogServices.DATABASESERVICE, "Undefined DB");
          return;
      }
      this.connect();

    } catch (Exception e) {
      LogManager.traceAlert(LogServices.SERVERSERVICE, new ServiceException("database connection error", e));
    }
  }
  
  public void connect() throws ServiceException {
    try {
	
	    poolManager = new ConnectionPoolManager(monitorInterval);
	    Class.forName(driver).newInstance();
	    poolManager.addAlias(poolAlias, driver, server, login, password, maxConnections, idleTimeout, checkoutTimeout, maxCheckout);
	    if (DatabaseConst.debug)  {
	      poolManager.getPool(poolAlias).setTracing(true);
	    }
	    openedConnection = new HashMap(maxConnections);
	    LogManager.traceInfo(LogServices.DATABASESERVICE, "DB open successfully");
    } catch (Exception e) {
      LogManager.traceAlert(LogServices.SERVERSERVICE, new ServiceException("database connection error", e));
      throw new ServiceException(e);
    }
  }

/**
* disconnect() method allows to close the database
* @param     None
* @return    None
* @exception SQLException
*/
  public void disconnect () {
    try {
      if (DatabaseConst.debug)  {
        // before closing, retreive the stats of the current db session
        ConnectionPool p = poolManager.getPool(poolAlias);
        LogManager.traceInfo(LogServices.SERVERSERVICE, " Current size: " + p.size() + " of " + p.getMaxConn());
        LogManager.traceInfo(LogServices.SERVERSERVICE, "  Connection requests: " + p.getNumRequests());
        LogManager.traceInfo(LogServices.SERVERSERVICE, "  Number of waits: " + p.getNumWaits());
        LogManager.traceInfo(LogServices.SERVERSERVICE, "  Number of timeouts: " +p.getNumCheckoutTimeouts());
      }
      // perform shutdown if necessary
      if (shutdownQuery != null) {
        Connection connection = this.getConnection();
        if (connection != null) {
          Statement stat = connection.createStatement();
          stat.executeUpdate(shutdownQuery);
          stat.close();
          connection.close();
        }
        if (DatabaseConst.debug)  {
          LogManager.traceInfo(LogServices.SERVERSERVICE, "executed shutdown : " + shutdownQuery);
        }
      }
      // close pool connections
      poolManager.removeAlias(poolAlias);
      openedConnection = null;
      if (DatabaseConst.debug)  {
        LogManager.traceInfo(LogServices.SERVERSERVICE, "The DB is closed");
      }
    }  catch (Exception e) {
      LogManager.traceAlert(LogServices.SERVERSERVICE, new ServiceException("Impossible to close the database", e));
    }
  }
  
  public boolean isConnected()	{
  	return (openedConnection !=null);
  }

/**
* getConnection() return a free connection or null if all the connections
* are used after a given time out
* @param  :   None
* @return :   Connection
* @exception : None
*/
  public Connection getConnection () throws java.sql.SQLException {
   // return DriverManager.getConnection(ConnectionPoolManager.URL_PREFIX + alias);
   if (! this.isConnected()) {
   		try	{
   			this.connect();
   		} catch (ServiceException ex)	{
   			LogManager.traceError(LogServices.SERVERSERVICE, ex);
   			throw new SQLException(ex.getMessage());
   		}
   }
   this.garbageOpenedConnection();
   Connection connection = poolManager.getPool(poolAlias).getConnection();
   openedConnection.put(Thread.currentThread(), connection);
   return connection;
  }

  public void garbageOpenedConnection()  {
    Thread currentThread = Thread.currentThread();
    Object obj = openedConnection.get(currentThread);
    while (obj != null) {
      try {
        ((Connection)obj).close();
      } catch (SQLException ex) {
        LogManager.trace(new ServiceException("JdbcPoolManagerDB error during garbage of connexion", ex));
      }
      openedConnection.remove(currentThread);
      obj = openedConnection.get(currentThread);
   }
  }


 }