/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 * 
 */


package org.openmobileis.common.util.exception;

import java.io.*;

import org.openmobileis.common.util.OpenMISSerializable;
import org.openmobileis.common.util.log.*;

/**
 * The root class of all exceptions for Java OpenMobileIS developement.
 *
 * Usage Note : exception can be created from a message, an another exeption (in this case the other
 * exception become nested) with or without a priority code.
 * The priority code correspond to the org.openmobileis.common.util.log.LogPriorities
 * class defined code. Default priority code is LogPriorities.ERROR.
 * To facilitate exception log a default service and priority is defined.
 * A Service can be associated to an exception. It can be a good practice to identify the part of the program
 * that throws the exception during exception log.
 * Default services are defined in the class org.openmobileis.common.util.log.LogServices
 *
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 * @see org.openmobileis.common.util.exception.DefaultException
 **/

//The class is abstract so you must specialize it to use it for your own purpose
public abstract class OpenMISException extends Exception implements	OpenMISSerializable {
  static final long serialVersionUID = 5521257935120563452L;

	//can be null
	private Throwable associatedException = null;

	private int priority;

	public OpenMISException(String msg) {
		super(msg);
		priority = LogPriorities.ERROR;
	}

	public OpenMISException(String msg, int logPriority) {
		super(msg);
		priority = logPriority;
	}

	public OpenMISException(String msg, Throwable t) {
		super(msg);
		priority = LogPriorities.ERROR;
		associatedException = t;
	}

	public OpenMISException(String msg, Throwable t, int logPriority) {
		super(msg);
		priority = logPriority;
		associatedException = t;
	}

	public OpenMISException() {
		super();
		priority = LogPriorities.ERROR;
	}

	public OpenMISException(int logPriority) {
		super();
		priority = logPriority;
	}

	public OpenMISException(Throwable t) {
		this(t.getMessage());
		associatedException = t;
	}

	public OpenMISException(Throwable t, int logPriority) {
		this(t.getMessage());
		associatedException = t;
		priority = logPriority;
	}

	public Throwable getNestedError() {
		return associatedException;
	}

	public final int getPriority() {
		return priority;
	}

	public void printStackTrace() {
		printStackTrace(System.err);
	}

	public void printStackTrace(PrintStream ps) {
		printStackTrace(new PrintWriter(new OutputStreamWriter(ps)));
	}

	public void printStackTrace(PrintWriter writer) {
		if (null != getNestedError()) {
			getNestedError().printStackTrace(writer);
		}
		//append the original message as well as the localized one (when relevant)
		//    String message = getMessage();
		writer.println(this.getClass().getName() + "; service="
				+ LogServices.toString(getService()) + "; priority="
				+ LogPriorities.toString(getPriority()) + " : " + getMessage());
		//the localized one is output-ed by the default stack trace...
		super.printStackTrace(writer);
		writer.flush();
	}

	/**
	 * return  the service assocated to the exception
	 * @return int the service int identified associated with the execption
	 * @see org.openmobileis.common.util.log.LogServices
	 */

	public abstract int getService();
}