/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 * 
 */

package org.openmobileis.common.util.log;

import java.util.*;

import org.openmobileis.common.context.ApplicationContextManager;
import org.openmobileis.common.util.exception.OpenMISException;

import java.io.CharArrayWriter;
import java.io.PrintWriter;
import java.io.PrintStream;



/**
* This class permits to manage log. LogManager is an abstract class to 
* hide the underneath log. It can be mapped to Log4J or other log API.
* We provide default log for console and file log.
* Any other log manager can be plug.
* 
* Default log id define with tree parameters :
*   service : the service that log. defined by the class LogServices.
*   priority : the log priority defined by the class LogPriorities.
*   the log object. The toString method is call on the object except for exception where the
*   printStackTrace methode is call.
*
* Some helper method is defined to faclitate the use of log that define default values for Service or priority.
*
* LogManager is initialized by calling the getInstance method.
* If null is provided, default console logging is set.
* All log are done using LogManager class.
*
* <code>
* <pre>
* exemple :
*   ....
*  public void methodRed () {
*    ....
*    LogManager.traceInfo(0, "Log in methodRed");
*    ....
*    ....
*    try {
*    } catch (java.lang.OpenMSPException exec ) {
*      LogManager.traceError(0, exec);
*    }
*
*  }
*</pre>
*</code>
*
* @author P.Delrieu
* @version 1.0
* @since JDK1.1
* @see:org.openmobileis.common.util.log.LogPriorities
* @see:org.openmobileis.common.util.log.LogServices
*/

public class LogManager {

  /**
  * Static instance of LogManager
	*/
  protected static LogManager instance = null;
  protected LogTracer tracer;
  protected static HashMap logTable;


  /**
	* The constructor is protected in order to use the class <code>LogManager</code>
    * like a singleton.
	*/
  protected LogManager(Properties props) {
    this();
    tracer.setLogTracerProperties(props);
    tracer.setLogStream(new PrintStream( System.out,true ));
    ApplicationContextManager.getManager().addManager(this);
  }

  protected LogManager() {
    super();
    logTable = new HashMap(10);
    tracer = new DefaultLogTracer();
  }

  /**
   * attache the specified logTracer to a priority. No attached priority use the default logTracer to trace.
   * Attached priority will use the specified tracer to log.
   */
  public static  void attachLogTracerToPriority(int priority, LogTracer t)  {
    logTable.put(new Integer(priority), t);
  }


  /**
	* <FONT COLOR="#009999"><B>
	* This methods return the only <code>LogManager</code> instance.
	* </B></FONT>
  * @return the only instance
	*/
  public synchronized static LogManager getInstance() {
    return instance;
  }
  
  public synchronized static void registerLogManager(Properties props) {
	instance = new LogManager(props);
	// put in application context to avoid class garbadge
	ApplicationContextManager.getManager().addManager(instance);
  }

  public static void trace(OpenMISException exp) {
    instance.trace(exp.getService(), exp.getPriority(), exp);
  }

  /**
   * Trace the Object with specified service and priority to the Log system.
   * If object is an exception the track is traced.
  */
  public void trace(int service, int priority, Object obj) {
    String message = null;
    if (obj == null) {
      message = "ERROR Null message Logged";
    } else if (obj instanceof Exception)  {
      CharArrayWriter stream = new CharArrayWriter();
      ((Exception)obj).printStackTrace(new PrintWriter(stream));
      stream.flush();
      stream.close();
      message = new String(stream.toCharArray());
    } else {
      message = obj.toString();
    }
    if (tracer == null) {
      System.out.println(message);
    } else  {
      LogTracer tracelog = (LogTracer) logTable.get(new Integer(priority));
      if (tracelog == null) {
        tracelog = tracer;
      }
      tracelog.trace(service, priority, message);
    }

  }

   /**
   * Trace the Object in EMERGENCY priority with specified service to the Log system.
   * If object is an exception the track is traced.
   */
  public static void traceEmergency(int service, Object obj) {
    instance.trace(service, LogPriorities.EMERGENCY, obj);
  }

   /**
   * Trace the Object in CRITIQUE priority with specified service to the Log system.
   * If object is an exception the track is traced.
   */
  public static void traceCritique(int service, Object obj) {
    instance.trace(service, LogPriorities.CRITIQUE, obj);
  }

   /**
   * Trace the Object in ALERT priority with specified service to the Log system.
   * If object is an exception the track is traced.
   */
  public static void traceAlert(int service, Object obj) {
    instance.trace(service, LogPriorities.ALERT, obj);
  }

  /**
   * Trace the Object in ERROR priority with specified service to the Log system.
   * If object is an exception the track is traced.
   */
  public static void traceError(int service, Object obj) {
    instance.trace(service, LogPriorities.ERROR, obj);
  }

  /**
   * Trace the Object in WARNING priority with specified service to the Log system.
   * If object is an exception the track is traced.
   */
  public static void traceWarning(int service, Object obj) {
    instance.trace(service, LogPriorities.WARNING, obj);
  }

  /**
   * Trace the Object in NOTICE priority with specified service to the Log system.
   * If object is an exception the track is traced.
   */
  public static void traceNotice(int service, Object obj) {
    instance.trace(service, LogPriorities.NOTICE, obj);
  }

  /**
   * Trace the Object in INFO priority with specified service to the Log system.
   * If object is an exception the track is traced.
   */
  public static void traceInfo(int service, Object obj) {
    instance.trace(service, LogPriorities.INFO, obj);
  }

  /**
   * Trace the Object in DEBUG priority with specified service to the Log system.
   * If object is an exception the track is traced.
   */
  public static void traceDebug(int service, Object obj) {
    instance.trace(service, LogPriorities.DEBUG, obj);
  }

  /**
   * Trace the Object in DEBUG priority with specified service to the Log system.
   * If object is an exception the track is traced.
   */
  public static void traceUserSynchro(int service, Object obj) {
    instance.trace(service, LogPriorities.USERSYNCHRO, obj);
  }

	public LogTracer getLogTracer()  {
		return tracer;
	}

	public void setLogTracer(LogTracer t)  {
		tracer = t;
	}

}
