/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */
 
package org.openmobileis.database.fastobjectdb;

import java.io.*;

import org.odbms.Query;
import org.openmobileis.database.fastobjectdb.db.CollectionManager;
import org.openmobileis.database.fastobjectdb.db.FODBCollection;
import org.openmobileis.database.fastobjectdb.db.exception.FODBException;
import org.openmobileis.database.fastobjectdb.db.index.FODBUniqueIndex;
import org.openmobileis.database.fastobjectdb.db.query.soda.FODBSodaQuery;
import org.openmobileis.database.fastobjectdb.db.store.FODBMainFile;
import org.openmobileis.database.fastobjectdb.db.transaction.TransactionManager;

import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.exception.BadDataFormatException;
import org.openmobileis.common.util.exception.DatabaseException;

/**
 * @author Philippe Delrieu
 * 
 * FastObjectDB is the main class of the librairie.
 * All base function are accessed by this class
 * 
 * FastObjectDB is an Object collection based database.
 * Object are stored in collection.
 * Object are retrieved using Index that are attached to collections.
 * A SODA like API is use to request collection.
 * 
 * FODB creation :
 * <CODE>
 * 		db = FastObjectDB.open(dbdir, "testdb");
 *		if (!FODBCollection.isCollectionExist(db, "TESTSEARCH"))	{  // if not created create it.
 *			db.createCollection("TESTSEARCH", TestSearchData.class);
 * 			FODBUniqueIntIndexDescriptor KeyDescriptor = new FODBUniqueIntIndexDescriptor("KEY", "getKey()", 5);
 *			db.addIndex("TESTSEARCH", KeyDescriptor);
 *			FODBMultipleIntIndexDescriptor debKeyDescriptor = new FODBMultipleIntIndexDescriptor("DIZ", "getDizaine()", 5, 5);
 *			db.addIndex("TESTSEARCH", debKeyDescriptor);
 *		}
 *</CODE>
 */
public class FastObjectDB {
	protected final static int ACTION_CREATE = 0;
	protected final static int ACTION_OPEN = 1;
	
	private String name;
	private String rootDir;
	protected CollectionManager colManager;
	
	
	private FODBMainFile databaseFile;
	
	private TransactionManager transactionManager;

	/**
	 * The Constructor.
	 * It is private, it is called by local methods create and open
	 * 
	 * @param dbRootDir - database root directory
	 * @param dbName - database name.
	 * @param fmf - database main file.
	 * @param action - ACTION_CREATE or ACTION_OPEN
	 * @throws BadDataFormatException
	 * @throws FODBException
	 */	
	protected FastObjectDB(String dbRootDir, String dbName, FODBMainFile fmf, int action) throws BadDataFormatException, FODBException {
		name = dbName;
		rootDir = dbRootDir;
		databaseFile = fmf;
		colManager =new CollectionManager();
		transactionManager = new TransactionManager();
		if (action == ACTION_OPEN) {
			String[] colList = databaseFile.getCollectionsList();
			if (colList != null) {
				for (int i = 0; i < colList.length; i++) {
//					System.out.println("" + colList[i]);
					try {
						openCollection(colList[i]);
          } catch (IOException ex) {
            throw new FODBException("Open collection error :"+colList[i], ex);
          } catch (ClassNotFoundException ex) {
            throw new FODBException("Open collection error :"+colList[i], ex);
          }
				}
			}
		}
	}
	
	/**
	 * The function is called to open a database. If the database does not exist, It's created.
	 * 
	 * @param dbRootDir - database directory
	 * @param dbName - database name
	 * @return the database object that has been created.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 * @throws BadDataFormatException
	 * @throws FODBException
	 */
	static public FastObjectDB open(String dbRootDir, String dbName) throws IOException, ClassNotFoundException, BadDataFormatException, FODBException {
		String completeFilePath = "" + dbRootDir + "/" + dbName + "/" + dbName + ".db";
		File file = new File(completeFilePath);
		FastObjectDB db = null;
		if (file.exists())	{
			FODBMainFile fmf = FODBMainFile.openDbFile(completeFilePath);
			db =  new FastObjectDB(dbRootDir, dbName, fmf, ACTION_OPEN);
		} else	{
			File filedir = new File(dbRootDir+ "/" + dbName);
			if (!filedir.exists())	{
				filedir.mkdirs();
			}			
			FODBMainFile fmf = FODBMainFile.createDbFile(completeFilePath);
			db =  new FastObjectDB(dbRootDir, dbName, fmf, ACTION_CREATE);
		}
		return db;
	}
	
	public String getRootDir() {
		return rootDir;
	}
	
	public String getName() {
		return name;
	}
	
	public FODBCollection getCollection(String name)	{
		return  colManager.getCollectionByName(name);
	}

	public boolean isCollectionExist(String colName) {
		String completeFilePath = "" + this.getRootDir() + "/" + this.getName() + "/" + colName + ".col";
		File collect = new File(completeFilePath);
		return collect.exists();

	}
  
	/**
	 * Add a new collection to the database.
	 * 
	 * @param collectionName - name of the new colletion
	 * @param objectType - type of objects stored in this collection.
	 * @param synchro - if set to false  the collection wont be synchronized when FODB is in synchro mode.
	 * @return true if it succeded, false otherwise.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 * @throws BadDataFormatException
	 * @throws FODBException
	 */
	public boolean createCollection(FODBCollectionDescriptor descriptor) throws IOException, ClassNotFoundException, BadDataFormatException, FODBException {
		if (colManager.getCollectionByName(descriptor.getCollectionName())!=null) {
			return false;
		}
		FODBCollection newCol;
		newCol = FODBCollection.createCollection(this, descriptor);
		colManager.addCollection(newCol);
		
		databaseFile.addCollection(descriptor.getCollectionName());
		
		return true;
	}
	
	/**
	 * Add a new collection to the database.
	 * the synchro parameter is set to true.
	 * @param collectionName - name of the new colletion
	 * @param objectType - type of objects stored in this collection.
	 * @return true if it succeded, false otherwise.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 * @throws BadDataFormatException
	 * @throws FODBDataFileException
	 * @throws FODBCollectionException
	 */
	public boolean createCollection(String collectionName, Class objectType) throws IOException, ClassNotFoundException, BadDataFormatException, FODBException {
		return this.createCollection(new FODBCollectionDescriptor(collectionName, objectType));
	}
	
	/**
	 * Open an existing collection.
	 * This is not used and should be suppressed.
	 * 
	 * @param collectionName - name of the collection
	 * @return true if it succeded, false otherwise.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 * @throws BadDataFormatException
	 * @throws FODBException
	 */
	protected boolean openCollection(String collectionName) throws IOException, ClassNotFoundException, BadDataFormatException, FODBException {
		if (colManager.getCollectionByName(collectionName)!=null) {
			return false;
		}
		FODBCollection newCol;
		newCol = FODBCollection.openCollection(this, collectionName);
		colManager.addCollection(newCol);
		return true;
	}
	
	public boolean isCollectionEmpty(String colName) throws DatabaseException	{
		boolean ret = false;
		FODBCollection collect = colManager.getCollectionByName(colName);
		if (collect != null)	{
			try	{
				this.getTransactionManager().begin();
				try	{
					this.getTransactionManager().enterTransaction(collect.getName());
					FODBUniqueIndex index = collect.getCollectionIdIndex();
					ret = index.isEmpty();
				} finally	{
					this.getTransactionManager().commit();
				}
			} catch (Throwable ex)	{
				throw new DatabaseException(ex);
			}			
		}
		return ret;
	}
	
	/**
	 * Add an index to a Collection of the Database.
	 * 
	 * @param colName - name of the Collection
	 * @param descriptor - descriptor representing the index to add.
	 * @return true if it succeded, false otherwise.
	 * @throws BadDataFormatException
	 * @throws FODBException
	 */	
	public boolean addIndex(String colName, FODBIndexDescriptor descriptor) throws BadDataFormatException, FODBException {
		FODBCollection col;
		if ( (col = colManager.getCollectionByName(colName)) == null) {
			return false;
		}
		this.getTransactionManager().enterTransaction(col.getName());
		return col.addIndex(descriptor);
	}
	
	/**
	 * Add an element to a Collection of the database.
	 * 
	 * @param colName - Name of the collection
	 * @param obj - object to add.
	 * @return - true if it succeded, false otherwise.
	 * @throws FODBException
	 */
	public boolean add(String colName, Object obj) throws FODBException {
		FODBCollection col;
		if ( (col = colManager.getCollectionByName(colName)) == null) {
			return false;
		}
		
		this.getTransactionManager().enterTransaction(col.getName());
		return col.addElement(obj);
	}
	
	/**
	 * Replace an existing element in a Collection of the database.
	 * 
	 * @param colName - Name of the Collection
	 * @param obj - Object to replace.
	 * @return true if it succeded, false otherwise.
	 * @throws FODBException
	 */
	public boolean replace(String colName, Object obj) throws FODBException {
		FODBCollection col;
		if ( (col = colManager.getCollectionByName(colName)) == null) {
			return false;
		}
		
		this.getTransactionManager().enterTransaction(col.getName());
		return col.replaceElement(obj);
	}

	/**
	 * use to get base class to query FODB. 
	 * The FODB query API use the SODA API (http://http://odbms.org/soda)
	 * as a base API. 
	 * From Soda specification the differences are :
	 * 	* only query on FODB collection : use query.constrain(MyCollectionObject.class) to declare it.
	 *  * only query using index. use query.descend("IndexedFieldName") to specify an index.
	 * 
	 * Exemple of use:
	 * Data class :TestSearchData 
	 * class signature
	 * public final class TestSearchData implements Serializable {
	 *   public int getKey(); 
	 *   public int getDizaine();
	 * }
	 * 
	 * FODB creation :
	 * <CODE>
	 * 		db = FastObjectDB.open(dbdir, "testdb");
	 *			if (!FODBCollection.isCollectionExist(db, "TESTSEARCH"))	{  // if not created create it.
	 *			db.createCollection("TESTSEARCH", TestSearchData.class);
	 * 			FODBUniqueIntIndexDescriptor KeyDescriptor = new FODBUniqueIntIndexDescriptor("KEY", "getKey()", 5);
	 *			db.addIndex("TESTSEARCH", KeyDescriptor);
	 *			FODBMultipleIntIndexDescriptor debKeyDescriptor = new FODBMultipleIntIndexDescriptor("DIZ", "getDizaine()", 5, 5);
	 *			db.addIndex("TESTSEARCH", debKeyDescriptor);
	 *		}
	 * </CODE>
	 * The collection TestData where keys are in the interval [0..30].
	 * <CODE>
	 * 		for (int i=0; i<NB_SEARCH_CREATED_DATA; i++)	{
	 *			db.add("TESTSEARCH", new TestSearchData(i));				
	 *		}
	 * </CODE>
	 *
	 * Exemple of query :
	 * 			
	 * <CODE>
	 *		//query SELECT * FROM TESTSEARCH
	 *		Query q = db.query();
	 *		q.constrain(TestSearchData.class);
	 * 		Query subq = q.descend("getKey()"); //specify an index to request. 
	 *		ObjectSet set = q.execute();
	 *		
	 *		//query SELECT * FROM TESTSEARCH where getKey<14
	 *		q = db.query();
	 *		q.constrain(TestSearchData.class);
	 * 		subq = q.descend("getKey()");
	 *		subq.constrain(new Integer(14)).smaller();
	 *		set  = q.execute();
	 *		
	 *		//query SELECT * FROM TESTSEARCH where getKey>16 AND getKey<23
	 *		q = db.query();
	 *		q.constrain(TestSearchData.class);
	 *		subq = q.descend("getKey()");
	 *		Constraint c = subq.constrain(new Integer(16)).greater();
	 *		Constraint d = subq.constrain(new Integer(23)).smaller();
	 *		c.and(d);
	 *		set  = q.execute();
	 *		
	 *		//query SELECT * FROM TESTSEARCH where getKey>16 AND getDizaine=2
	 *		q = db.query();
	 *		q.constrain(TestSearchData.class);
	 *		subq = q.descend("getKey()");
	 *		c = subq.constrain(new Integer(16)).greater();
	 *		Query subq2 = subq.descend("getDizaine()");
	 *		d = subq2.constrain(new Integer(2)).equal();
	 *		c.and(d);
	 *		set = q.execute();
	 * </CODE>
	 *
	 * @see org.odbms.Constraint
	 * @see org.odbms.ObjectSet
	 * @see org.odbms.Query
	 * @see org.openmobileis.database.fastobjectdb.db.query.soda.FODBSodaQuery
	 * 
	 * @author	Philippe Delrieu
	 * @since	JDK 1.1
	 * @version 1.0.
	 */
	public Query query() {
		return new FODBSodaQuery(colManager, this);
	}
	
	/**
	 * Delete an element of a Collection
	 * 
	 * @param colName - name of the Collection.
	 * @param obj - object to delete.
	 * @return true if it succeded, false otherwise.
	 * @throws FODBException
	 */
	public boolean delete(String colName, Object obj) throws FODBException {
		if (obj == null) throw new  FODBException("Error FastObjectDB delete object null");
		FODBCollection col;
		if ( (col = (FODBCollection) colManager.getCollectionByName(colName)) == null) {
			return false;
		}
		
		this.getTransactionManager().enterTransaction(col.getName());
		return col.deleteSingleElement(obj);
	}
	
	/**
	 * Delete an object of a Collection with its id
	 * 
	 * @param colName - name of the Collection.
	 * @param id - id of the object to delete.
	 * @return true if it succeded, false otherwise.
	 * @throws FODBException
	 */
	public boolean deleteWithId(String colName, Object id) throws FODBException {
		if (id == null) throw new  FODBException("Error FastObjectDB delete object null");
		FODBCollection col;
		if ( (col = (FODBCollection) colManager.getCollectionByName(colName)) == null) {
			return false;
		}
		
		this.getTransactionManager().enterTransaction(col.getName());
		return col.deleteObjectWithId(id);
	}
	
	public String toString() {
		Object[] collectionsKey = null;
		String result = "FastObjectDB [" + name + ", " + rootDir + "]\n";
		
		collectionsKey = colManager.getCollectionList().keySet().toArray();
		for (int i = 0; i < collectionsKey.length; i++) {
			result = result.concat("  " + (String)collectionsKey[i] + "\n" + colManager.getCollectionByName((String)collectionsKey[i]) + "\n");
		}
		System.out.println("");
		
		return result;
	}
	
	public FODBCollection[] getDatabaseCollectionArray()	{
		Array arr =  colManager.getCollectionArray();
		FODBCollection[] retList = new FODBCollection[arr.size()];
		arr.toArray(retList);
		return retList;
	}

	public TransactionManager getTransactionManager() {
		return transactionManager;
	}

}
