/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 * 
 */

package org.openmobileis.database.fastobjectdb.db;

import java.io.IOException;


import org.openmobileis.database.fastobjectdb.db.exception.FODBException;
import org.openmobileis.database.fastobjectdb.db.index.FODBIndex;
import org.openmobileis.database.fastobjectdb.db.index.FODBUniqueIndex;
import org.openmobileis.database.fastobjectdb.db.index.SearchResult;
import org.openmobileis.database.fastobjectdb.db.index.node.Node;

/**
 * Use to enumerate an index without to loading in memory all index key or referenced object.
 * Only the current object is loaded.
 * Start at the end and enumerate backward
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class BackwardUniqueIndexIterator {
	private FODBCollection collection;
	private Node currentNode;
	private int position = 0;
	private boolean begin = false;
	private FODBIndex index;

  /**
   * 
   */
  BackwardUniqueIndexIterator(FODBIndex index, FODBCollection collection) throws IOException, ClassNotFoundException, FODBException {
    super();
    if (!(index instanceof FODBUniqueIndex))	{
    	throw new FODBException("Not unique index to iterator awailable");
    }
    this.collection = collection;
    this.index = index;
		collection.getDatabase().getTransactionManager().begin();
    try	{
			collection.getDatabase().getTransactionManager().enterTransaction(this.collection.getName());
    	this.currentNode = index.readRoot();
    	if (this.currentNode.nbKey == 0)	{ //empty index
    		begin = true;
    	}
    	//position at the last element
			this.position = this.currentNode.nbKey-1;
			long childBranch = currentNode.branchs[this.position+1];
			while (childBranch != Node.NO_NODE)	{
				this.currentNode = (Node)collection.getNodeAtPos(childBranch);
				this.position = this.currentNode.nbKey-1; 
				childBranch = currentNode.branchs[this.position+1];
			}
    } finally	{
			collection.getDatabase().getTransactionManager().commit();
    }
  }  
  
	public boolean hasPrev() throws FODBException	{
		return !begin;
	}
  
	public Object prev() throws FODBException	{
		if (!this.hasPrev()) return null;
		collection.getDatabase().getTransactionManager().begin();
		try	{
			collection.getDatabase().getTransactionManager().enterTransaction(this.collection.getName());
			
			//read current obj
			long objPos= currentNode.ptr[this.position];
			Object ret = collection.getElementAtPos(objPos);
			
			//get next pos
			if (currentNode.branchs[(this.position)] != Node.NO_NODE)	{ //goes down
				long childBranch = currentNode.branchs[this.position];
				while (childBranch != Node.NO_NODE)	{
					this.currentNode = (Node)collection.getNodeAtPos(childBranch);
					this.position = currentNode.nbKey-1; 
					childBranch = currentNode.branchs[this.position+1];
				}
			} else	{
				this.position --;  //goes next
				while (this.position < 0)	{ //goes up
					long parent = currentNode.parentPtr;
					if (parent == Node.NO_NODE)	{ //root node end index traversal.
						begin = true; 
						break;
					}
					long nodepos = currentNode.filePtr;
					this.currentNode = (Node)collection.getNodeAtPos(parent);
					//child node position
					int i=0;
					for (; i<this.currentNode.branchs.length; i++)	{
						if (nodepos == this.currentNode.branchs[i]) break;
					}
					this.position = i-1;
				}
			}
			
			return ret;
		} finally	{
			collection.getDatabase().getTransactionManager().commit();
		}  	
	}
  
	public void initCurrentNode(Object indexKey) throws FODBException	{
		collection.getDatabase().getTransactionManager().begin();
		try	{
			collection.getDatabase().getTransactionManager().enterTransaction(this.collection.getName());
			SearchResult result = ((FODBUniqueIndex)index).getNodeForKey(indexKey);
			if (result.found)	{
				this.currentNode = result.node;
				this.position = result.pos;
				begin = false;
			}
			this.prev(); //position to next record.
		} finally	{
			collection.getDatabase().getTransactionManager().commit();
		}
  	
	}
}
