/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */
package org.openmobileis.database.fastobjectdb.db.index;

import org.openmobileis.common.util.collection.Array;
import org.openmobileis.database.fastobjectdb.FODBIndexDescriptor;
import org.openmobileis.database.fastobjectdb.db.exception.FODBException;
import org.openmobileis.database.fastobjectdb.db.exception.FODBQueryException;
import org.openmobileis.database.fastobjectdb.db.index.node.Node;
import org.openmobileis.database.fastobjectdb.db.query.soda.SodaIndexComparator;
import org.openmobileis.database.fastobjectdb.db.store.FODBCollectionIndexFile;

import java.io.IOException;
import java.lang.reflect.*;


/**
 *
 * FODBIndex is an abstract Class on which each type of index
 * will be based.
 * 
 */

public abstract class FODBIndex	{
  public static final int UNIQUE = 1;
  public static final int MULTIPLE = 2;

  protected String name;
  protected final AccessibleObject accessObj;
  protected FODBCollectionIndexFile colFile;
  protected FODBIndexHeader header;
	protected int minKey;

  /**
   * Constructor of FODBIndex, using an index header.
   * Using an index header, means that we're opening (and not
   * creating) an index.
   */
  public FODBIndex(FODBIndexHeader newHeader, FODBCollectionIndexFile cFile, AccessibleObject accObj) {
    header = newHeader;
	minKey = (int) Math.ceil((double) header.descriptor.getOrder() / 2D) - 1;
    name = header.getName();
    colFile = cFile;
    accessObj = accObj;
  }

  public Node readRoot() throws IOException, ClassNotFoundException {
    return (Node) colFile.readNode(header.rootPtr);
  }

  /**
   * Constructor of FODBIndex, using an index descriptor.
   * Using an index descriptor, means that we're creating (and not
   * opening) an index.
   * This is the reason why this constructor is little bit more
   * complicated that the upper one.
   */
  public FODBIndex(FODBIndexDescriptor descriptor, FODBCollectionIndexFile cFile, AccessibleObject accObj) throws FODBException {
    name = descriptor.getName();
    colFile = cFile;
    accessObj = accObj;
    try {

			// descriptor verification (specific to each index type).
			specificDescriptorVerifications(descriptor);
      // open collection file
        // Logically, we should rewrite the root page as we initialised
        // its filePtr field.
        // This is not needed as FODBObjectDatabaseFile already
        // does it. Each time it writes an object instance of Page, it
        // first writes it, fill its filePtr field with the right value, and then
        // rewrites it.
        // We actually even shouldn't fill the filePtr field ourselves as
        // it is already done.

        // new header
        if (accessObj instanceof Method) {
          descriptor.setMemberName((((Method) accessObj).getName()).concat("()"));
          header = new FODBIndexHeader(descriptor);
          // we concat '()' at the end of Methods name, in order to be
          // able to recognise it is a method, next time we open
          // the index.
        } else { // accessObj instanceof Field
          descriptor.setMemberName(((Field) accessObj).getName());
          header = new FODBIndexHeader(descriptor);
        }

        header.descriptor.setType(this.getType()); // specific to each index type.
        header.rootPtr = Node.NO_NODE;

        // write header to file
        // First to get it's position
        long pos = colFile.writeHeader(header);
        // Then to store it with it's position
        header.pos = pos;
        colFile.rewriteHeader(header, pos);
        // Then we store it's pointer to the IndexTable
        colFile.addIndexPointerToTable(pos);

        // new root page (specific to each index type).
        Node root = initRoot(descriptor);
        // write root to collection file
        //update header rootPtr.
        root.filePtr = colFile.writeNode(root, header);

        //close collection file
    } catch (Throwable ex) {
      throw new FODBException(ex);
    }
		minKey = (int) Math.ceil((double) header.descriptor.getOrder() / 2D) - 1;
  }

  public String getName() {
    return name;
  }

  public long getPos() {
    return header.pos;
  }

  /**
   * return true if the index is empty.
   * throws StoreBTreeException if the index is closed
   */
  public boolean isEmpty() throws FODBException {
    try {
      Node root = colFile.readNode(header.rootPtr);
      if (root == null) {
        throw new FODBException("Index closed");
      }

      return (root.nbKey == 0);
    } catch (Throwable ex) {
      throw new FODBException(ex);
    }
  }

  /**
   * Add a new pointer in our index.
   * 
   * @param obj - helps us getting the key.
   * @param pos - position of the element in the collection file.
   * @throws FODBDataFileException
   * @throws FODBIndexException
   * @throws BTreeException
   */
  public void add(Object obj, long pos) throws FODBException {
  	try	{
     	Class accessObjType=getAccessObjType();
    	if(accessObjType.getName().startsWith("[")){
        	//insert keys for all elements in array
    		Array keys=getArrayKey(obj);
    		for(int i=0;i<keys.size();i++){
    			Object key=keys.get(i);
    			if(key!=null){
    				insertKey(key,pos);
    			}
    		}
    	}else{
    		//insert key for obj
    		Object key=getKey(obj);
    		insertKey(key,pos);
    	}
  	} catch (Throwable ex)	{
  		throw new FODBException(ex);
  	}
  }
  
  public Class getAccessObjType(){
  	if(accessObj instanceof Method)
  		return ((Method)accessObj).getReturnType();
  	else
  		return ((Field)accessObj).getType();
  }
  

  /**
   * Remove a pointer from our index.
   * 
   * @param obj - helps us getting the key.
   * @param pos - position of the element in the collection file.
   * @return true if the operation succeed, false otherwise.
   * @throws FODBIndexException
   */
  public boolean delObjectWithKey(Object obj, long pos) throws FODBException {
  	try	{
  		boolean result=true;
  		Class accessObjType=getAccessObjType();
    	if(accessObjType.getName().startsWith("[")){
        	//insert keys for all elements in array
    		Array keys=getArrayKey(obj);
    		for(int i=0;i<keys.size();i++){
    			Object key=keys.get(i);
    			if(key!=null){
    				result=result && deleteKey(key,pos);
    			}
    		}
    	}else{
    		//insert key for obj
    		Object key=getKey(obj);
    		result=deleteKey(key,pos);
    	}
    	return result;
  	} catch (Throwable ex)	{
  		throw new FODBException(ex);
  	}
  }

  public FODBIndexHeader getIndexHeader() {
    return this.header;
  }

	public abstract long[] query(SodaIndexComparator comparator) throws FODBQueryException;
  protected abstract void specificDescriptorVerifications(FODBIndexDescriptor descriptor) throws FODBException;
  protected abstract Node initRoot(FODBIndexDescriptor descriptor) throws FODBException;
  public abstract int getType();

  protected abstract void insertKey(Object key, long pos) throws IOException, ClassNotFoundException, FODBException;
  protected abstract boolean deleteKey(Object obj, long pos) throws IOException, ClassNotFoundException, FODBException;
  
	public abstract Object getKey(Object obj) throws FODBException;
	public abstract org.openmobileis.common.util.collection.Array getArrayKey(Object obj) throws FODBException;
    
  public FODBIndexDescriptor getIndexDescriptor()   {
    return this.header.descriptor;
  }

  /**
   * Helps us debugging.
   */
  public String toString() {
    String result = "FODBIndex[" + name + ", ";

    if (accessObj instanceof Method) {
      result = result.concat((((Method) accessObj).getName()).concat("()"));
    } else { // accessObj instanceof Field
      result = result.concat(((Field) accessObj).getName());
    }

    result =
      result.concat(
        ", "
          + header.getName()
          + ", "
          + header.getMemberName()
          + ", "
          + header.descriptor.getType()
          + ", "
          + header.descriptor.getOrder()
          + ", "
          + header.pos
          + ", "
          + header.rootPtr);

    result = result.concat("]");

    return result;
  }

}
