/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */
package org.openmobileis.database.fastobjectdb.db.index;

import java.lang.reflect.*;

import org.openmobileis.common.util.exception.BadDataFormatException;
import org.openmobileis.database.fastobjectdb.FODBIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FODBIntIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FODBLongIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FODBStringIndexDescriptor;
import org.openmobileis.database.fastobjectdb.db.exception.FODBException;
import org.openmobileis.database.fastobjectdb.db.store.FODBCollectionIndexFile;
import org.openmobileis.database.fastobjectdb.db.util.FODBUtils;

/**
 * Factory use to create index 
 *
 */
public final class FODBIndexFactory {
	public static final int UNIQUE = 1;
	public static final int MULTIPLE = 2;
	
	private static FODBIndexFactory factory;
	
	public FODBIndexFactory() {
	}
	
	public static FODBIndexFactory getFactory() {
		if (factory == null)  {
			synchronized(FODBIndexFactory.class) {
				if (factory == null)  {
					factory = new FODBIndexFactory();
				}
			}
		}
		return factory;
	}

	protected boolean isValidKeyType(Class keyType,Class indexType){
		boolean valid=false;
		if(keyType==indexType)
			valid=true;
		else if(keyType.getComponentType()==indexType)
			valid=true;
		return valid;
	}
	public FODBIndex createIndex(FODBIndexDescriptor descriptor, Class objectType, FODBCollectionIndexFile cFile) throws BadDataFormatException, FODBException {
		Class keyType;
		AccessibleObject accessObj = FODBUtils.getMember(objectType, descriptor.getMemberName());

		if (accessObj == null) {
			throw new BadDataFormatException("Method or Field does not exist");
		}
		
		if (accessObj instanceof Method) {
			keyType = ((Method)accessObj).getReturnType();
		}
		else { // accessObj instanceof Field
			keyType = ((Field)accessObj).getType();
		}
		
		if (descriptor instanceof FODBStringIndexDescriptor) {
			if (!isValidKeyType(keyType,String.class)) {
				throw new BadDataFormatException("Invalid Method or Field for type String");
			}
			FODBIndex newIndex;
			if (descriptor.getType() == UNIQUE) {
				newIndex = new FODBUniqueStringIndex((FODBStringIndexDescriptor)descriptor, cFile, accessObj);
				return newIndex;
			} else {
				newIndex = new FODBMultipleStringIndex((FODBStringIndexDescriptor)descriptor, cFile, accessObj);
				return newIndex;
			}
		}
		else if (descriptor instanceof FODBIntIndexDescriptor) {
			if (!isValidKeyType(keyType,int.class)) {
				throw new BadDataFormatException("Invalid Method or Field for type int");
			}
			FODBIndex newIndex;
			if (descriptor.getType() == UNIQUE) {
				newIndex = new FODBUniqueIntIndex((FODBIntIndexDescriptor)descriptor, cFile, accessObj);
				return newIndex;
			} else {
				newIndex = new FODBMultipleIntIndex((FODBIntIndexDescriptor)descriptor, cFile, accessObj);
				return newIndex;
			}
		}
		else if (descriptor instanceof FODBLongIndexDescriptor) {
			if (!isValidKeyType(keyType,long.class)) {
				throw new BadDataFormatException("Invalid Method or Field for type long");
			}
			FODBIndex newIndex;
			if (descriptor.getType() == UNIQUE) {
				newIndex = new FODBUniqueLongIndex((FODBLongIndexDescriptor)descriptor, cFile, accessObj);
				return newIndex;
			} else {
				newIndex = new FODBMultipleLongIndex((FODBLongIndexDescriptor)descriptor, cFile, accessObj);
				return newIndex;
			}
		} else {
			throw new BadDataFormatException("Unrecognised IndexDescriptor");
		}

	}
	
	public FODBIndex createIndex(FODBIndexHeader header, Class objectType, FODBCollectionIndexFile cFile) throws BadDataFormatException, FODBException {
		Class keyType;
		AccessibleObject accessObj = FODBUtils.getMember(objectType, header.getMemberName());
		
		if (accessObj == null) {
			throw new BadDataFormatException("Method or Field does not exist");
		}
		
		if (accessObj instanceof Method) {
			keyType = ((Method)accessObj).getReturnType();
		}
		else { // accessObj instanceof Field
			keyType = ((Field)accessObj).getType();
		}

		//if (keyType == String.class) {
		if (isValidKeyType(keyType,String.class)) {
			FODBIndex newIndex;
			if (header.descriptor.getType() == UNIQUE) {
				newIndex = new FODBUniqueStringIndex(header, cFile, accessObj);
				return newIndex;
			} else {
				newIndex = new FODBMultipleStringIndex(header, cFile, accessObj);
				return newIndex;
			}
		}
		else if (isValidKeyType(keyType,int.class)) {
			FODBIndex newIndex;
			if (header.descriptor.getType() == UNIQUE) {
				newIndex = new FODBUniqueIntIndex(header, cFile, accessObj);
				return newIndex;
			} else {
				newIndex = new FODBMultipleIntIndex(header, cFile, accessObj);
				return newIndex;
			}
		}
		else if (isValidKeyType(keyType,long.class)) {
			FODBIndex newIndex;
			if (header.descriptor.getType() == UNIQUE) {
				newIndex = new FODBUniqueLongIndex(header, cFile, accessObj);
				return newIndex;
			} else {
				newIndex = new FODBMultipleLongIndex(header, cFile, accessObj);
				return newIndex;
			}
		} else {
			throw new BadDataFormatException("Unrecognised key type for creation of index");
		}

	}
}
