/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */

package org.openmobileis.database.fastobjectdb.db.index.node;

import java.io.*;

import org.openmobileis.database.fastobjectdb.db.index.SearchResult;

/**
 * StringNode.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public class StringNode extends Node {
	// serialization constant
	static final long serialVersionUID = 5521257935120563452L;

	public static String SHARE_STRING = new String();

	//-------------------------
	// fields
	//-------------------------
	private String[] keys; // array of key values

	private int keyLength;

	//-------------------------
	// constructors
	//-------------------------
	public StringNode() {
		super();
	}

	public StringNode(int order, int keyLength) {
		super(order);
		// allocate arrays
		keys = new String[order - 1];

		for (int n = 0; n < keys.length; ++n)
			keys[n] = StringNode.SHARE_STRING;

		this.keyLength = keyLength;

	}

	public String getKeyAtPos(int pos) {
		return this.keys[pos];
	}

	public void setKeyPtrAtPos(String key, long ptr, int pos) {
		this.keys[pos] = key;
		super.ptr[pos] = ptr;
	}

	public SearchResult searchNode(String key, SearchResult search) {
		search.node = this;
		if (key.compareTo(this.getKeyAtPos(0)) < 0) {
			search.pos = 0;
			search.found = false;
		} else {
			int i;
			for (i = nbKey - 1; i >= 0
					&& key.compareTo(this.getKeyAtPos(i)) < 0; i--)
				;
			search.found = key.equals(this.getKeyAtPos(i));
			if (search.found)
				search.pos = i;
			else
				search.pos = i + 1;
		}
		return search;
	}

	public boolean pushInLeaf(String key, int pos, long newPtr) {
		this.branchs[this.nbKey + 1] = Node.NO_NODE;
		for (int j = this.nbKey - 1; j >= pos; j--) {
			this.setKeyPtrAtPos(this.getKeyAtPos(j), this.getNodePtrAtPos(j),
					j + 1);
		}
		this.setKeyPtrAtPos(key, newPtr, pos);
		this.nbKey++;
		return this.nbKey >= this.getMaxNbKey();
	}

	public boolean promote(String newkey, long keyptr, int pos, long nodePtr) {
		for (int j = this.nbKey; j >= pos + 1; j--) {
			this.setKeyPtrAtPos(keys[j - 1], this.getNodePtrAtPos(j - 1), j);
			this.branchs[j + 1] = this.branchs[j];
		}
		this.branchs[pos + 1] = nodePtr;
		this.setKeyPtrAtPos(newkey, keyptr, pos);
		this.nbKey++;
		return this.nbKey >= this.getMaxNbKey();
	}

	public void removeKeyAtPpos(int pos) {
		for (int j = pos + 1; j < this.nbKey; j++) {
			this.setKeyPtrAtPos(this.getKeyAtPos(j), this.getNodePtrAtPos(j),
					j - 1);
			this.branchs[j] = this.branchs[j + 1];
		}
		this.branchs[this.nbKey] = Node.NO_NODE;

		this.nbKey--;
	}

	//---------------------
	// serialization
	//---------------------
	public void writeExternal(ObjectOutput out) throws IOException {
		super.writeExternal(out);
		// create raw key data
		out.writeInt(keyLength);

		for (int n = 0; n < keys.length; ++n) {
			// copy characters from string to raw data
			int i = 0;

			while ((i < keys[n].length()) && (i < keyLength)) {
				out.writeChar(keys[n].charAt(i));
				++i;
			}

			// fill any extra space with nulls
			while (i < keyLength) {
				out.writeChar('\0');
				++i;
			}
		}

		// write raw keys
		//   out.writeObject(rawKeys);
	}

	public void readExternal(ObjectInput in) throws IOException,
			ClassNotFoundException {
		super.readExternal(in);
		keyLength = in.readInt();

		// get and check raw key length
		int maxKey = super.getMaxNbKey();

		// allocate key arrays
		keys = new String[maxKey];

		// extract strings from their raw format
		int n = 0;
		//  int pos = 0;

		StringBuffer newKey = new StringBuffer(keyLength);
		for (n = 0; n < maxKey; ++n) {
			// create a buffer

			int i;
			char currentread;
			// copy characters to buffer
			for (i = 0; i < keyLength; ++i) {
				currentread = in.readChar();
				if (currentread == '\0')
					break;
				newKey.append(currentread);
			}
			//skip remainding char
			for (i++; i < keyLength; i++) {
				in.readChar();
			}

			// create string from buffer
			keys[n] = newKey.toString();
			newKey.setLength(0);
		}
	}

	/**
	 * @return
	 */
	public int getKeyLength() {
		return keyLength;
	}

	/**
	 * @param i
	 */
	public void setKeyLength(int i) {
		keyLength = i;
	}

}
