/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */
package org.openmobileis.database.fastobjectdb.db.query.soda;

import org.odbms.Constraint;
import org.odbms.Constraints;
import org.odbms.ObjectSet;
import org.odbms.Query;
import org.openmobileis.common.util.collection.LongArray;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.database.fastobjectdb.FastObjectDB;
import org.openmobileis.database.fastobjectdb.db.CollectionManager;
import org.openmobileis.database.fastobjectdb.db.FODBCollection;
import org.openmobileis.database.fastobjectdb.db.index.FODBIndex;

/**
 * Base class to query FODB. Use the SODA API (http://http://odbms.org/soda)
 * as a base API. 
 * From Soda specification the differences are :
 * 	* only query on FODB collection : use query.constrain(MyCollectionObject.class) to declare it.
 *  * only query using index. use query.descend("IndexedFieldName") to specify an index.
 * 
 * Exemple of use:
 * Data class :TestSearchData 
 * class signature
 * public final class TestSearchData implements Serializable {
 *   public int getKey(); 
 *   public int getDizaine();
 * }
 * 
 * FODB creation :
 *   <CODE>
 * 		db = FastObjectDB.open(dbdir, "testdb");
 *		if (!FODBCollection.isCollectionExist(db, "TESTSEARCH"))	{  // if not created create it.
 *			db.createCollection("TESTSEARCH", TestSearchData.class);
 * 			FODBUniqueIntIndexDescriptor KeyDescriptor = new FODBUniqueIntIndexDescriptor("KEY", "getKey()", 5);
 *			db.addIndex("TESTSEARCH", KeyDescriptor);
 *			FODBMultipleIntIndexDescriptor debKeyDescriptor = new FODBMultipleIntIndexDescriptor("DIZ", "getDizaine()", 5, 5);
 *			db.addIndex("TESTSEARCH", debKeyDescriptor);
 *		}
 *  </CODE>
 * The collection TestData where keys are in the interval [0..30].
 * <CODE>
 * 		for (int i=0; i<NB_SEARCH_CREATED_DATA; i++)	{
 *			db.add("TESTSEARCH", new TestSearchData(i));				
 *		}
 * </CODE>
 *
 * Exemple of query :
 * 		<CODE>	
 *		//query SELECT * FROM TESTSEARCH
 *		Query q = db.query();
 *		q.constrain(TestSearchData.class);
 * 		Query subq = q.descend("getKey()"); //specify an index to request. 
 *		ObjectSet set = q.execute();
 *		
 *		//query SELECT * FROM TESTSEARCH where getKey<14
 *		q = db.query();
 *		q.constrain(TestSearchData.class);
 * 		subq = q.descend("getKey()");
 *		subq.constrain(new Integer(14)).smaller();
 *		set  = q.execute();
 *		
 *		//query SELECT * FROM TESTSEARCH where getKey>16 AND getKey<23
 *		q = db.query();
 *		q.constrain(TestSearchData.class);
 *		subq = q.descend("getKey()");
 *		Constraint c = subq.constrain(new Integer(16)).greater();
 *		Constraint d = subq.constrain(new Integer(23)).smaller();
 *		c.and(d);
 *		set  = q.execute();
 *		
 *		//query SELECT * FROM TESTSEARCH where getKey>16 AND getDizaine=2
 *		q = db.query();
 *		q.constrain(TestSearchData.class);
 *		subq = q.descend("getKey()");
 *		c = subq.constrain(new Integer(16)).greater();
 *		Query subq2 = subq.descend("getDizaine()");
 *		d = subq2.constrain(new Integer(2)).equal();
 *		c.and(d);
 *		set = q.execute();
 *		</CODE>
 * @see org.odbms.Constraint
 * @see org.odbms.ObjectSet
 * @see org.odbms.Query
 * 
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class FODBSodaQuery implements Query {
	private CollectionManager colManager;
	private boolean queryNoError = true;
	private FODBSodaConstraint firstConstraint;
	private int constraintId = 0;
	private FODBCollection collection; 
	private FastObjectDB db;
	
  /**
   * 
   */
  public FODBSodaQuery(CollectionManager colManager, FastObjectDB db) {
    super();
    this.colManager = colManager;
    this.db = db;
  }

  /* (non-Javadoc)
   * @see org.odbms.Query#constrain(java.lang.Object)
   */
  public Constraint constrain(Object constraint) {
  	if (constraint instanceof Class)	{
			FODBCollection col = colManager.getCollectionByType((Class) constraint);
			if (col != null)	{
				FODBSodaConstraint constr = new FODBSodaConstraint(this, constraintId++);
				firstConstraint = constr;
				this.setCollection(col);
			}
		} else 	{
			if (constraint instanceof String)	{
				FODBCollection col = colManager.getCollectionByName((String) constraint);
				if ((col != null) && (this.collection == null))	{
					this.setCollection(col);
					if (firstConstraint == null)	{
						FODBSodaConstraint constr = new FODBSodaConstraint(this, constraintId++);
						firstConstraint = constr;
					}
					return firstConstraint;
				}
			}
			FODBSodaConstraint path = (FODBSodaConstraint)firstConstraint;
			if (path.hasOperande)	{ //if already declared create a new constraint. One constraint per operation.
				path = new FODBSodaConstraint(this, constraintId++);
				FODBIndex index = collection.getIndexByMember(firstConstraint.getIndexMember());
				if (index != null)	{
					FODBSodaIndexPath ipath = new FODBSodaIndexPath(index);
					path.setIndex(ipath);
				}
			}
			this.setError(path.setOperand(constraint));
			return path;
  	}
  	return firstConstraint;
  }

  /* (non-Javadoc)
   * @see org.odbms.Query#constraints()
   */
  public Constraints constraints() {
		return null;
  }

  /* (non-Javadoc)
   * @see org.odbms.Query#execute()
   */
  public ObjectSet execute() {
		FODBSodaObjectSet set = new FODBSodaObjectSet();
  	if ((queryNoError) && (firstConstraint != null))	{
  		try	{
				db.getTransactionManager().begin();
	  		try	{
          db.getTransactionManager().enterTransaction(this.collection.getName());
	  			LongArray arrayLong = new LongArray();
	    		firstConstraint.execute(arrayLong);
	    		set.addCollectionPtrs(arrayLong.toArray(), this.collection);
	  		} finally	{
					db.getTransactionManager().commit();
	  		}
  		} catch (Throwable ex)	{
  			LogManager.traceError(0, ex);
  		}
  	}
    return set;
  }

  /* (non-Javadoc)
   * @see org.odbms.Query#descend(java.lang.String)
   */
  public Query descend(String fieldName) {
  	if (firstConstraint != null)	{
			FODBSodaQuery retQuery = this;
			FODBSodaConstraint constr = firstConstraint;
  		if (firstConstraint.hasIndex())	{
				retQuery = new FODBSodaQuery(colManager, db);
				retQuery.setCollection(this.collection);
				constr = new FODBSodaConstraint(this, constraintId++);
				retQuery.setFirstConstraint(constr);
//				this.firstConstraint.setNextConstraint(constr, BaseConstraint.CONSTRAINT_AND);
  		}
			FODBIndex index = collection.getIndexByMember(fieldName);
			if (index != null)	{
				FODBSodaIndexPath ipath = new FODBSodaIndexPath(index);
				retQuery.getFirstConstraint().setIndex(ipath);
				return retQuery;
			} else	{
				this.setError(false);
			}
  	} else	{
  		this.setError(false);
  	}
		return null;
  }
  
  private void setError(boolean error)	{
  	if (!error)	{
  		queryNoError = error;
  	}
  	
  }

  /* (non-Javadoc)
   * @see org.odbms.Query#orderAscending()
   */
  public Query orderAscending() {
    // TODO Auto-generated method stub
    return null;
  }

  /* (non-Javadoc)
   * @see org.odbms.Query#orderDescending()
   */
  public Query orderDescending() {
    // TODO Auto-generated method stub
    return null;
  }

	/* (non-Javadoc)
	 * @see org.odbms.Constraint#or(org.odbms.Constraint)
	 */
	Constraint or(Constraint with) {
		// TODO Auto-generated method stub
		return null;
	}

  /**
   * @param collection
   */
  void setCollection(FODBCollection collection) {
    this.collection = collection;
  }

  /**
   * @param constraint
   */
  void setFirstConstraint(FODBSodaConstraint constraint) {
    firstConstraint = constraint;
  }

  FODBSodaConstraint getFirstConstraint() {
    return firstConstraint;
  }

}
