/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */
package org.openmobileis.database.fastobjectdb.db.store;

import java.io.*;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.openmobileis.database.fastobjectdb.FODBCollectionDescriptor;
import org.openmobileis.database.fastobjectdb.FastObjectDB;
import org.openmobileis.database.fastobjectdb.db.crypto.FODBCypher;
import org.openmobileis.database.fastobjectdb.db.exception.FODBException;
import org.openmobileis.database.fastobjectdb.db.index.FODBIndexHeader;
import org.openmobileis.database.fastobjectdb.db.index.FODBIndexTable;
import org.openmobileis.database.fastobjectdb.db.index.node.Node;


/**
 *
 * FODBCollectionFile is a Class that defines simple functions
 * access our data file.
 * 
 */
public class FODBCollectionFile {
	private final static int ACTION_CREATE = 0;
	private final static int ACTION_OPEN = 1;
	
	private FODBObjectDatabaseFile databaseFile = null;
	private FODBCollectionFileHeader header = null;
	private FODBIndexTable indexTable = null;
	private String collectionFilePath;
	private int openCounter = 0;
	private FODBCypher collCypher = null;
	
	/**
	 * The Constructor.
	 * It is private, it is called by local methods createCollection and
	 * openCollection.
	 * 
	 * @param filePath - path to the file (to create or open).
	 * @param objType - the type of Objects stored in the collection.
	 * @param action - create or open.
	 * @throws FODBDataFileException
	 */
	private FODBCollectionFile (String filePath, FODBCollectionDescriptor descriptor, int action, FastObjectDB db) throws IOException, ClassNotFoundException {
		if (action == ACTION_CREATE) {
			File f1 = new File(filePath);
			f1.delete();
		}
		
		collectionFilePath = filePath;
		
		// Let's open our file (or create it if it doesn't exist)
		databaseFile = new FODBObjectDatabaseFile(collectionFilePath);
		db.getTransactionManager().registerTransactionFile(databaseFile, descriptor.getCollectionName());
		databaseFile.open();
		try	{
			if (action == ACTION_CREATE) {
				// create and write header
				header = new FODBCollectionFileHeader(descriptor);
				header.setIndexTablePos(Node.NO_NODE);
	
				writeHeader();
							
				// create and write indexTable
				indexTable = new FODBIndexTable();
				long tablePos = databaseFile.writeObject(indexTable);
	
				
				// set indexTable's pos in the header
				header.setIndexTablePos(tablePos);
				
				// rewrite header
				rewriteHeader();
			}
			else if (action == ACTION_OPEN) {
				header = readHeader();
				
				indexTable = readIndexTable();
			}
		} finally	{
	
			// We read or wrote our datas, let's close the file.		
			databaseFile.close();
		}

	}
	
	public boolean isSynchronized()	{
		return header.getDescriptor().isSynchronize();
	}


	/**
	 * return the collection cypher if any, null otherwise.
	 * 
	 * @return cypher use to encrypt data, null otherwise.
	 */
	public FODBCypher getCollectionCypher() throws FODBException	{
		return collCypher;
	}
  
  /**
   * if isCompressed return true, all stored data object are zipped.
   * The compressed behaviour is defined by the FODBCollection descriptor.
   * To gzip collection allow to gain around 25% of the collection file length but increase the
   * synhro file size for complete synchro.
   * @param gzip
   */
  public boolean isCompressed() {
    return header.getDescriptor().isCompressed();
  }
 
	/**
	 * Init collection encryption with the specified cypher.
	 * Must be call just after collection initialization before storing/reading any data.
	 * Same cypher must be provided to read already stored object.
	 * 
	 * @param cypher : cypher to be use to encrypt data.
	 */
	public void setCollectionCypher(FODBCypher cypher) throws FODBException	{
		if (collCypher != null) throw new FODBException("FODBCollection Cypher already initilazed error.");
		collCypher = cypher;
		try	{
			databaseFile.open();
			try	{
				rewriteHeader();
			} finally	{
				databaseFile.close();
			}
		} catch (Throwable ex)	{
			throw new FODBException(ex);
		}
	}
	
	/**
	 * The function that is called to create a new Collection file.
	 * It just calls the constructor with the argument ACTION_CREATE.
	 * 
	 * @param filePath - same as constructor
	 * @param objType - same as constructor
	 * @return the object created.
	 * @throws FODBDataFileException
	 */
	public static FODBCollectionFile createCollection (String filePath, FODBCollectionDescriptor descriptor, FastObjectDB db) throws IOException, ClassNotFoundException	{
		return new FODBCollectionFile(filePath, descriptor, ACTION_CREATE, db);
	}
	/**
	 * The function that is called to open an existing Collection file.
	 * It just calls the constructor with the argument ACTION_OPEN.
	 * 
	 * @param filePath - same as constructor
	 * @return the object created or null if the file does not exist.
	 * @throws FODBDataFileException
	 */
	public static FODBCollectionFile openCollection (String filePath, String colname, FastObjectDB db) throws IOException, ClassNotFoundException {
		File collectionFile = new File(filePath);
		if (!collectionFile.exists()) {
			return null;
		}
		return new FODBCollectionFile(filePath, new FODBCollectionDescriptor(colname, null), ACTION_OPEN, db);
	}
	
	/**
	 * Add an index to the indexs table.
	 * 
	 * @param indexPos - position of the new index in the file.
	 * @return indexPos (might be suppressed).
	 * @throws FODBDataFileException
	 */
	public synchronized long addIndex(long indexPos)  throws IOException, ClassNotFoundException  {
		boolean modified;
		modified = indexTable.add(indexPos);
		
		rewriteIndexTable(modified);
		return indexPos;
	}
	
	/**
	 * Rewrite the indexs table.
	 * If it grew up (bigger), it's position in the file changes. Then we
	 * have to rewrite the header, with this new position.
	 * 
	 * @param bigger - did it grow up.
	 * @throws FODBDataFileException
	 */
	private synchronized void rewriteIndexTable(boolean bigger)  throws IOException, ClassNotFoundException  {
		if (!bigger) {
			databaseFile.seek(header.getIndexTablePos());
			databaseFile.rewriteObject(indexTable);
		}
		else {
				deleteObject(header.getIndexTablePos());
				long newPos = databaseFile.writeObject(indexTable);
				header.setIndexTablePos(newPos);
				rewriteHeader();
		}
	}
	/**
	 * Read Object at pos.
	 * 
	 * @param pos - position of the object to read.
	 * @return the read object.
	 * @throws FODBDataFileException
	 */
	public Object readNode(long pos) throws IOException, ClassNotFoundException {
		databaseFile.seek(pos);
		return databaseFile.readObject();
	}
	
	/**
	 * Write an object.
	 * 
	 * @param obj - object to wright.
	 * @return - the position of the object written.
	 * @throws FODBDataFileException
	 */
	public long writeNode(Object obj) throws IOException, ClassNotFoundException  {
		return databaseFile.writeObject(obj);
	}	
	
	/**
	 * Rewrite an object
	 * The object has to be the same size it used to be. The function
	 * doesn't check this. It is the role of the developper.
	 * 
	 * @param obj - object to write
	 * @param pos - position in the collection file.
	 * @throws FODBDataFileException
	 */
	public void rewriteNode(Object obj, long pos) throws IOException, ClassNotFoundException  {
		databaseFile.seek(pos);
		databaseFile.rewriteObject(obj);
	}
	
	/**
	 * Read Object at pos.
	 * 
	 * @param pos - position of the object to read.
	 * @return the read object.
	 * @throws FODBDataFileException
	 */
  public Object readObject(long pos) throws IOException, ClassNotFoundException {
    databaseFile.seek(pos);
    if (collCypher == null) {
      ByteArrayInputStream bstr = new ByteArrayInputStream((byte[])databaseFile.readObject());
      ObjectInputStream objstr;
      if (header.getDescriptor().isCompressed()) {
        GZIPInputStream gzstr = new GZIPInputStream(bstr);
        objstr = new ObjectInputStream(gzstr);
      } else {
        objstr = new ObjectInputStream(bstr);
      }
      return objstr.readObject();
    } else  {
      byte[]todecrypt = (byte[])databaseFile.readObject();
      byte[] decrypted = collCypher.decryptObject(todecrypt);
      ByteArrayInputStream bstr = new ByteArrayInputStream(decrypted);
      ObjectInputStream objstr;
      if (header.getDescriptor().isCompressed()) {
        GZIPInputStream gzstr = new GZIPInputStream(bstr);
        objstr = new ObjectInputStream(gzstr);
      } else {
        objstr = new ObjectInputStream(bstr);
      }
      return objstr.readObject();
    }
  }
	
	/**
	 * Write an object.
	 * 
	 * @param obj - object to wright.
	 * @return - the position of the object written.
	 * @throws FODBDataFileException
	 */
  public long writeObject(Object obj) throws IOException, ClassNotFoundException  {
    if (collCypher == null) {
      ByteArrayOutputStream bstr = new ByteArrayOutputStream();
      ObjectOutputStream objstr;
      if (header.getDescriptor().isCompressed()) {
        GZIPOutputStream gzstr = new GZIPOutputStream(bstr);
        objstr = new ObjectOutputStream(gzstr);
      } else {
        objstr = new ObjectOutputStream(bstr);
      }
      objstr.writeObject(obj);
      objstr.flush();
      objstr.close();
      return databaseFile.writeObject(bstr.toByteArray());
    } else  {
      ByteArrayOutputStream bstr = new ByteArrayOutputStream();
      ObjectOutputStream objstr;
      if (header.getDescriptor().isCompressed()) {
        GZIPOutputStream gzstr = new GZIPOutputStream(bstr);
        objstr = new ObjectOutputStream(gzstr);
      } else {
        objstr = new ObjectOutputStream(bstr);
      }
      objstr.writeObject(obj);
      objstr.flush();
      objstr.close();
      byte[] toencrypt = bstr.toByteArray();
      byte[] encrypted = collCypher.encryptObject(toencrypt);
      return databaseFile.writeObject(encrypted);
      
    }
  }
	
	/**
	 * Rewrite an object
	 * The object has to be the same size it used to be. The function
	 * doesn't check this. It is the role of the developper.
	 * 
	 * @param obj - object to write
	 * @param pos - position in the collection file.
	 * @throws FODBDataFileException
	 */
	public void rewriteObjectt(Object obj, long pos) throws IOException, ClassNotFoundException  {
		databaseFile.seek(pos);
		if (collCypher == null)	{
			databaseFile.rewriteObject(obj);
		} else	{
			ByteArrayOutputStream bstr = new ByteArrayOutputStream();
			ObjectOutputStream ostr = new ObjectOutputStream(bstr);
			ostr.writeObject(obj);
			byte[] toencrypt = bstr.toByteArray();
			byte[] encrypted = collCypher.encryptObject(toencrypt);
			databaseFile.rewriteObject(encrypted);
			
		}
	}
	
	/**
	 * Delete an Object
	 * 
	 * @param pos - position of the object to delete.
	 * @throws FODBDataFileException
	 */
	public void deleteObject(long pos) throws IOException, ClassNotFoundException  {
				databaseFile.seek(pos);
				databaseFile.delete();
	}
	
	/**
	 * Write the header at the beginning of the collection file.
	 * 
	 * @throws FODBDataFileException
	 */
	public void writeHeader() throws IOException, ClassNotFoundException  {
				databaseFile.rewind();
				databaseFile.writeObject(header);
	}
	
	public void rewriteHeader() throws IOException, ClassNotFoundException  {
				databaseFile.rewind();
				databaseFile.rewriteObject(header);
	}
	
	/**
	 * Read the header which is situated at the beginning of the
	 * collection file and return it.
	 * 
	 * @return - The header of the file.
	 * @throws FODBDataFileException
	 */
	public FODBCollectionFileHeader readHeader()  throws IOException, ClassNotFoundException  {
				databaseFile.rewind();
				return (FODBCollectionFileHeader)databaseFile.readObject();
	}
	
	/**
	 * Read the indexs table and return it.
	 * 
	 * @return - the indexs table of the collection file.
	 * @throws FODBDataFileException
	 */
	private FODBIndexTable readIndexTable() throws IOException, ClassNotFoundException {
		databaseFile.seek(header.getIndexTablePos());
		return (FODBIndexTable)databaseFile.readObject();
	}
	
	public Class getObjectType() {
		return header.getObjectType();
	}
	
	/**
	 * Read the indexs table. Then get all indexs header, and return
	 * them as a table.
	 * 
	 * @return - a table of indexsHeader.
	 * @throws FODBDataFileException
	 */
	public FODBIndexHeader[] getIndexsHeaders() throws IOException, ClassNotFoundException {
		FODBIndexHeader[] headers;
			int numHeaders = indexTable.size();

			if (numHeaders <= 0) {
				return null;
			}
			headers = new FODBIndexHeader[numHeaders];
			
			for (int i = 0; i < numHeaders; i++) {
				long pos = indexTable.getPosByPlace(i);
				databaseFile.seek(pos);
				headers[i] = (FODBIndexHeader)databaseFile.readObject();
			}
		return headers;
	}
	
  public FODBCollectionDescriptor getDescriptor() {
    return header.getDescriptor();
  }

  /**
	 * Used for debugging.
	 */
	public String toString() {
		String result = "     FODBCollectionFile[" + collectionFilePath + ", " + openCounter + "]\n";
		result = result.concat("" + header + "\n");
		result = result.concat("" + indexTable);
		return result;
	}
	
}
