/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */
package org.openmobileis.database.fastobjectdb.synchro.client;

import java.io.File;
import java.io.IOException;

import org.odbms.ObjectSet;
import org.odbms.Query;
import org.openmobileis.common.util.exception.BadDataFormatException;
import org.openmobileis.common.util.exception.DatabaseException;
import org.openmobileis.database.fastobjectdb.FODBCollectionDescriptor;
import org.openmobileis.database.fastobjectdb.FODBStringIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FastObjectDB;
import org.openmobileis.database.fastobjectdb.db.FODBCollection;
import org.openmobileis.database.fastobjectdb.db.exception.FODBException;
import org.openmobileis.database.fastobjectdb.db.index.FODBIndex;
import org.openmobileis.database.fastobjectdb.db.index.FODBIntIndex;
import org.openmobileis.database.fastobjectdb.db.index.FODBLongIndex;
import org.openmobileis.database.fastobjectdb.db.index.FODBStringIndex;
import org.openmobileis.database.fastobjectdb.db.index.FODBUniqueIndex;
import org.openmobileis.database.fastobjectdb.db.index.node.Node;
import org.openmobileis.database.fastobjectdb.db.store.FODBMainFile;
import org.openmobileis.synchro.client.SynchroManager;
import org.openmobileis.synchro.client.SynchroProcessor;
import org.openmobileis.synchro.journal.JournalManager;
import org.openmobileis.synchro.journal.SimpleLogRenderer;
import org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener;
import org.openmobileis.synchro.openmsp.client.OpenMSPSynchroManager;
import org.openmobileis.synchro.openmsp.client.core.NumSyncManagerDB;
import org.openmobileis.synchro.openmsp.client.db.ActionDB;
import org.openmobileis.synchro.openmsp.client.db.SyncActionDBManager;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Synchronisable version of FastDBObject
 * 
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public class SynchroFastObjectDB extends FastObjectDB {
  private static SynchroFODBConnector syncConnector;
  /**
   * @param dbRootDir
   * @param dbName
   * @param fmf
   * @param action
   * @throws BadDataFormatException
   * @throws FODBDataFileException
   * @throws FODBCollectionException
   */
  private SynchroFastObjectDB(String dbRootDir, String dbName, FODBMainFile fmf, int action)
    throws BadDataFormatException, FODBException {
    super(dbRootDir, dbName, fmf, action);
  }
	
	/**
	 * The function is called to open a database. If the database does not exist, It's created.
	 * 
	 * @param dbRootDir - database directory
	 * @param dbName - database name
	 * @return the database object that has been created.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 * @throws BadDataFormatException
	 * @throws FODBDataFileException
	 * @throws FODBCollectionException
	 */
	public static FastObjectDB open(String dbRootDir, String dbName) throws IOException, ClassNotFoundException, BadDataFormatException, FODBException {
		String dirPath = "" + dbRootDir + "/" + dbName;
		String completeFilePath =  dirPath + "/" + dbName + ".db";
		File file = new File(completeFilePath);
		FastObjectDB db = null;
        SynchroFastObjectDB.initDefaultOpenMSPSynchro();
		if (file.exists())	{
			FODBMainFile fmf = FODBMainFile.openDbFile(completeFilePath);
			db =  new SynchroFastObjectDB(dbRootDir, dbName, fmf, ACTION_OPEN);
		} else	{
			//create directory path if not present.
			File dirFile = new File(dirPath);
			if (!dirFile.exists())	{
				dirFile.mkdirs();
			}
			FODBMainFile fmf = FODBMainFile.createDbFile(completeFilePath);
			db =  new SynchroFastObjectDB(dbRootDir, dbName, fmf, ACTION_CREATE);
		}
		return db;
	}
  
  public static void registerSynchroFODBConnector(SynchroFODBConnector connector)  {
    SynchroFastObjectDB.syncConnector = connector;
  }
	
	/**
	 * Add a new collection to the database.
	 * 
	 * @param collectionName - name of the new colletion
	 * @param objectType - type of objects stored in this collection.
	 * @param synchro - if set to false  the collection wont be synchronized when FODB is in synchro mode.
	 * @return true if it succeded, false otherwise.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 * @throws BadDataFormatException
	 * @throws FODBDataFileException
	 * @throws FODBCollectionException
	 */
	public boolean createCollection(FODBCollectionDescriptor descriptor) throws IOException, ClassNotFoundException, BadDataFormatException, FODBException {
		boolean ret =  super.createCollection(descriptor);
		if (ret)	{
			if (descriptor.isSynchronize())	{
				// remove all sync number.
				NumSyncManagerDB.getManager().deleteSyncNumberForService(descriptor.getCollectionName());
			  syncConnector.registerCollection(descriptor.getCollectionName());
			}
		}
		return ret;
	}
	
	/**
	 * Add a new collection to the database.
	 * the synchro parameter is set to true.
	 * @param collectionName - name of the new colletion
	 * @param objectType - type of objects stored in this collection.
	 * @return true if it succeded, false otherwise.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 * @throws BadDataFormatException
	 * @throws FODBDataFileException
	 * @throws FODBCollectionException
	 */
	public boolean createCollection(String collectionName, Class objectType) throws IOException, ClassNotFoundException, BadDataFormatException, FODBException {
		return this.createCollection(new FODBCollectionDescriptor(collectionName, objectType));
	}
	
	/**
	 * Open an existing collection.
	 * This is not used and should be suppressed.
	 * 
	 * @param collectionName - name of the collection
	 * @return true if it succeded, false otherwise.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 * @throws BadDataFormatException
	 * @throws FODBDataFileException
	 * @throws FODBCollectionException
	 */
	protected boolean openCollection(String collectionName) throws IOException, ClassNotFoundException, BadDataFormatException, FODBException {
		boolean ret =  super.openCollection(collectionName);
		if (ret)	{
			FODBCollection collection = this.getCollection(collectionName);
			if (collection.isSynchronized())	{
              //don't register collection if synchro listener already exist.
              // keep created listener 
              //this method is call after each synchro to reload complete synchro files.
              OpenMSPSyncListener listener = OpenMSPSynchroManager.getManager().getListenerByName(collectionName);
              if (listener == null) {
                syncConnector.registerCollection(collectionName);
              }
			}
		}
		return ret;
	}
	
	/**
	 * Register a synchroListener call during synchronisation processus.
	 * The listener is not persistent, so it must be registered at each application start.
	 * @param collectionName
	 * @param listener
	 */
	public void registerSynchroFODBReturnListener(String collectionName, SynchroFODBReturnListener listener)	{
	  syncConnector.registerSynchroFODBReturnListener(collectionName, listener);
	}
  
  /**
   * Register a a simpe journal log renderer to the synchro journal manager for the specified collection.
   * journalinfo message will be displayed in the journal to identify the collection.
   * The listener is not persistent, so it must be registered at each application start.
   * @param collectionName
   * @param nodeObject
   */
 public void registerJournalLogRenderer(String collectionName, String journalinfo) {
    JournalManager.getManager().registerJournalLogRenderer(new SimpleLogRenderer(collectionName, journalinfo));
  }

	/**
	 * return true if the specified object from the specified collection is in the collection.
	 * return false otherwise
	 * 
	 * @param colName - name of the Collection
	 * @param object - object to look for.
	 * @return true if the object is found.
	 */	
	public boolean isObjectInCollection(String colName, Object obj) throws DatabaseException	{
		FODBCollection collect = colManager.getCollectionByName(colName);
		if (collect != null)	{
			try	{
				this.getTransactionManager().begin();
				try	{
					this.getTransactionManager().enterTransaction(collect.getName());
					FODBUniqueIndex index = collect.getCollectionIdIndex();
					Object key  = ((FODBIndex)index).getKey(obj);
					long pos = index.getKeyPosition(key);
					return (pos != Node.NO_NODE);
				} finally	{
					this.getTransactionManager().commit();
				}
			} catch (Throwable ex)	{
				throw new DatabaseException(ex);
			}			
		}
		return false;
	}
	
	/**
	 * return the object from the specified collection with specified id.
	 * id is the string format of the id object.
	 * If id is an int for exemple.  The specified parameter is converted to an int.
	 * 
	 * @param colName - name of the Collection
	 * @param id - object id.
	 * @return the object with specified id or null if not found.
	 */	
	public Object getObjectFromCollectionWithId(String colName, String id) throws DatabaseException	{
		FODBCollection collect = colManager.getCollectionByName(colName);
		Object objId = id;
        if (collect != null)	{
			try	{
				FODBUniqueIndex index = collect.getCollectionIdIndex();
                if (index instanceof FODBLongIndex) {
                  objId = Long.getLong(id);
                } else if (index instanceof FODBIntIndex) {
                  objId = Integer.getInteger(id);
                }
				Query q = this.query();
				q.constrain(collect.getCollectionObjectClass());
				Query subq = q.descend(((FODBIndex)index).getIndexHeader().getMemberName());
				subq.constrain(objId).equal();
				ObjectSet set = q.execute(); 
				if (set.hasNext())	{
					return set.next();
				} else	{
					return null;
				}
			} catch (Throwable ex)	{
				throw new DatabaseException(ex);
			}			
		}
		return null;
	}
	
	/**
	 * Add an element to a Collection of the database without saving modification
	 * data needed to synchronize synchronizable collection.
	 * Use by synchro to add data that come from the server.
	 * 
	 * @param colName - Name of the collection
	 * @param obj - object to add.
	 * @return - true if it succeded, false otherwise.
	 * @throws FODBDataFileException
	 * @throws FODBCollectionException
	 * @throws FODBIndexException
	 * @throws BTreeException
	 */
	public boolean addFODB(String colName, Object obj) throws FODBException {
		return super.add(colName, obj);
	}
	
	/**
	 * Add an element to a Collection of the database.
	 * 
	 * @param colName - Name of the collection
	 * @param obj - object to add.
	 * @return - true if it succeded, false otherwise.
	 * @throws FODBException
	 */
	public boolean add(String colName, Object obj) throws FODBException {
		boolean ret = false;
		this.getTransactionManager().begin();
		try	{
			ret = super.add(colName, obj);
			if (ret)	{
				FODBCollection col = this.colManager.getCollectionByName(colName);
				if (col.isSynchronized())	{
					this.getTransactionManager().enterTransaction("FWKACTIONDB");
					FODBIndex index = (FODBIndex)col.getCollectionIdIndex();
                    if ((index != null) && (index instanceof FODBStringIndex)){
                      FODBStringIndex strindex = (FODBStringIndex)index;
                        String key = (String)strindex.getKeySensitive(obj, ((FODBStringIndexDescriptor)strindex.getIndexDescriptor()).isCaseSensitive());
                        SyncActionDBManager.getManager().addActionDB(new ActionDB(colName, key, ActionDB.ADD_ACTION));
                    } else if ((index != null) && (index instanceof FODBLongIndex)){
                      Long key = (Long)((FODBLongIndex)index).getKey(obj);
                      SyncActionDBManager.getManager().addActionDB(new ActionDB(colName, key.toString(), ActionDB.ADD_ACTION));
                    } else if ((index != null) && (index instanceof FODBIntIndex)){
                      Integer key = (Integer)((FODBIntIndex)index).getKey(obj);
                      SyncActionDBManager.getManager().addActionDB(new ActionDB(colName, key.toString(), ActionDB.ADD_ACTION));
					} else	{
						col.deleteSingleElement(obj);
						throw new FODBException("No string unique index. Object not Synchronized, cancel db writing");
					}
				}
			}
		} finally	{
			this.getTransactionManager().commit();
		}
		return ret;
	}
	
	/**
	 * Replace an existing element in a Collection of the database without saving modification
	 * data needed to synchronize synchronizable collection.
	 * Use by synchro to replace data that come from the server.
	 * 
	 * @param colName - Name of the Collection
	 * @param obj - Object to replace.
	 * @return true if it succeded, false otherwise.
	 * @throws FODBException
	 */
	public boolean replaceFODB(String colName, Object obj) throws FODBException {
		return super.replace(colName, obj);
	}
	
	/**
	 * Replace an existing element in a Collection of the database.
	 * 
	 * @param colName - Name of the Collection
	 * @param obj - Object to replace.
	 * @return true if it succeded, false otherwise.
	 * @throws FODBException
	 */
	public boolean replace(String colName, Object obj) throws FODBException {
		boolean ret = false;
		this.getTransactionManager().begin();
		try	{
			ret = super.replace(colName, obj);
			if (ret)	{
				FODBCollection col = this.colManager.getCollectionByName(colName);
				if (col.isSynchronized())	{
					this.getTransactionManager().enterTransaction("FWKACTIONDB");
					FODBIndex index = (FODBIndex)col.getCollectionIdIndex();
					if ((index != null)	&& (index instanceof FODBStringIndex)){
                         FODBStringIndex strindex = (FODBStringIndex)index;
						String key = (String)((FODBStringIndex)index).getKeySensitive(obj, ((FODBStringIndexDescriptor)strindex.getIndexDescriptor()).isCaseSensitive());
						SyncActionDBManager.getManager().addActionDB(new ActionDB(colName, key, ActionDB.UPDATE_ACTION));
                     } else if ((index != null) && (index instanceof FODBLongIndex)){
                        Long key = (Long)((FODBLongIndex)index).getKey(obj);
                        SyncActionDBManager.getManager().addActionDB(new ActionDB(colName, key.toString(), ActionDB.UPDATE_ACTION));
                      } else if ((index != null) && (index instanceof FODBIntIndex)){
                        Integer key = (Integer)((FODBIntIndex)index).getKey(obj);
                        SyncActionDBManager.getManager().addActionDB(new ActionDB(colName, key.toString(), ActionDB.UPDATE_ACTION));
					} else	{
						col.deleteSingleElement(obj);
						throw new FODBException("No string unique index. Object not Synchronized, cancel db writing");
					}
				}
			}
		} finally	{
			this.getTransactionManager().commit();
		}
		return ret;
	}
	
	/**
	 * Delete an element of a Collection
	 * 
	 * @param colName - name of the Collection.
	 * @param obj - object to delete.
	 * @return true if it succeded, false otherwise.
	 * @throws FODBCollectionException
	 * @throws FODBIndexException
	 */
	public boolean delete(String colName, Object obj) throws FODBException {
		boolean ret = false;
		this.getTransactionManager().begin();
		try	{
			ret = super.delete(colName, obj);
			if (ret)	{
				FODBCollection col = this.colManager.getCollectionByName(colName);
				if (col.isSynchronized())	{
					this.getTransactionManager().enterTransaction("FWKACTIONDB");
					FODBIndex index = (FODBIndex)col.getCollectionIdIndex();
					if ((index != null)	&& (index instanceof FODBStringIndex)){
						String key = (String)((FODBStringIndex)index).getKeySensitive(obj, false);
						SyncActionDBManager.getManager().addActionDB(new ActionDB(colName, key, ActionDB.DELETE_ACTION));
                    } else if ((index != null) && (index instanceof FODBLongIndex)){
                      Long key = (Long)((FODBLongIndex)index).getKey(obj);
                      SyncActionDBManager.getManager().addActionDB(new ActionDB(colName, key.toString(), ActionDB.DELETE_ACTION));
                    } else if ((index != null) && (index instanceof FODBIntIndex)){
                      Integer key = (Integer)((FODBIntIndex)index).getKey(obj);
                      SyncActionDBManager.getManager().addActionDB(new ActionDB(colName, key.toString(), ActionDB.DELETE_ACTION));
					} else	{
						throw new FODBException("No string unique index. Object not Synchronized, cancel db writing");
					}
				}
			}
		} finally	{
			this.getTransactionManager().commit();
		}
		return ret;
	}
	
	/**
	 * Delete an object of a Collection with its id
	 * Doas not generate syncho action.
	 * 
	 * @param colName - name of the Collection.
	 * @param id - id of the object to delete.
	 * @return true if it succeded, false otherwise.
	 * @throws FODBCollectionException
	 * @throws FODBIndexException
	 */
	public boolean deleteWithIdFODB(String colName, Object id) throws FODBException {
		return super.deleteWithId(colName, id);
	}
	/**
	 * Delete an object of a Collection with its id
	 * 
	 * @param colName - name of the Collection.
	 * @param id - id of the object to delete.
	 * @return true if it succeded, false otherwise.
	 * @throws FODBCollectionException
	 * @throws FODBIndexException
	 */
	public boolean deleteWithId(String colName, Object id) throws FODBException {
		boolean ret = false;
		this.getTransactionManager().begin();
		try	{
			ret = super.deleteWithId(colName, id);
			if (ret)	{
				FODBCollection col = this.colManager.getCollectionByName(colName);
				if (col.isSynchronized())	{
					this.getTransactionManager().enterTransaction("FWKACTIONDB");
					FODBIndex index = (FODBIndex)col.getCollectionIdIndex();
					if ((index != null)	&& (index instanceof FODBStringIndex)){
						SyncActionDBManager.getManager().addActionDB(new ActionDB(colName, (String)id, ActionDB.DELETE_ACTION));
					} else	{
						throw new FODBException("No string unique index. Object not Synchronized, cancel db writing");
					}
				}
			}
		} finally	{
			this.getTransactionManager().commit();
		}
		return ret;
	}
  
  private static void initDefaultOpenMSPSynchro() {
    SynchroProcessor openMSPproc = SynchroManager.getManager().getSynchroProcessor(OpenMSPSynchroManager.OPENMSPPROCESSORNAME);
    if (openMSPproc == null)  {
      SynchroManager.getManager().registerSynchroProcessor(OpenMSPSynchroManager.getManager());
    }
    if (syncConnector == null)  {
      syncConnector = new DefaultOpenMSPSynchroFODBConnector();
    }

  }

}
