/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */
package org.openmobileis.database.fastobjectdb.synchro.server;

import org.openmobileis.common.user.UserNotFoundException;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.exception.ServiceException;
import org.openmobileis.database.fastobjectdb.FastObjectDB;
import org.openmobileis.synchro.algo.replication.SynchroAtomicObject;
import org.openmobileis.synchro.algo.replication.SynchroConflicResolver;
import org.openmobileis.synchro.openmsp.OpenMSPException;
import org.openmobileis.synchro.security.auth.Credential;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: For each Terminal synchronized collection you have to register a FODBSyncTarget. This target act has a database connector for the synchro service.
 * To specify the temp directory where generated db file (during complete synchronisation) is use the property :org.openmobileis.database.fastobjectdb.synchro.server.generateddb.installpath
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public interface FODBSyncTarget {
  
    /**
     * @return the name to the terminal FODB collection.
     */
	public String getCollectionName();
    
    /**
     * Synchro meta data added to the synchro on the terminal side are given be the 
     * synchro process.
     * This method is call before all other synchro method.
     * @param metadata
     * @throws OpenMSPException
     */
    public void setSendSynchroMetaData(Object metadata)throws OpenMSPException;

    /**
     * It is use by the synchro service to get an object from the database when it has to be updated on the terminal.
     * @return the collection object with the specified ID.
     */
	public Object getCollectionObjectWithId(String id) throws OpenMSPException ;

    /**
     * Implement the update part of the database. Call by the synchro service when a object has been modified on the terminal.
     * There's only an update method, no insert because on the server side we can't know if it's an update on an insert. Because of synchronisation problems,
     * a created object on the terminal can be synchronized two time. So for each modified object on the terminal you have to validate if it's already exist.
     * @param obj is the instance of the modified object.
     * @throws OpenMSPException : throws when there is a database update problem. The synchro collection will be in error in the terminal synchro journal.
     * The object will come back in the next synchronisation.
     */
    public void updateCollectionObject(Object obj) throws OpenMSPException ;
	
    /**
     * Implement the delete part of the database. Call by the synchro service when a object has been deleted on the terminal.
     * @param id is the identifiant of the deleted object.
     * @throws OpenMSPException :throws when there is a database delete problem. The synchro collection will be in error in the terminal synchro journal.
     * The id will come back in the next synchronisation.
     */
    public void deleteCollectionObject(String id) throws OpenMSPException ;

    /**
     * When a conflic is detected, the synchro service use this conflic resolver to determine the action to do.
     *  We provide default conflic resolver behaviours.
     * @see org.openmobileis.synchro.algo.replication.AlwaysUpdateClientSynchroConflicResolver
     * @see org.openmobileis.synchro.algo.replication.AlwaysUpdateServerSynchroConflicResolver
     * @return Return the conflic resolver use to manage synchro conflic.
     */
    public SynchroConflicResolver getConflicResolver();
    
    /**
     * It's call when a complete synchronisation is done to fill the terminal database with all object. If no complete synchronisation can occurs (getUpdateMaxNbRow() always return -1),
     * the method is never call. So no implementation is needed.
     * Complete synchronisation is a convenient way to optimize the terminal synchronisation. some time it's make less time to generate
     * the Terminal complete collection on the server side than to synchronize all modifications.
     * @return Return all the object of the collection for the user that initiate the collection.
     * @throws OpenMSPException : throws when there is a problem to invalidate the synchronisation. The synchro collection will be in error in the terminal synchro journal.
     */
	public Array getAllCollectionObject()  throws OpenMSPException;

	/**
	 * Return all atomic object uid modified since specified syncnumber. Syncnumber is a number incremented for each synchro request. It can be converted to
     * a timestamp with the SyncNumberManager. It's preferable to use number increment for synchronisation than timestamp to avoid desynchronized clock problem.
     * Terminal and server can be desynchronized at the system cloc level of a few second. A synchronisation can elapse less than a few second. This can leads to 
     * synchronisation problem that can be difficult to track.
	 */
	public SynchroAtomicObject[] getAllModifiedAtomicObjectSince(long syncnumber) throws OpenMSPException;

	/**
		Return the max number of row that update file can contains
		If more than max row must be updated on the PDA, a complete synchronisation will be started, the database file
		will be send.
		If return -1 (default implementation) max row in not taken into account (ie always update file)
	*/
	public int getUpdateMaxNbRow();
	
	/**
	 * call by synchronisation service during complete synchronisation to update the generated db collection.
	 * Use to modifiy db objects before the db files are send to the Terminal.
     * Empty is nothing has to ne done.
	 * @param db the database collection (thedb contains only the synchronized collection).
	 */
	public void updateSynchroDB(FastObjectDB db) throws OpenMSPException;
  
    /**
     * call at the beginning of the collection synchronisation.
     * @param cred : synchronisation credential
     * @throws UserNotFoundException
     * @throws ServiceException
     */
    public void connect(Credential cred) throws UserNotFoundException, ServiceException;
    
    /**
     * call at the end of the collection synchronisation.
     * @param cred : synchronisation credential
     * @throws UserNotFoundException
     * @throws ServiceException
     */
    public void disconnect();

}
