/**
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */
package org.openmobileis.examples.mycrm.data.fodb;

import org.odbms.ObjectSet;
import org.odbms.Query;
import org.openmobileis.common.context.ApplicationContextManager;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.exception.DatabaseException;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.database.fastobjectdb.FODBCollectionDescriptor;
import org.openmobileis.database.fastobjectdb.FODBIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FODBStringIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FastObjectDB;
import org.openmobileis.database.fastobjectdb.FastObjectDBManager;
import org.openmobileis.database.fastobjectdb.db.query.soda.FODBSodaObjectSet;
import org.openmobileis.database.fastobjectdb.synchro.client.SynchroFastObjectDB;
import org.openmobileis.examples.mycrm.data.Leads;

public final class LeadsFactory {

  private static LeadsFactory manager;

  private LeadsFactory() {
    super();
    this.initDB();
  }

  public static LeadsFactory getManager() {
    if (manager == null) {
      synchronized (LeadsFactory.class) {
        if (manager == null) {
          manager = new LeadsFactory();
          //register the manager to avoid class garbage.
          ApplicationContextManager.getManager().addManager(manager);
        }
      }
    }
    return manager;
  }

  private void initDB() {
    try {
      FastObjectDB db = FastObjectDBManager.getCurrentFODB();
      // test if the collection has already been created.
      //if not create it. Created collection are automatiquely 
      // loaded when FODB is initialized.
      if (!db.isCollectionExist("leads")) {
        //SimpleOpenMISInit intialize the synchronized version of FODB.
        // all created collection are synchronized by default.
        // Synchro listener are automatiquely registered when the collection is created or loaded.
        // use the FODBCollectionDescriptor to define a collection that is not synchronized.
        FODBCollectionDescriptor coldes = new FODBCollectionDescriptor("leads", Leads.class);
        db.createCollection(coldes);
        //add index to query the collection by id
        FODBStringIndexDescriptor IDDescriptor = new FODBStringIndexDescriptor("ID", FODBIndexDescriptor.UNIQUE, "getId()", 12, 26);
        db.addIndex("leads", IDDescriptor);
        FODBStringIndexDescriptor AccountrDescriptor = new FODBStringIndexDescriptor("ACCOUNT", FODBIndexDescriptor.MULTIPLE, "getIdaccount()", 12, 26);
        db.addIndex("leads", AccountrDescriptor);
        //only not closed leads are synchronized.
      }
      //register the log renderer.
      //logrenderer are use to show synchronization result for this collection.
      ((SynchroFastObjectDB) db).registerJournalLogRenderer("leads", "Leads synchronization :");

    } catch (Throwable ex) {
      LogManager.traceError(0, "account Factory FODB ERROR : Unknown error during creation");
      LogManager.traceError(0, ex);
    }
  }

  public Leads getLeads(String id) throws DatabaseException {
    Query q = FastObjectDBManager.getCurrentFODB().query();
    q.constrain(Leads.class);
    Query subq = q.descend("getId()");
    subq.constrain(id).equal();
    ObjectSet set = q.execute();
    if (set.hasNext()) {
      return (Leads) set.next();
    } else {
      return null;
    }
  }

  public Array getAllLeads() throws DatabaseException {
    Query q = FastObjectDBManager.getCurrentFODB().query();
    q.constrain(Leads.class);
    q.descend("getId()");
    ObjectSet set = q.execute(); 
    return ((FODBSodaObjectSet)set).getArrayFromSet();
  }

  public Array getLeadsForAccount(String accountid) throws DatabaseException {
    Query q = FastObjectDBManager.getCurrentFODB().query();
    q.constrain(Leads.class);
    Query subq = q.descend("getIdaccount()");
    subq.constrain(accountid).equal();
    ObjectSet set = q.execute(); 
    return ((FODBSodaObjectSet)set).getArrayFromSet();
  }
}
