/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */
package org.openmobileis.examples.server.synchro;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.openmobileis.common.user.UserNotFoundException;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.database.ManagerDB;
import org.openmobileis.common.util.exception.ServiceException;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.common.util.log.LogServices;
import org.openmobileis.database.fastobjectdb.FastObjectDB;
import org.openmobileis.database.fastobjectdb.synchro.server.FODBSyncTarget;
import org.openmobileis.examples.server.database.JDBCBaseDataFactory;
import org.openmobileis.examples.simpleappli.data.BaseData;
import org.openmobileis.synchro.algo.replication.AlwaysUpdateServerSynchroConflicResolver;
import org.openmobileis.synchro.algo.replication.SynchroAtomicObject;
import org.openmobileis.synchro.algo.replication.SynchroConflicResolver;
import org.openmobileis.synchro.algo.replication.utils.DefaultSynchroAtomicObject;
import org.openmobileis.synchro.algo.replication.utils.SynchroAtomicObjectManager;
import org.openmobileis.synchro.openmsp.OpenMSPException;
import org.openmobileis.synchro.security.auth.Credential;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Example of a FODBSyncTarget. FODB synchronisation is done per collection. Each collection is synchronized independently.
 * FODBSyncTarget is the server side collection synchro connector to the server database.
 * In the example the server database is a HSQL db with one table. We use a JDBCBaseDataFactory that make SQL query and
 * the mapping beween BaseData object and SQL in made by BaseDataFODBSyncTarget method.
 * There is two type of synchronisation :
 *  * incrementale synchronisation: Terminal modification is updated server side and if everything is ok, server side modifications 
 *     are send to the terminal and updated in the terminal.
 *  * full or complete synchronisation : Terminal modification is updated server side and if everything is ok, the terminal collection
 *     file is generated server side and send to the terminal and updated in one file copy. This mode is done for the terminal init database (first synchro)
 *     or when there is too much server side object getUpdateMaxNbRow(). When there is too much object it's more efficient to generate the object collection
 *     on the server side and to copy the collection file on the terminal than to update each object one at a time. getUpdateMaxNbRow() return the threshold. When there is more object
 *     to update, the synchronization is forced to be full for the collection.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class BaseDataFODBSyncTarget implements FODBSyncTarget {
  private JDBCBaseDataFactory dbfactory;
  /**
   * 
   */
  public BaseDataFODBSyncTarget() {
    super();
    dbfactory = new JDBCBaseDataFactory();
    try {
      dbfactory.createSyncTable();
    }  catch (Throwable ex) {
      LogManager.traceError(LogServices.SERVERSERVICE, "Error during BaseData db creationr :"+ex.getMessage());
    }
  }

  /* (non-Javadoc)
   * @see org.openmobileis.database.fastobjectdb.synchro.server.FODBSyncTarget#getCollectionName()
   */
  public String getCollectionName() {
    // TODO Auto-generated method stub
    return "basedata";
  }
  
  public void setSendSynchroMetaData(Object metadata)throws OpenMSPException    {
    
  }

  /* (non-Javadoc)
   * @see org.openmobileis.database.fastobjectdb.synchro.server.FODBSyncTarget#getCollectionObjectWithId(java.lang.String)
   */
  public Object getCollectionObjectWithId(String id) throws OpenMSPException {
      String[] params = {id};
      BaseData data = null;  //if return object is null, an error is generated by the synchro algo.
      try {
        ResultSet result = dbfactory.getBaseData(params);
        if (result.next()) {
          data = new BaseData(result.getString(1));
          data.setData(result.getString(2));
        }
      }  catch (Throwable ex) {
         throw new OpenMSPException(ex);
      } finally {
        //for select query the ManagerDB connection must be garbaged explicitly.
        ManagerDB.getManager().garbageOpenedConnection();;
      }
      return data;
  }

  /* (non-Javadoc)
   * @see org.openmobileis.database.fastobjectdb.synchro.server.FODBSyncTarget#updateCollectionObject(java.lang.Object)
   */
  public void updateCollectionObject(Object obj) throws OpenMSPException {
    BaseData data = (BaseData)obj;
   
    try {
      Connection connection = dbfactory.getDbManager().getConnection();
      connection.setAutoCommit(false);
      // set manuel commit management to put in the same transaction db modication and SynchroAtomicObject update
      try {
        BaseData dbData = (BaseData)this.getCollectionObjectWithId(data.getId());
        if (dbData == null) {
          String[] params = {data.getId(), data.getData()};
          dbfactory.create(params);       
        } else  {
          String[] params = {data.getData(), data.getId()};
          dbfactory.update(params);       
        }        
        SynchroAtomicObjectManager.getManager().updateAtomicObject(new DefaultSynchroAtomicObject(data.getId(), ""), this.getCollectionName(), null);
        //user link is not use.
        connection.commit();
      } catch (Throwable ex)  {
        connection.rollback();
        throw new OpenMSPException(ex);
      } finally {
        connection.setAutoCommit(true);
        //for insert, update, delete query the ManagerDB connection is automaticly garbaged. Explicit garbage is not needed
      }
    }  catch (SQLException ex) {
       throw new OpenMSPException(ex);
    }
 }

  /* (non-Javadoc)
   * @see org.openmobileis.database.fastobjectdb.synchro.server.FODBSyncTarget#deleteCollectionObject(java.lang.String)
   */
  public void deleteCollectionObject(String id) throws OpenMSPException {
    try {
        String[] params = {id};
        dbfactory.delete(params);       
    }  catch (Throwable ex) {
       throw new OpenMSPException(ex);
    }
    //for insert, update, delete query the ManagerDB connection is automaticly garbaged. Explicit garbage is not needed
  }

  /* (non-Javadoc)
   * @see org.openmobileis.database.fastobjectdb.synchro.server.FODBSyncTarget#getConflicResolver()
   */
  public SynchroConflicResolver getConflicResolver() {
    //if there is a conflic always update the server.
   return new AlwaysUpdateServerSynchroConflicResolver();
  }

  /* (non-Javadoc)
   * @see org.openmobileis.database.fastobjectdb.synchro.server.FODBSyncTarget#getAllCollectionObject()
   */
  public Array getAllCollectionObject() throws OpenMSPException {
    Array ret = new Array();
    try {
      ResultSet result = dbfactory.getAllBaseData();
      while (!result.next()) {
        BaseData data = new BaseData(result.getString(1));
        data.setData(result.getString(2));
        ret.add(data);
      }
    }  catch (Throwable ex) {
       throw new OpenMSPException(ex);
    } finally {
      //for select query the ManagerDB connection must be garbaged explicitly.
      ManagerDB.getManager().garbageOpenedConnection();;
    }
    return ret;
  }

  /* (non-Javadoc)
   * @see org.openmobileis.database.fastobjectdb.synchro.server.FODBSyncTarget#getAllModifiedAtomicObjectSince(long)
   */
  public SynchroAtomicObject[] getAllModifiedAtomicObjectSince(long syncnumber) throws OpenMSPException {
    try {
      return SynchroAtomicObjectManager.getManager().getAllModifiedAtomicObjectForServiceSince(this.getCollectionName(), syncnumber, null);
    } catch (Throwable ex)  {
      throw new OpenMSPException(ex);
    }
  }

  /* (non-Javadoc)
   * @see org.openmobileis.database.fastobjectdb.synchro.server.FODBSyncTarget#getUpdateMaxNbRow()
   */
  public int getUpdateMaxNbRow() {
    return 2;
  }

  /* (non-Javadoc)
   * @see org.openmobileis.database.fastobjectdb.synchro.server.FODBSyncTarget#updateSynchroDB(org.openmobileis.database.fastobjectdb.FastObjectDB)
   */
  public void updateSynchroDB(FastObjectDB db) throws OpenMSPException {
    //do nothing because we have nothiing to change on the server side generated db.
  }

  /* (non-Javadoc)
   * @see org.openmobileis.database.fastobjectdb.synchro.server.FODBSyncTarget#connect(org.openmobileis.synchro.security.auth.Credential)
   */
  public void connect(Credential cred) throws UserNotFoundException, ServiceException {
  }

  /* (non-Javadoc)
   * @see org.openmobileis.database.fastobjectdb.synchro.server.FODBSyncTarget#disconnect()
   */
  public void disconnect() {
  }

}
