/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */
package org.openmobileis.examples.simpleappli.data.FODB;

import org.odbms.ObjectSet;
import org.odbms.Query;
import org.openmobileis.common.context.ApplicationContextManager;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.exception.DatabaseException;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.database.fastobjectdb.FODBCollectionDescriptor;
import org.openmobileis.database.fastobjectdb.FODBIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FODBStringIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FastObjectDB;
import org.openmobileis.database.fastobjectdb.FastObjectDBManager;
import org.openmobileis.database.fastobjectdb.db.query.soda.FODBSodaObjectSet;
import org.openmobileis.database.fastobjectdb.synchro.client.SynchroFastObjectDB;
import org.openmobileis.examples.simpleappli.data.BaseData;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Provide all BaseData FODB access. It use the singleton pattern to
 * facilitate its use. The singleton is registered in the ApplicationContextManager to
 * avoid class GC.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class FODBBaseDataFactory {
  private static FODBBaseDataFactory manager;

  /**
   * 
   */
  private FODBBaseDataFactory() {
    super();
    this.initDB();
  }

  public static FODBBaseDataFactory getManager() {
    if (manager == null) {
      synchronized (FODBBaseDataFactory.class) {
        if (manager == null) {
          manager = new FODBBaseDataFactory();
          //register the manager to avoid class garbage.
          ApplicationContextManager.getManager().addManager(manager);
        }
      }
    }
    return manager;
  }

  private void initDB() {
    try {
      FastObjectDB db = FastObjectDBManager.getCurrentFODB();
      // test if the collection has already been created.
      //if not create it. Created collection are automatiquely 
      // loaded when FODB is initialized.
      if (!db.isCollectionExist("basedata")) {
        //SimpleOpenMISInit intialize the synchronized version of FODB.
        // all created collection are synchronized by default.
        // Synchro listener are automatiquely registered when the collection is created or loaded.
        // use the FODBCollectionDescriptor to define a collection that is not synchronized.
        FODBCollectionDescriptor coldes = new FODBCollectionDescriptor("basedata", BaseData.class);
        db.createCollection(coldes);
        //add index to query the collection
        FODBStringIndexDescriptor IDDescriptor = new FODBStringIndexDescriptor("ID", FODBIndexDescriptor.UNIQUE, "getId()", 12, 26);
        db.addIndex("basedata", IDDescriptor);
      }
      //register the log renderer.
      //logrenderer are use to show synchronization result for this collection.
      ((SynchroFastObjectDB) db).registerJournalLogRenderer("basedata", "Base data synchronization :");

    } catch (Throwable ex) {
      LogManager.traceError(0, "BaseData Factory FODB ERROR : Unknown error during creation");
      LogManager.traceError(0, ex);
    }
  }

  public BaseData[] getAllBaseData() throws DatabaseException {
    Query q = FastObjectDBManager.getCurrentFODB().query();
    q.constrain(BaseData.class);
    q.descend("getId()");
    ObjectSet set = q.execute();
    Array ret = ((FODBSodaObjectSet)set).getArrayFromSet();
    BaseData[] datalist = new BaseData[ret.size()];
    ret.toArray(datalist);
    return datalist;
  }

  public BaseData getBaseData(String id) throws DatabaseException {
    Query q = FastObjectDBManager.getCurrentFODB().query();
    q.constrain(BaseData.class);
    Query subq = q.descend("getId()");
    subq.constrain(id).equal();
    ObjectSet set = q.execute();
    if (set.hasNext()) {
      return (BaseData) set.next();
    } else {
      return null;
    }
  }

  /**
   * call by synchro server to create the representant DB
   */
  public void storeBaseData(BaseData data) throws DatabaseException {
    BaseData dbdata = this.getBaseData(data.getId());
    if (dbdata == null) {
      FastObjectDBManager.getCurrentFODB().add("basedata", data);
    } else {
      FastObjectDBManager.getCurrentFODB().replace("basedata", data);
    }
  }
  
  // delete a data
  public void deleteBaseData (String id) throws DatabaseException {
    try {     
      FastObjectDB db = FastObjectDBManager.getCurrentFODB();
      db.deleteWithId("basedata", id); // can only be call if the first collection index is an UNIQUE index.
    } catch (Throwable ex)  {
      throw new DatabaseException(ex);
    }
  }
}
