/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2005 Modification F.Salque
 * 
 */
package org.openmobileis.modules.common.database.fodb;

import org.odbms.Constraint;
import org.odbms.ObjectSet;
import org.odbms.Query;
import org.openmobileis.database.fastobjectdb.FODBIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FODBStringIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FastObjectDB;
import org.openmobileis.database.fastobjectdb.FastObjectDBManager;
import org.openmobileis.database.fastobjectdb.db.exception.FODBException;
import org.openmobileis.database.fastobjectdb.synchro.client.SynchroFastObjectDB;
import org.openmobileis.common.context.SessionContext;
import org.openmobileis.common.context.SessionContextManager;
import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.common.util.PersistentPropertiesManager;
import org.openmobileis.modules.common.data.TerminalUser;
import org.openmobileis.modules.common.data.TerminalUserFactory;
import org.openmobileis.common.util.exception.BadDataFormatException;
import org.openmobileis.common.util.exception.DatabaseException;
import org.openmobileis.common.util.exception.ServiceException;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.common.util.log.LogServices;

public abstract class FODBTerminalUserFactory implements TerminalUserFactory {

  /**
   * Constructor for FODBTerminalUserFactory.
   */
  public FODBTerminalUserFactory() {
    super();
    String dbPath = PropertiesManager.getManager().getProperty("fastobjectdb.database.path");
    if (dbPath == null) {
      LogManager.traceError(0, "TerminalUser Factory FODB ERROR : FODB path property is not set. Can't create db");
    }
    try {
      FastObjectDB db = FastObjectDBManager.getCurrentFODB();
      if (!db.isCollectionExist(this.getCollectionName())) {
        db.createCollection(this.getCollectionName(), this.getDataType());
        FODBStringIndexDescriptor IDDescriptor = new FODBStringIndexDescriptor("ID", FODBIndexDescriptor.UNIQUE, "getTerminalUserId()", 12, this
            .getMaxTerminalUserIdLength());
        db.addIndex(this.getCollectionName(), IDDescriptor);

        this.initDB(db);
      }
      // always register notification listener because it's not persistent.
      ((SynchroFastObjectDB) db).registerSynchroFODBReturnListener(this.getCollectionName(), this.getSynchroListener());
      // register the log renderer.
      ((SynchroFastObjectDB) db).registerJournalLogRenderer(this.getCollectionName(), "Users");
      this.updateSession();

    } catch (Throwable ex) {
      LogManager.traceError(0, "TerminalUser Factory FODB ERROR : Unknown error during creation");
      LogManager.traceError(0, ex);
    }
  }

  private void updateSession() {
    SessionContext context = (SessionContext) SessionContextManager.getManager().getSessionContext();
    TerminalUser rep = this.getInstallTerminalUser();
    if (rep != null) {
      context.setUserId(rep.getTerminalUserId());
      context.setUserLocale(rep.getLangue());
    }
  }

  public String[] getTerminalUserIdsList() {
    try {
      TerminalUser[] repList = this.getAllTerminalUsers();
      String[] repIds = new String[repList.length];
      for (int i = 0; i < repList.length; i++) {
        repIds[i] = repList[i].getTerminalUserId();
      }
      return repIds;
    } catch (Throwable ex) {
      LogManager.traceError(0, ex);
    }
    return new String[0];
  }

  public TerminalUser getTerminalUser(String id) {
    try {
      Query q = FastObjectDBManager.getCurrentFODB().query();
      q.constrain(this.getDataType());
      Query subq = q.descend("getTerminalUserId()");
      Constraint c = subq.constrain(id).equal();
      ObjectSet set = q.execute();
      if (set.hasNext()) {
        return (TerminalUser) set.next();
      } else {
        return null;
      }
    } catch (Exception e) {
      LogManager.traceError(LogServices.DATABASESERVICE, e);
    }
    return null;
  }

  public TerminalUser getInstallTerminalUser() {
    String repId = PersistentPropertiesManager.getManager().getProperty("FWKTerminalUser", "installrepid");
    if (repId != null) {
      TerminalUser rep = this.getTerminalUser(repId);
      return rep;
    }
    return null;
  }

  public void setInstallTerminalUser(String id) throws ServiceException, DatabaseException {
    PersistentPropertiesManager.getManager().saveProperty("FWKTerminalUser", "installrepid", id);
  }

  public TerminalUser[] getAllTerminalUsers() throws DatabaseException {
    try {
      Query q = FastObjectDBManager.getCurrentFODB().query();
      q.constrain(this.getDataType());
      Query subq = q.descend("getTerminalUserId()");
      ObjectSet set = q.execute();
      TerminalUser[] repList = new TerminalUser[set.size()];
      int i = 0;
      while (set.hasNext()) {
        repList[i++] = (TerminalUser) set.next();
      }
      return repList;
    } catch (Exception e) {
      LogManager.traceError(LogServices.DATABASESERVICE, e);
    }
    return new TerminalUser[0];
  }

  /**
   * call by synchro server to create the TerminalUser DB
   */
  public void addTerminalUser(TerminalUser rep) throws ServiceException, DatabaseException {
    TerminalUser dbrep = this.getTerminalUser(rep.getTerminalUserId());
    try {
      if (dbrep == null) {
        FastObjectDBManager.getCurrentFODB().add(this.getCollectionName(), rep);
      } else {
        FastObjectDBManager.getCurrentFODB().replace(this.getCollectionName(), rep);
      }
    } catch (Throwable ex) {
      throw new ServiceException(ex);
    }
  }

  public void removeTerminalUser(String repId) throws ServiceException, DatabaseException {
    try {
      FastObjectDBManager.getCurrentFODB().deleteWithId(this.getCollectionName(), repId);
    } catch (Throwable ex) {
      throw new ServiceException(ex);
    }
  }

  protected TerminalUserSynchroFODBReturnListener getSynchroListener() {
    return new TerminalUserSynchroFODBReturnListener();
  }

  public String getCollectionName() {
    return "terminaluser";
  }

  protected abstract int getMaxTerminalUserIdLength();

  protected abstract Class getDataType();

  protected abstract void initDB(FastObjectDB db) throws FODBException, BadDataFormatException;
}
