/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  
 *  Modifications :
 *  2006 Creation P.Delrieu
 * 
 */

package org.openmobileis.modules.profiles.old;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;

import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.collection.tree.DependTree;
import org.openmobileis.common.util.exception.ServiceException;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.database.fastobjectdb.synchro.client.SynchroFastObjectDBManager;
import org.openmobileis.modules.profiles.old.embedded.core.DependProfileModule;
import org.openmobileis.modules.profiles.old.embedded.core.DependProfileRubric;
import org.openmobileis.modules.profiles.old.embedded.core.ProfileParser;
import org.openmobileis.modules.profiles.old.embedded.core.ProfileRubricSynchroListener;
import org.openmobileis.modules.profiles.old.embedded.core.ProfileSynchroListenerAddNotifier;
import org.openmobileis.services.common.ServiceManager;
import org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener;
import org.openmobileis.synchro.openmsp.client.OpenMSPSynchroManager;
import org.xmlpull.v1.XmlPullParserException;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Manage profil on the terminal side. Profile are store in a XML file.
 * The filename and path is defined by the PropertiesManager property org.openmobileis.profil.file
 * By default (ie the property is not defined), the file is stored in the directory "user.dir"/WEB-INF/conf/properties/profilData.xml.
 * @author  Philippe Delrieu
 * @since JDK 1.1
 * @version 1.0.
 */
public class ProfileManager {

  private static ProfileManager manager;

  private String profilFile;

  private Hashtable profilRubrics = null;

  private Hashtable profilModules = null;

  ProfileSynchroListenerAddNotifier profileSyncListenerNotifier = new ProfileSynchroListenerAddNotifier();

  private ProfileManager() {
    profilFile = PropertiesManager.getManager().getProperty("org.openmobileis.profil.file");
    if (profilFile == null) {
      profilFile = System.getProperty("user.dir") + File.separator + "WEB-INF" + File.separator + "conf" + File.separator + "properties" + File.separator
          + "profildata.xml";
    }
    profilRubrics = new Hashtable(20);
    profilModules = new Hashtable(10);

    profileSyncListenerNotifier = new ProfileSynchroListenerAddNotifier();
  }

  public static ProfileManager getManager() {
    if (manager == null) {
      synchronized (org.openmobileis.modules.profiles.old.ProfileManager.class) {
        if (manager == null) {
          manager = new ProfileManager();
          try {
            manager.initManager();
          } catch (Throwable ex) {
            LogManager.traceError(0, ex);
          }
        }
      }
    }
    return manager;
  }

  public void saveProfilFile(String fileData) throws ServiceException {
    try {
      FileWriter out = new FileWriter(this.profilFile);
      try {
        out.write(fileData.toCharArray());
        out.flush();
      } finally {
        out.close();
      }
      profilRubrics.clear();
      profilModules.clear();
      this.parseFile();
    } catch (Throwable ex)  {
      throw new ServiceException(ex);
    }
  }

  public Collection getProfilRubrics() {
    return profilRubrics.values();
  }

  public Collection getProfilModules() {
    return profilModules.values();
  }

  public ProfileRubric getRubric(String rubricName) {
    return (ProfileRubric) profilRubrics.get(rubricName);
  }

  public ProfileModule getModule(String moduleName) {
    return (ProfileModule) profilModules.get(moduleName);
  }

  public void addRubric(ProfileRubric rubric) {
    profilRubrics.put(rubric.getName(), rubric);
  }

  public void addModule(ProfileModule module) {
    profilModules.put(module.getName(), module);
  }

  private void initManager() throws ServiceException {
    try {
      this.parseFile();

      //init synchro db

      OpenMSPSynchroManager.getManager().registerSynchroListenerAddNotifier(profileSyncListenerNotifier);
      profileSyncListenerNotifier.setBlocking(true);
      try {
        SynchroFastObjectDBManager.getCurrentFODB(); //init synchro db to be use as default db
      } finally {
        profileSyncListenerNotifier.setBlocking(false);
      }
    } catch (Throwable ex) {
      throw new ServiceException(ex);
    }
  }

  public void loadProfile() throws ServiceException {
    try {
      //determine module loading order
      DependTree moduledepend = new DependTree();
      Enumeration enume = profilModules.elements();
      while (enume.hasMoreElements()) {
        ProfileModule module = (ProfileModule) enume.nextElement();
        Array moduleDependList = module.getDependModuleList();
        Array moduleNameDependList = new Array();
        int depsize = moduleDependList.size();
        boolean error = false;
        for (int j = 0; j < depsize; j++) {
          //verify dependance module version
          DependProfileModule dep = (DependProfileModule) moduleDependList.get(j);
          ProfileModule depModule = this.getModule(dep.getModuleName());
          if ((depModule != null) && (depModule.getVersion() == null || depModule.getVersion().equals(dep.getModuleVersion()))) {
            moduleNameDependList.add(dep.getModuleName());
          } else {
            LogManager.traceError(0, "Can't load module :" + module.getName() + " Missing depend's module:" + dep.getModuleName() + " or version:"
                + dep.getModuleVersion());
            error = true;
            break;
          }
        }
        if (error)
          continue;
        String[] smodulenamedeplist = new String[moduleNameDependList.size()];
        moduleNameDependList.toArray(smodulenamedeplist);
        moduledepend.addObject(module.getName(), module, smodulenamedeplist);
      }

      //load each module in rigth order
      moduledepend.resetPhaseList();
      Array moduleList = moduledepend.getNextPhaseObjectList();
      while (moduleList.size() != 0) {
        int modsize = moduleList.size();
        for (int i = 0; i < modsize; i++) {
          ProfileModule module = (ProfileModule) moduleList.get(i);
          try {
            Array moduledeplist = module.getDependModuleList();
            String[] smoduledeplist = new String[moduledeplist.size()];
            int depsize = moduledeplist.size();
            for (int j = 0; j < depsize; j++) {
              DependProfileModule dep = (DependProfileModule) moduledeplist.get(j);
              smoduledeplist[j] = dep.getModuleName();
            }
            this.loadModule(module, smoduledeplist);
          } catch (Throwable ex) {
            LogManager.traceError(0, "Error during rubric loading :" + module.getName() + " Exception:" + ex.getClass() + " Message :" + ex.getMessage());
            LogManager.traceError(0, ex);
          }
        }
        moduleList = moduledepend.getNextPhaseObjectList();
      }

      //load remainding synchro listener that has not been loaded by the profil.
      Enumeration syncenum = profileSyncListenerNotifier.listenerlist.elements();
      while (syncenum.hasMoreElements()) {
        OpenMSPSyncListener toloadlistener = (OpenMSPSyncListener) syncenum.nextElement();
        String[] syncdepends = (String[]) profileSyncListenerNotifier.dependlist.get(toloadlistener.getSyncName());
        OpenMSPSynchroManager.getManager().addListener(toloadlistener, syncdepends);
      }
      profileSyncListenerNotifier.listenerlist.clear();
      profileSyncListenerNotifier.dependlist.clear();

    } catch (Throwable ex) {
      LogManager.traceError(0, "Error during loading profile file :" + ex.getMessage());
      throw new ServiceException(ex);
    }

  }

  private void loadModule(ProfileModule module, String[] moduleDependList) throws ServiceException {
    Array rublist = module.getModuleRubric();
    //determine rubric loading order
    DependTree rubricDependTree = new DependTree();
    for (int k = 0; k < rublist.size(); k++) {
      DependProfileRubric rubric = (DependProfileRubric) rublist.get(k);
      rubricDependTree.addObject(rubric.getRubricName(), rubric, rubric.getRubricDepends());
    }

    //construct base listener depend list based on depend module.
    Array synclistenerdependlist = new Array();
    for (int i = 0; i < moduleDependList.length; i++) {
      ProfileModule depmodule = this.getModule(moduleDependList[i]);
      Array deprublist = depmodule.getModuleRubric();
      for (int k = 0; k < deprublist.size(); k++) {
        DependProfileRubric deprubric = (DependProfileRubric) deprublist.get(k);
        ProfileRubric rubric = this.getRubric(deprubric.getRubricName());
        Array synclistlist = rubric.getSynchroListenerList();
        for (int j = 0; j < synclistlist.size(); j++) {
          ProfileRubricSynchroListener synclistener = (ProfileRubricSynchroListener) synclistlist.get(j);
          synclistenerdependlist.add(synclistener.name);
        }
      }

    }

    //block synchro listener loading and load it with profile depend info.
    rubricDependTree.resetPhaseList();
    Array rubricList = rubricDependTree.getNextPhaseObjectList();
    while (rubricList.size() != 0) {
      int rubsize = rubricList.size();
      for (int i = 0; i < rubsize; i++) {
        DependProfileRubric deprubric = (DependProfileRubric) rubricList.get(i);
        ProfileRubric rubric = this.getRubric(deprubric.getRubricName());
        try {
          // add synchro listerner of depend rubric to depend from module list
          Array rubsynclistenerdependlist = (Array) synclistenerdependlist.clone();
          String rubricdependlist[] = deprubric.getRubricDepends();
          if (rubricdependlist != null) {
            for (int k = 0; k < rubricdependlist.length; k++) {
              ProfileRubric deprub = this.getRubric(rubricdependlist[k]);
              Array synclistlist = deprub.getSynchroListenerList();
              for (int j = 0; j < synclistlist.size(); j++) {
                ProfileRubricSynchroListener synclistener = (ProfileRubricSynchroListener) synclistlist.get(j);
                rubsynclistenerdependlist.add(synclistener.name);
              }
            }
          }

          //load rubric as in ServiceManagerand block listener loading
          profileSyncListenerNotifier.setBlocking(true);
          try {
            ServiceManager.getManager().loadRubricLoader(rubric.getLoader());
          } finally {
            profileSyncListenerNotifier.setBlocking(false);
          }

          //load synchro listener
          Array synchroListenerList = rubric.getSynchroListenerList();
          //contruct rubric synchro lister loading order list
          DependTree syncDependTree = new DependTree();
          for (int k = 0; k < synchroListenerList.size(); k++) {
            ProfileRubricSynchroListener synclistener = (ProfileRubricSynchroListener) synchroListenerList.get(k);
            syncDependTree.addObject(synclistener.name, synclistener, synclistener.dependsList);
          }

          //load rubric listener with its depend list. Depend list = module sync list + rubric depend sync list + sync depend list
          syncDependTree.resetPhaseList();
          Array syncList = syncDependTree.getNextPhaseObjectList();
          while (syncList.size() != 0) {
            int syncsize = syncList.size();
            for (int j = 0; j < syncsize; j++) {
              ProfileRubricSynchroListener listener = (ProfileRubricSynchroListener) syncList.get(j);
              //add sync listener depend list.
              Array syncrubsynclistenerdependlist = (Array) rubsynclistenerdependlist.clone();
              String[] depends = listener.dependsList;
              if (depends != null) {
                for (int k = 0; k < depends.length; k++) {
                  syncrubsynclistenerdependlist.add(depends[k]);
                }
              }
              //load sync listener.
              String[] cursyncdeplist = new String[syncrubsynclistenerdependlist.size()];
              syncrubsynclistenerdependlist.toArray(cursyncdeplist);
              String name = listener.name;
              OpenMSPSyncListener toloadlistener = (OpenMSPSyncListener) profileSyncListenerNotifier.listenerlist.get(name);
              if (toloadlistener != null) {
                OpenMSPSynchroManager.getManager().addListener(toloadlistener, cursyncdeplist);
                profileSyncListenerNotifier.listenerlist.remove(toloadlistener.getSyncName());
                profileSyncListenerNotifier.dependlist.remove(toloadlistener.getSyncName());
              } else {
                LogManager.traceError(0, "Error during rubric loading " + rubric.getName() + ". Synchro listener not found :" + name + ". Listener not loaded");
              }
            }
            syncList = syncDependTree.getNextPhaseObjectList();

          }

        } catch (Throwable ex) {
          LogManager.traceError(0, "Error during rubric loading :" + rubric.getName() + " Exception:" + ex.getClass() + " Message :" + ex.getMessage());
          LogManager.traceError(0, ex);
        }
      }
      rubricList = rubricDependTree.getNextPhaseObjectList();
    }
  }

  private void parseFile() throws IOException, FileNotFoundException, XmlPullParserException {
    FileReader reader = new FileReader(this.profilFile);
    ProfileParser parser = new ProfileParser(reader);
    parser.parse();
  }

  public ProfileSynchroListenerAddNotifier getProfileSyncListenerNotifier() {
    return profileSyncListenerNotifier;
  }

  public void setProfileSyncListenerNotifier(ProfileSynchroListenerAddNotifier profileSyncListenerNotifier) {
    this.profileSyncListenerNotifier = profileSyncListenerNotifier;
  }
}
