/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  
 *  Modifications :
 *  2006 Creation P.Delrieu
 * 
 */
package org.openmobileis.modules.profiles.old.embedded.core;

import java.io.IOException;
import java.io.Reader;
import java.util.StringTokenizer;

import org.kxml2.io.KXmlParser;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.common.util.log.LogServices;
import org.openmobileis.modules.profiles.old.ProfileManager;
import org.openmobileis.modules.profiles.old.ProfileModule;
import org.openmobileis.modules.profiles.old.ProfileRubric;
import org.openmobileis.modules.profiles.terminal.RubricLoader;
import org.xmlpull.v1.XmlPullParserException;

/**
 * Title: OpenMobileIS project source <BR>
 * Description: Profile file parser. Extract module and Rubric from the file.
 * 
 * @author Philippe Delrieu
 * @since JDK 1.1
 * @version 1.0.
 */
public class ProfileParser {

  private final static String TAG_XML = "xml";

  private final static String TAG_PROFIL = "profil";

  private final static String TAG_RUBRIC = "rubric";

  private final static String TAG_PROPERTY = "property";

  private final static String TAG_RUBRIC_LOADER = "loader";

  private final static String TAG_RUBRIC_SYNCHROLISTENER = "synchrolistener";

  private final static String TAG_MODULE = "module";

  private final static String TAG_MODRUBRIC = "modrubric";

  private final static String TAG_MODDEPEND = "moddepend";

  private Reader reader;

  private ProfileRubric currentRubric = null;

  private ProfileModule currentModule = null;

  private boolean propertyTagOpened = false;

  private String propertyName = null;

  private String propertyValue = null;

//  private Array rubrics = new Array(5);

//  private Array modules = new Array(5);

  String currentTag;

  public ProfileParser(Reader reader) {
    this.reader = reader;
  }

  public void parse() throws XmlPullParserException, IOException {
    KXmlParser parser = new KXmlParser();
    parser.setInput(this.reader);

    int eventType = parser.getEventType();
    do {
      if (eventType == KXmlParser.START_DOCUMENT) {
      } else if (eventType == KXmlParser.END_DOCUMENT) {
      } else if (eventType == KXmlParser.START_TAG) {
        this.processStartElement(parser);
      } else if (eventType == KXmlParser.END_TAG) {
        this.processEndElement(parser);
      } else if (eventType == KXmlParser.TEXT) {
        this.processText(parser);
      }
      eventType = parser.next();
    } while (eventType != KXmlParser.END_DOCUMENT);
  }

  private void openRubricTag(KXmlParser parser) {
    if (currentRubric != null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, starting a new rubric while the preceding one isn't closed");
      return;
    }
    if (currentModule != null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, starting a new rubric while the preceding module isn't closed");
      return;
    }
    currentRubric = new ProfileRubric();
    int count = parser.getAttributeCount();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("name")) {
        currentRubric.setName(parser.getAttributeValue(i));
      } else {
        LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, unrecognized attribute " + parser.getAttributeName(i) + " for rubric");
      }
    }
  }

  private void closeRubricTag() {
    if (currentRubric == null) {
      LogManager.traceError(LogServices.SYNCHROSERVICE, "Corrupted profil file, found rubric closing tag while no rubric is open");
      return;
    }
    ProfileManager.getManager().addRubric(currentRubric);
    currentRubric = null;
  }

  private void openModuleTag(KXmlParser parser) {
    if (currentRubric != null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, starting a new module while the preceding rubric isn't closed");
      currentRubric = null;
    }
    if (currentModule != null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, starting a new module while the preceding one isn't closed");
      return;
    }
    currentModule = new ProfileModule();
    int count = parser.getAttributeCount();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("name")) {
        currentModule.setName(parser.getAttributeValue(i));
      } else if (parser.getAttributeName(i).equals("version")) {
        currentModule.setVersion(parser.getAttributeValue(i));
      } else {
        LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, unrecognized attribute " + parser.getAttributeName(i) + " for modrubric");
      }
    }
  }

  private void closeModuleTag() {
    if (currentModule == null) {
      LogManager.traceError(LogServices.SYNCHROSERVICE, "Corrupted profil file, found module closing tag while no module is open");
      return;
    }
    ProfileManager.getManager().addModule(currentModule);
    currentModule = null;
  }

  private void openPropertyTag(KXmlParser parser) {
    if (currentRubric != null && propertyTagOpened == true) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, opening a new rubric property while the preceding one isn't closed");
      return;
    } else if (currentModule != null && propertyTagOpened == true) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, opening a new rubric property while the preceding one isn't closed");
      return;
    } else if (currentModule == null && currentRubric == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, argument property must be included in a rubric or in a module");
      return;
    }

    propertyTagOpened = true;
    propertyName = null;
    propertyValue = null;

    int count = parser.getAttributeCount();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("name")) {
        propertyName = parser.getAttributeValue(i);
      } else if (parser.getAttributeName(i).equals("value")) {
        propertyValue = parser.getAttributeValue(i);
      } else {
        LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, unrecognized attribute " + parser.getAttributeName(i) + " for modrubric");
      }
    }

    if (propertyName == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, property tag with no name !");
    }
    if (propertyValue == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, property tag " + propertyName + " with no value !");
    }
  }

  private void closePropertyTag() {
    if (propertyTagOpened == false) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, closing tag property while it's not opened");
    } else if (propertyName == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, closing tag property while it has no name defined");
    } else if (propertyValue == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, closing tag property " + propertyName + " while it has no value defined");
    } else {
      if (currentRubric != null) {
        currentRubric.addProperty(propertyName, propertyValue);
      } else if (currentModule != null) {
        currentModule.addProperty(propertyName, propertyValue);
      } else {
        LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, property " + propertyName + " outside rubrics and modules");
      }
    }
    propertyTagOpened = false;
    propertyName = null;
    propertyValue = null;
  }


  private void openLoaderTag(KXmlParser parser) {
    if (currentRubric == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, loader tag found outside a rubric");
      return;
    }
   int count = parser.getAttributeCount();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("class")) {
        String classname = parser.getAttributeValue(i).trim();
        try {
          Object toload = Class.forName(classname).newInstance();
          if (toload instanceof RubricLoader) {
            currentRubric.setLoader((RubricLoader)toload) ;
          } else  {
            LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Bad profil file, rubric loader class is not a RubricLoader :"+classname);
          }      
        } catch (Throwable ex)  {
          LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Bad profil file, error during creating rubric loader instance :"+ex.getMessage());
        }
      }
    }
  }

  private void openSynchroListenerTag(KXmlParser parser) {
    if (currentRubric == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, synchrolistener tag found outside a rubric");
      return;
    }
    ProfileRubricSynchroListener synclistener = new ProfileRubricSynchroListener();
    int count = parser.getAttributeCount();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("name")) {
        synclistener.name = parser.getAttributeValue(i).trim();
      } else if (parser.getAttributeName(i).equals("depends")) {
        String dep = parser.getAttributeValue(i).trim();
        Array deplist = new Array();
        StringTokenizer token = new StringTokenizer(dep, ",");
        while (token.hasMoreElements()) {
          deplist.add(token.nextElement());
        }
        String[] strdeplist = new String[deplist.size()];
        deplist.toArray(strdeplist);
        synclistener.dependsList = strdeplist;
      }
    }
    currentRubric.addSynchroListener(synclistener);
  }

  private void openModDependTag(KXmlParser parser) {
    if (currentModule == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, moddepend tag found outside a module");
      return;
    }
    int count = parser.getAttributeCount();
    DependProfileModule depend = new DependProfileModule();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("modulename")) {
        depend.setModuleName(parser.getAttributeValue(i).trim());
      } else if (parser.getAttributeName(i).equals("version")) {
        depend.setModuleVersion(parser.getAttributeValue(i).trim());
      }
    }
    currentModule.addDependProfilModule(depend);
  }

  private void openModrubricTag(KXmlParser parser) {
    if (currentModule == null) {
      LogManager.traceWarning(LogServices.SYNCHROSERVICE, "Corrupted profil file, argument modrubric must be included in a module");
      return;
    }
 //   modrubricName = new ProfilModuleRubric();
    int count = parser.getAttributeCount();
    DependProfileRubric depend = new DependProfileRubric();
    for (int i = 0; i < count; i++) {
      if (parser.getAttributeName(i).equals("name")) {
        String name = parser.getAttributeValue(i);
        depend.setRubricName(name);
      } else if (parser.getAttributeName(i).equals("depends")) {
        String dep = parser.getAttributeValue(i).trim();
        Array deplist = new Array();
        StringTokenizer token = new StringTokenizer(dep, ",");
        while (token.hasMoreElements()) {
          deplist.add(token.nextElement());
        }
        String[] strdeplist = new String[deplist.size()];
        deplist.toArray(strdeplist);
        depend.setRubricDepends(strdeplist);
      }
    }
    currentModule.addModuleRubric(depend);
  }

  private void processStartElement(KXmlParser parser) {
    String name = parser.getName();

    if (name.equals(ProfileParser.TAG_XML)) {
      // Do nothing
    } else if (name.equals(ProfileParser.TAG_PROFIL)) {
      // Do nothing
    } else if (name.equals(ProfileParser.TAG_RUBRIC)) {
      this.openRubricTag(parser);
    } else if (name.equals(ProfileParser.TAG_MODULE)) {
      this.openModuleTag(parser);
    } else if (name.equals(ProfileParser.TAG_PROPERTY)) {
      this.openPropertyTag(parser);
    } else if (name.equals(ProfileParser.TAG_MODRUBRIC)) {
      this.openModrubricTag(parser);
    } else if (name.equals(ProfileParser.TAG_RUBRIC_LOADER)) {
      this.openLoaderTag(parser);
    } else if (name.equals(ProfileParser.TAG_RUBRIC_SYNCHROLISTENER)) {
      this.openSynchroListenerTag(parser);
    } else if (name.equals(ProfileParser.TAG_MODDEPEND)) {
      this.openModDependTag(parser);
    }
  }

  private void processEndElement(KXmlParser parser) {
    String name = parser.getName();

    if (name.equals(ProfileParser.TAG_XML)) {
      // Do nothing
    } else if (name.equals(ProfileParser.TAG_PROFIL)) {
      // Do nothing
    } else if (name.equals(ProfileParser.TAG_RUBRIC)) {
      this.closeRubricTag();
    } else if (name.equals(ProfileParser.TAG_MODULE)) {
      this.closeModuleTag();
    } else if (name.equals(ProfileParser.TAG_PROPERTY)) {
      this.closePropertyTag();
    }
  }

  private void processText(KXmlParser parser) throws XmlPullParserException {
    // Do nothing
  }
}
