/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */

package org.openmobileis.services.servlet;

import java.io.*;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.openmobileis.common.intl.IntlResourceManager;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description:  Load the URL specified file and send it.
 * handle index.html if no file is specified.
 * Does not allow directory listing.
 * 
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public class FileServlet extends HttpServlet {
  static final long serialVersionUID = 5521257935120563452L;

	boolean headOnly = false;
	private String supPath = "";

	/// Constructor.
	public FileServlet() {
	}
	
    public void init() throws ServletException {
        
        // This property gives FileServlet' sup. path.
        String suppath = this.getInitParameter("htdocs");
        if (suppath != null) {
            this.supPath = suppath;
        }
    }

	/// Returns a string containing information about the author, version, and
	// copyright of the servlet.
	public String getServletInfo() {
		return "servlet similar to a standard httpd";
	}

	/// Services a single request from the client.
	// @param req the servlet request
	// @param req the servlet response
	// @exception ServletException when an exception has occurred
	public void service(HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException {
		if (req.getMethod().equalsIgnoreCase("get"))
			headOnly = false;
		else
			headOnly = true;
		//String path = req.getServletPath();
		String path = req.getPathInfo();
		if (path == null || path.charAt(0) != '/') {
			res.sendError(HttpServletResponse.SC_BAD_REQUEST);
			return;
		}
		if (path.indexOf("/../") != -1 || path.endsWith("/..")) {
			res.sendError(HttpServletResponse.SC_FORBIDDEN);
			return;
		}

		// Make a version without the leading /.
/*		String pathname = path.substring(1);
		if (pathname.length() == 0)
			pathname = "./";*/
		// Try to find the directory or the file and store in HttpServletResponse
		managePathname(req, res, headOnly, path, path);
	}

	private void managePathname(
			HttpServletRequest req,
			HttpServletResponse res,
			boolean headOnly,
			String path,
			String pathname)
			throws IOException {
		String filename = pathname;
		filename = getServletConfig().getServletContext().getRealPath(this.supPath+filename);

		//translate file to system local.
		IntlResourceManager resourceManager =
			IntlResourceManager.getManager();
		filename = resourceManager.getLocalizedFileName(filename, true);

		File file = new File(filename);
		if (file.exists()) {
			if (!file.isDirectory())
				serveFile(req, res, headOnly, path, filename, file);
			else {
				String indexFilename = null;
				if (pathname.charAt(pathname.length() - 1) == '/')
					indexFilename = filename + "index.html";
				else 
					indexFilename = filename + File.separatorChar + "index.html";
				File indexFile = new File(indexFilename);
				if (indexFile.exists())
					serveFile(req, res, headOnly, path, indexFilename, indexFile);
				else
					res.sendError(HttpServletResponse.SC_UNAUTHORIZED);
			} // end else !file.isDirectory
		} else { // else file.exists
			if (pathname.endsWith("/index.html"))
				managePathname(
					req,
					res,
					headOnly,
					path,
					pathname.substring(0, pathname.length() - 10));
			else if (pathname.equals("index.html"))
				managePathname(req, res, headOnly, path, "./");
			else
				res.sendError(HttpServletResponse.SC_NOT_FOUND);
		} // end else
	}

	private void serveFile(
			HttpServletRequest req,
			HttpServletResponse res,
			boolean headOnly,
			String path,
			String filename,
			File file)
			throws IOException {
		try {
			if (!file.canRead()) {
				res.sendError(HttpServletResponse.SC_FORBIDDEN);
				return;
			}

			res.setStatus(HttpServletResponse.SC_OK);
	//		long lastMod = file.lastModified();
			String mimeType = getServletConfig().getServletContext().getMimeType(filename);
			res.setContentType(mimeType);
			res.setContentLength((int) file.length());
		//	res.setDateHeader("Last-modified", lastMod);
			if (this.headOnly) // body is not required
				return;
			OutputStream out = res.getOutputStream();
			InputStream in = new FileInputStream(file);
			copyStream(in, out);
			in.close();
			out.close();
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	/// Copy the input to the output until EOF.
	private void copyStream(InputStream in, OutputStream out) throws IOException {
		byte[] buf = new byte[512];
		int len;
		while ((len = in.read(buf)) != -1) {
			out.write(buf, 0, len);
		}
	}

}
