/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */

package org.openmobileis.services.servlet;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Properties;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.openmobileis.common.context.ApplicationContextManager;
import org.openmobileis.common.context.SessionContext;
import org.openmobileis.common.context.SessionContextManager;
import org.openmobileis.common.intl.IntlResourceManager;
import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.common.util.log.FileLogManager;
import org.openmobileis.common.util.log.FileOpenCloseLogManager;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.common.util.log.LogServices;
import org.openmobileis.embedded.webserver.ServeConfig;
import org.openmobileis.embedded.webserver.WebServerConnection;
import org.openmobileis.services.Service;
import org.openmobileis.services.common.CallingServiceManager;
import org.openmobileis.services.common.ServiceManager;

public class OpenMISServlet extends HttpServlet {
  static final long serialVersionUID = 5521257935120563452L;
    
    //inited to avoid class garbage collection.
    private static ApplicationContextManager applicationContextManager;
 
    /*
     * Start service from GET method, same behavior as POST @param req the
     * servlet request @param req the servlet response @exception
     * ServletException when an exception has occurred
     */
    
    public OpenMISServlet() {
      applicationContextManager = ApplicationContextManager.getManager();
      applicationContextManager.getClass(); // to be read.
    }
    
    public void init() throws ServletException {
        
        // This property gives OpenMIS' install path.
        String installpath = this.getInitParameter("org.openmis.services.installpath");
        if (installpath==null) {
            installpath="";
        }
        String userdir = this.getServletContext().getRealPath(installpath);
        File file = new File(userdir);
        if (!file.isDirectory()) {
            throw new ServletException("Invalid installation path : "+userdir+". Can't Initialize application !!!");
        }
		System.setProperty("user.dir", userdir);
        
        // This property gives OpenMIS' property file.
        // It can be relative from the install path defined before, or absolute.
        String propsFile = this.getInitParameter("org.openmis.services.conffile");
        if (propsFile == null || propsFile.trim().length() == 0) {
            throw new ServletException("Servlet didn't find 'org.openmis.services.conffile' init parameter. Can't Initialize application !!!");
        }
        
        initOpenMIS(propsFile);
    }
    
    private void initOpenMIS(String propertiesFileName) throws ServletException {
        OpenMISInit openmisInit = null;



        // The property filename can be relative to user.dir or absolute. We test both.
        String absolutePath = System.getProperty("user.dir") + propertiesFileName;
        File confFile = new File(absolutePath);
        
        if ((!confFile.exists()) || (!confFile.isFile())) {
            
            confFile = new File(propertiesFileName);
            
            if ((!confFile.exists()) || (!confFile.isFile())) {
                throw new ServletException("Servlet didn't find property file. Didn't find "+absolutePath+" neither "+propertiesFileName+". Can't Initialize application !!!");
            }
            absolutePath = propertiesFileName;
        }

        try {
    	    System.out.println("Webserver start property file : " + absolutePath);
    	    
    	    PropertiesManager.getManager().addPropertiesFileFromFilePath(absolutePath);
    	    

 /*   	    String logsFile = PropertiesManager.getManager().getProperty("org.openmobileis.common.log.file");
            // log either on console or in a log file defined in the property
            // file
            if (logsFile == null)
                LogManager.registerLogManager(null);
            else {
                logsFile = org.openmobileis.common.util.file.FileUtilities.convertFileNameToSystem(logsFile);
                FileOpenCloseLogManager.registerLogManager(PropertiesManager.getManager().getProperties());
            }  */
    	    
    	    // First of all, we try to identify if an OpenMISInit class has been declared.
    	    // If yes, we run its preLoadingInit method first.
    	    // Then, its postLoadingInit when all initialization done.
    	    String openmisInitClass = PropertiesManager.getManager().getProperty("org.openmis.services.initclass");
    	    if (openmisInitClass != null && openmisInitClass.length() > 0) {
    	        Object init = Class.forName( openmisInitClass ).newInstance();
    	        if (init instanceof OpenMISInit) {
    	            openmisInit = (OpenMISInit)init;
    	            openmisInit.preLoadingInit();
    	        } else {
    	            LogManager.traceError(LogServices.WEBSERVICE, openmisInitClass+" isn't an instance of OpenMISInit !!!");
    	        }
    	    }

        } catch (Exception ex) {
            LogManager.traceError(0, ex);
            throw new ServletException(ex);
        }

        //init application context
        applicationContextManager = ApplicationContextManager.getManager();
        
        String listServices = PropertiesManager.getManager().getProperty("listServicesFile");
        if (listServices==null) {
            throw new ServletException("Servlet didn't find property listServicesFile. Can not load services.");
        }
        this.loadServices(listServices);
        
        if (openmisInit != null) {
            try {
                openmisInit.postLoadingInit();
            } catch (Exception ex) {
                throw new ServletException("error during init server properties and log", ex);
            }
        }
    }
    
    private void loadServices(String listServicesFile) {
        java.io.File file = new java.io.File(listServicesFile);
        if (!file.exists())
            LogManager.traceAlert(LogServices.WEBSERVICE,
                      "OpenMIS service list file not found. " + listServicesFile);
        else {
            Properties props = new Properties();
            FileInputStream fileservice = null;
            try {
                fileservice = new java.io.FileInputStream(file);
                props.load(fileservice);
                ServiceManager.getManager().init(props, "/"+this.getServletName());
            } catch (Exception ex) {
                LogManager.traceAlert(LogServices.WEBSERVICE, "Error loading services");
                LogManager.traceAlert(LogServices.WEBSERVICE, ex);
            } finally {
                if (fileservice != null) {
                    try {
                        fileservice.close();
                    } catch (Exception ex) {
                        LogManager.traceAlert(LogServices.WEBSERVICE, "Unable to close file : "+listServicesFile);
                    }
                }
            }
        }
    }
    
    public void doGet(HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException {
        this.processRequest(req, res);
    }

    /**
     * Call service from POST method
     * 
     * @param req
     *            the servlet request
     * @param res
     *            the servlet response
     * @exception ServletException
     *                when an exception has occurred
     */
    public void doPost(HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException {
        this.processRequest(req, res);
    }

    private void processRequest(HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException {
        String serviceName = req.getPathInfo();
        if (serviceName == null) {
            res.sendError(HttpServletResponse.SC_BAD_REQUEST);
        }
         Service service = ServiceManager.getManager().getServiceByURI(serviceName);
        if (service == null) {
        
            LogManager.traceNotice(0, "ServiceError Service not found " + serviceName);
            IntlResourceManager resourceManager = IntlResourceManager.getManager();
            String title = resourceManager.getLocalizedProperty("ServiceManagerServlet.UnknownService");
            res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, title + serviceName);
         
        } else {
            if (this.isAuthorized(service)) {
	          	  // manage session context
	          this.getValideSessionContext(req.getSession(true).getId());
            
            try	{
	              service.runService(new OpenmisHttpServletRequest(req), res);
              } finally {
                SessionContextManager.getManager().leaveSessionContext();
              }

          } else    {// Not allow to execute the service
    			 LogManager.traceNotice(0, "ServiceError unauthorized service error "+serviceName);
           IntlResourceManager resourceManager = IntlResourceManager.getManager();
            String title = resourceManager.getLocalizedProperty("ServiceManagerServlet.AccessDeny");
            res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, title + service.getServiceUri());
          }
        }
    }
    /**
     * Return wether or not the user is authroized to launch the service
     * NOT IMPLMENTED in V1.0 (always true)
     */
     private boolean isAuthorized (Service service) {
      return true;
      }

      /**
       *  validate or init the context for this session
       * Only one session per server.
       */
      protected SessionContext getValideSessionContext(String contextID) {
        SessionContext session = SessionContextManager.getManager().getSessionContext(contextID);
        if (session == null)  {
          synchronized (SessionContextManager.getManager()) {
            session = SessionContextManager.getManager().getSessionContext(contextID);
            if (session == null)  {
              session = SessionContextManager.getManager().createSessionContext(contextID);
            }
          }
        }
        SessionContextManager.getManager().joinSessionContext(session.getId());
        return session;
      }

}