/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */

package org.openmobileis.synchro.journal;

import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.log.*;
import org.openmobileis.synchro.openmsp.protocol.Status;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description:  Data manager for all journal entry. Use to store and get journal entry.
 * Entry are persistent.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 * @see DefaultLogRenderer, JournalService, JournalEntry
 */
public class JournalManager {
	private static final String INTMOKMESSAGE="OK";
	private static final String INTMNOKMESSAGE="echec";
   

  public static String STATUS_SYNC_IN_PROGRESS="en cours";
  public static String STATUS_OK = "OK";
  public static String STATUS_NO_CONTENT = "pas de contenu";
  public static String STATUS_PARTIAL_CONTENT = "contenu partiel";
  public static String STATUS_WRONG_FORMAT = "erreur:format de fichier incorrecte";
  public static String STATUS_UNAUTHORIZED = "erreur d'authentification";
  public static String STATUS_NOT_FOUND = "erreur : service non touv.";
  public static String STATUS_AUTHENTIFICATION_REQUIRED = "erreur : authorisation requise.";
  public static String STATUS_TOO_LARGE = "erreur : mmoire insuffisante.";
  public static String STATUS_UNSUPPORTED_MEDIA = "erreur : format de fichier non support";
  public static String STATUS_EXIST_TARGET = "erreur de protocol";
  public static String STATUS_DEVICE_FULL = "erreur : mmoire insuffisante.";
  public static String STATUS_FAILED = "erreur durant la synchronisation";
  public static String STATUS_COMMAND_NOT_IMPLEMENTED = "erreur : commande non supporte.";
  public static String STATUS_SYNCHRO_CONFLIC = "erreur : conflit de synchronisation.";
  public static String STATUS_UNKNOWN = "status non reconnu.";
	
	public static String OK_MESSAGE;
	public static String NOK_MESSAGE;
	
  private static JournalManager manager;
  private JournalQueryManager queryManager;
  private Array journalRendererList;

  private JournalManager() {
  	OK_MESSAGE = INTMOKMESSAGE;
  	NOK_MESSAGE = INTMNOKMESSAGE;
    journalRendererList = new Array(10);
    queryManager = new JournalQueryManager();
    try {
      queryManager.createJournalTable();
    } catch (Exception e) {
      LogManager.traceError(LogServices.WEBSERVICE, e);
    }
  }

  public static JournalManager getManager() {
    if (manager == null)  {
      synchronized(JournalManager.class)  {
        if (manager == null)  {
          manager = new JournalManager();
        }
      }
    }
    return manager;
  }

  public void registerJournalLogRenderer(JournalLogRenderer renderer) {
    journalRendererList.add(renderer);
  }

  public Array getRendererList() {
    return journalRendererList;
  }

  public void saveJournalEntry(String serviceName, short id, String data, int status)  {
    try {
      // check data length
      if (data.length() > 200) {
        data = data.substring(0, 200);
      }
      JournalEntry entry = new JournalEntry(serviceName, id, data, status, org.openmobileis.embedded.util.SystemAPI.currentTimeInMillis());
      queryManager.createOrUpdate(entry);
    } catch (Exception e) {
      LogManager.traceError(LogServices.WEBSERVICE, e);
    }
  }


  public JournalEntry[] getAllJournalEntryForService(String serviceName)  {
    try {
      org.openmobileis.common.util.collection.Array result = queryManager.getAllEntryForService(serviceName);
      JournalEntry[] entries = new JournalEntry[result.size()];
      result.toArray(entries);
      return entries;
    } catch (Exception e) {
      LogManager.traceError(LogServices.WEBSERVICE, e);
    }
    return null;
  }


  public void deleteAllJournalEntryForService(String serviceName)  {
    try {
      queryManager.deleteAllEntryForService(serviceName);
    } catch (Exception e) {
      LogManager.traceError(LogServices.WEBSERVICE, e);
    }
  }

  /*
  * Return error message from a status
  */
  public String getErrorMessageForSynchroStatus (int status) {
    String ret = JournalManager.STATUS_UNKNOWN;
    switch (status) {
    case Status.STATUS_SYNC_IN_PROGRESS:
      ret = JournalManager.STATUS_SYNC_IN_PROGRESS;
      break;
    case Status.STATUS_OK:
    ret = JournalManager.STATUS_OK;
    break;
    case Status.STATUS_NO_CONTENT:
    ret = JournalManager.STATUS_NO_CONTENT;
    break;
    case Status.STATUS_PARTIAL_CONTENT:
    ret = JournalManager.STATUS_PARTIAL_CONTENT;
    break;
    case Status.STATUS_WRONG_FORMAT:
    ret = JournalManager.STATUS_WRONG_FORMAT;
    break;
    case Status.STATUS_UNAUTHORIZED:
    ret = JournalManager.STATUS_UNAUTHORIZED;
    break;
   case Status.STATUS_NOT_FOUND:
   ret = JournalManager.STATUS_NOT_FOUND;
   break;
    case Status.STATUS_AUTHENTIFICATION_REQUIRED:
    ret = JournalManager.STATUS_AUTHENTIFICATION_REQUIRED;
    break;
    case Status.STATUS_TOO_LARGE:
    ret = JournalManager.STATUS_TOO_LARGE;
    break;
    case Status.STATUS_UNSUPPORTED_MEDIA:
    ret = JournalManager.STATUS_UNSUPPORTED_MEDIA;
    break;
    case Status.STATUS_EXIST_TARGET:
    ret = JournalManager.STATUS_EXIST_TARGET;
    break;
    case Status.STATUS_DEVICE_FULL:
    ret = JournalManager.STATUS_DEVICE_FULL;
    break;
    case Status.STATUS_FAILED:
    ret = JournalManager.STATUS_FAILED;
    break;
    case Status.STATUS_COMMAND_NOT_IMPLEMENTED:
    ret = JournalManager.STATUS_COMMAND_NOT_IMPLEMENTED;
    break;
    case Status.STATUS_SYNCHRO_CONFLIC:
    ret = JournalManager.STATUS_SYNCHRO_CONFLIC;
    break;
    }
    return ret;
  }

}