/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */

package org.openmobileis.synchro.openmsp.client;

import org.openmobileis.synchro.openmsp.OpenMSPException;
import org.openmobileis.synchro.openmsp.protocol.*;
import org.openmobileis.synchro.security.auth.*;
import org.openmobileis.common.intl.IntlResourceManager;
import org.openmobileis.common.util.PersistentPropertiesManager;
import org.openmobileis.common.util.codec.GeneralCoder;
import org.openmobileis.synchro.client.SynchroDescriptor;
import org.openmobileis.synchro.journal.*;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description:  Default Synchro listener that handle OpenMS protocol for synchronisation.
 * Provide helper classes to facilitate the use of OpenMS protocol.
 * The synchroStatus is made persistent by using the ServicePropertiesManager.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public abstract class DefaultOpenMSPSyncListener implements OpenMSPSyncListener {
  protected static String ADD_ERROR_MESSAGE;
  protected static String UPDATE_ERROR_MESSAGE;
  protected static String DELETE_ERROR_MESSAGE;
  protected static String ITEM_ERROR_MESSAGE;
  protected static String SYNC_ERROR_MESSAGE;
  protected static String SYNC_UNAUTHORIZED_MESSAGE;

 // private boolean beginned;

  /**
   * status of last synchro
   */
  private int synchroStatus;

  public boolean isSyncOK() {
    return (synchroStatus == Status.STATUS_OK);
  }

  public DefaultOpenMSPSyncListener() {
  }
  
	public void initListener() throws OpenMSPException	{
    IntlResourceManager resourceManager = IntlResourceManager.getManager();
    SYNC_UNAUTHORIZED_MESSAGE = resourceManager.getLocalizedProperty("OpenMSPDbSyncListener.SYNC_UNAUTHORIZED_MESSAGE", "Erreur accs non authoris, mauvaise authentification.");
    SYNC_ERROR_MESSAGE = resourceManager.getLocalizedProperty("OpenMSPDbSyncListener.SYNC_ERROR_MESSAGE", "Erreur gnrale de la synchronisation sur le serveur distant.");
    ITEM_ERROR_MESSAGE = resourceManager.getLocalizedProperty("OpenMSPDbSyncListener.ITEM_ERROR_MESSAGE", "Erreur pendant la rcupration des donnes du serveur distant.");
    DELETE_ERROR_MESSAGE = resourceManager.getLocalizedProperty("OpenMSPDbSyncListener.DELETE_ERROR_MESSAGE", "Erreur pendant la suppression d'une donne sur le serveur distant.");
    UPDATE_ERROR_MESSAGE = resourceManager.getLocalizedProperty("OpenMSPDbSyncListener.UPDATE_ERROR_MESSAGE", "Erreur pendant la mise  jour d'une donne sur le serveur distant.");
    ADD_ERROR_MESSAGE = resourceManager.getLocalizedProperty("OpenMSPDbSyncListener.ADD_ERROR_MESSAGE", "Erreur pendant l'ajout d'une donne sur le serveur distant.");
    String status = PersistentPropertiesManager.getManager().getProperty(this.getSyncName(), "synchroStatus");
    if (status == null) {
      synchroStatus = Status.STATUS_OK;
    }  else {
      synchroStatus = Integer.parseInt(status);
    }
    JournalLogRenderer renderer = this.getJournalLogRenderer();
    if (renderer != null)  JournalManager.getManager().registerJournalLogRenderer(renderer);
  }

  /**
  * Fire this event once a synchro is starting
  */
  public void startSync(Credential cred, SynchroDescriptor synchrodescriptor) throws OpenMSPException {
   synchroStatus = Status.STATUS_OK;
   JournalManager.getManager().deleteAllJournalEntryForService(this.getSyncName());
  }

  public void endSync()  throws OpenMSPException {
    String status  = Integer.toString(synchroStatus);
    PersistentPropertiesManager.getManager().saveProperty(this.getSyncName(), "synchroStatus", status);
   // log journal only if a synchro has been processed
   if (this.isSyncOK()) {
       JournalManager.getManager().saveJournalEntry(this.getSyncName(), JournalEntry.ERROR_SYNC, JournalManager.OK_MESSAGE, Status.STATUS_OK);
   }
 }

  /**
  * Return the synchro status of the listener. The return value is defined by the OpenMSP Status class.
  * @see org.openmobileis.synchro.openmsp.protocol.Status
  */
  public int getSynchroStatus() {
    return synchroStatus;
  }

  public void notifySynchroFailure() {
    this.setSynchroStatus(Status.STATUS_FAILED);
    JournalManager.getManager().saveJournalEntry(

      this.getSyncName()

      ,JournalEntry.ERROR_SYNC

      ,"Failure"

      , Status.STATUS_FAILED

    );
  }
  
  public boolean equals(Object obj) {
    if (
      (obj != null)
      && (obj instanceof OpenMSPSyncListener)
      && ((OpenMSPSyncListener)obj).getSyncName().equals(this.getSyncName())
    ) {
      return true;
    }
    return false;
  }

  public int hashCode() {
    return this.getSyncName().hashCode();
  }
  /**
   * Event sent when the sync manager receives Sync command openML messages from the server
   * @param ContainerMessage container : container containing the message
   * @param long newSessionID : the session id use for this synchro for this listener message.
   */
  public void receiveSyncCommand( ContainerMessage syncContainer, long newSyncNumber)  throws OpenMSPException  {
    Command sync = (Command)syncContainer.getElement();
    Credential cred = CredentialCodec.getOpenMSPCredential(sync.getCrendential());
    while (syncContainer.hasMoreMessage())  {
      ContainerMessage commandContainer = syncContainer.nextMessage();
      Element command = commandContainer.getElement();
      if (command.getElementType() == Element.ADD) {
        this.receiveAddCommand( cred, commandContainer, newSyncNumber);
      } else if (command.getElementType() == Element.DELETE) {
        this.receiveDeleteCommand( cred, commandContainer, newSyncNumber);
      } else if (command.getElementType() == Element.REPLACE) {
        this.receiveReplaceCommand( cred, commandContainer, newSyncNumber);
      } else if (command.getElementType() == Element.ITEM) {
        this.receiveSyncItem( cred, (Item)command, newSyncNumber);
      }
    }
 }

  /**
   * Utility method that add the synchro credential to a RequestCommand.
   */
  protected static void addCredentialToCommand(Credential cred, RequestCommand command) {
    if (cred == null) return;
    String currentCredential = cred.getPrincipal()+":"+cred.getPassword();
    try {
      command.setCredential("<Type>OpenMSP:auth-basic</Type><Format>b64</Format>", new String(GeneralCoder.encodeBase64(currentCredential.getBytes())));
    } catch (Throwable e) {
    }

  }

  /**
   * update the synchro status if the specified status is less than current synchro status (more serious status)
   */
  protected void setSynchroStatus(int status)  {
    if (status > synchroStatus) {
      synchroStatus = status;
      if (status >= 300){  //listener in error
        OpenMSPSynchroManager.getManager().notifyFailure(this.getSyncName());
      }
    }
  }
  
  protected JournalLogRenderer getJournalLogRenderer()  {
    return null;
  }


  /**
   * Call when the server answer SYNC command container contains a ADD item.
   * Extends this method to process ADD item.
   * @param ContainerMessage container : container containing the message
   * @param long newSessionID : the session id use for this synchro for this listener message.
   */
  protected void receiveAddCommand( Credential cred,ContainerMessage addContainer, long newSyncNumber)  throws OpenMSPException {
  }

  /**
   * Call when the server answer SYNC command container contains a SYNC item.
   * Extends this method to process SYNC item.
   * @param ContainerMessage container : container containing the message
   * @param long newSessionID : the session id use for this synchro for this listener message.
   */
  protected void receiveSyncItem( Credential cred, Item item , long newSyncNumber)  throws OpenMSPException {
  }

  /**
   * Call when the server answer SYNC command container contains a DELETE item.
   * Extends this method to process DELETE item.
   * @param ContainerMessage container : container containing the message
   * @param long newSessionID : the session id use for this synchro for this listener message.
   */
  protected void receiveDeleteCommand( Credential cred, ContainerMessage deleteContainer, long newSyncNumber)  throws OpenMSPException {
  }

  /**
   * Call when the server answer SYNC command container contains a REPLACE item.
   * Extends this method to process REPLACE item.
   * @param ContainerMessage container : container containing the message
   * @param long newSessionID : the session id use for this synchro for this listener message.
   */
  protected void receiveReplaceCommand( Credential cred, ContainerMessage replaceContainer, long newSyncNumber)  throws OpenMSPException {
  }

  /**
   * Call when the server answer SYNC command container contains a GET item.
   * Extends this method to process GET item.
   * @param ContainerMessage container : container containing the message
   * @param long newSessionID : the session id use for this synchro for this listener message.
   */
  public void receiveGetCommand( Credential cred, ContainerMessage replaceContainer, long newSyncNumber)  throws OpenMSPException {
  }
  
  /**
   * Event sent when the sync manager receives MAP command openML messages from the server
   * @param ContainerMessage container : container containing the message
   * @param long newSessionID : the session id use for this synchro for this listener message.
   */
  public void receiveMapCommand( ContainerMessage mapContainer)   throws OpenMSPException   {
  }

  /**
   * Event sent when the sync manager receives RESULT openML messages from the server
   * @param Result resultCommande : resultCommande containing the result of a command
   * @param long newSessionID : the session id use for this synchro for this listener message.
   * @param ContainerMessage : command sent that generate the status. 
   */
  public void receiveResultCommand(ContainerMessage resultContainer, ContainerMessage initialCommand)  throws OpenMSPException  {
  }
}