/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */
package org.openmobileis.synchro.openmsp.client;

import java.io.ByteArrayOutputStream;
import java.io.OutputStreamWriter;

import org.openmobileis.common.util.codec.GeneralCoder;
import org.openmobileis.common.util.exception.DatabaseException;
import org.openmobileis.common.util.exception.SynchroException;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.synchro.client.SynchroDescriptor;
import org.openmobileis.synchro.journal.JournalEntry;
import org.openmobileis.synchro.journal.JournalManager;
import org.openmobileis.synchro.openmsp.OpenMSPException;
import org.openmobileis.synchro.openmsp.client.core.NumSyncManagerDB;
import org.openmobileis.synchro.openmsp.client.db.ActionDB;
import org.openmobileis.synchro.openmsp.client.db.DBImportFileCoder;
import org.openmobileis.synchro.openmsp.client.db.ImportQueryManager;
import org.openmobileis.synchro.openmsp.client.db.SyncActionDBManager;
import org.openmobileis.synchro.openmsp.protocol.AbstractCommand;
import org.openmobileis.synchro.openmsp.protocol.Command;
import org.openmobileis.synchro.openmsp.protocol.ContainerMessage;
import org.openmobileis.synchro.openmsp.protocol.DataItem;
import org.openmobileis.synchro.openmsp.protocol.Element;
import org.openmobileis.synchro.openmsp.protocol.Item;
import org.openmobileis.synchro.openmsp.protocol.Message;
import org.openmobileis.synchro.openmsp.protocol.RequestCommand;
import org.openmobileis.synchro.openmsp.protocol.Result;
import org.openmobileis.synchro.openmsp.protocol.Status;
import org.openmobileis.synchro.security.auth.Credential;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: 
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public abstract class OpenMSPDBSyncListener extends DefaultOpenMSPSyncListener  {

  private boolean syncGoOn = false;
  protected String installPath;

  /**
   * 
   */
  public OpenMSPDBSyncListener() {
    super();
  }

  public void startSync(Credential cred, SynchroDescriptor synchrodescriptor) throws OpenMSPException {
    super.startSync(cred, synchrodescriptor);
    this.syncGoOn = true;
    installPath = synchrodescriptor.getSynchroConduit().getIntallPath();
  }

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener#sendData(org.openmobileis.synchro.openmsp.protocol.Message)
   */
  public void sendData(Message message) throws OpenMSPException {
    if (!this.isSyncOK()) {
      return; // return if an error occurs during synchro.
    }

    if (!syncGoOn) { // synchronize data only for the first call.
    return;
    }
    syncGoOn = false;
    // construct Atomic Object action list and resolve conflicts
    ActionDB[] actionArray = SyncActionDBManager.getManager().getActionsForService(this.getSyncName());
     // genererate sync message.
    try {

      // add GET command to make synchro
      Command getCommand = new Command(Element.GET, this.getSyncName(), this.getSynchroTarget());
      long ns = NumSyncManagerDB.getManager().getSyncNumberForService(this.getSyncName());
      getCommand.setSourceSessionID(ns);
      Credential cred = this.getSyncroCredential();
      if (cred != null) {
        DefaultOpenMSPSyncListener.addCredentialToCommand(cred, getCommand);
      }
      ContainerMessage getContainer = new ContainerMessage(getCommand);
      Item newItem = new DataItem(
                  Element.ITEM
                  , ""
                  , ""
                  , null
                  , null
              );
      getContainer.add(newItem);
      message.add(getContainer); // add before sync container because GET command must be proceed before the SYNC command.
      
      //add sync command
      Command syncCommand = new Command(Element.SYNC, this.getSyncName(), this.getSynchroTarget());
      syncCommand.setSourceSessionID(ns);
      String commandMetadata = this.getSendCommandMetaData();
      if (commandMetadata != null)  {
        syncCommand.setMetaInformation(commandMetadata);
      }
      if (cred != null) {
        DefaultOpenMSPSyncListener.addCredentialToCommand(cred, syncCommand);
      }
      ContainerMessage syncContainer = new ContainerMessage(syncCommand);

      newItem = null;
      byte[] actionData = null;
      RequestCommand newCommand = null;
      ContainerMessage newContainer = null;
      for (int i=0; i<actionArray.length; i++) {
        if (actionArray[i].getActionType() == ActionDB.ADD_ACTION) { // ADD
          actionData = this.getSyncDataForAtomicObjectUID(actionArray[i].getActionObjectUID(), ActionDB.ADD_ACTION);
          if (actionData != null) {
            newCommand = new RequestCommand(Element.ADD);
            newCommand.setMetaInformation(new String(GeneralCoder.encodeBase64(actionArray[i].getActionObjectUID().getBytes())));
            newContainer = new ContainerMessage(newCommand);
            newItem = new DataItem(
                        Element.ITEM
                        , new String(GeneralCoder.encodeBase64(("AOUID:"+actionArray[i].getActionObjectUID()).getBytes()))
                        , new String(GeneralCoder.encodeBase64(actionData))
                        , null
                        , null
                    );
            newContainer.add(newItem);
            syncContainer.add(newContainer);
          }
        } else if (actionArray[i].getActionType() ==ActionDB.UPDATE_ACTION)  { // REPLACE
          actionData = this.getSyncDataForAtomicObjectUID(actionArray[i].getActionObjectUID(), ActionDB.UPDATE_ACTION);
          if (actionData != null) {
            String serverID = this.getServerUIDForAOUID(actionArray[i].getActionObjectUID());
            if (serverID != null) {
              newCommand = new RequestCommand(Element.REPLACE);
              newCommand.setMetaInformation(new String(GeneralCoder.encodeBase64(actionArray[i].getActionObjectUID().getBytes())));
              newContainer = new ContainerMessage(newCommand);
              newItem = new DataItem(
                          Element.ITEM
                          , new String(GeneralCoder.encodeBase64(("AOUID:"+serverID).getBytes()))
                          , new String(GeneralCoder.encodeBase64(actionData))
                          , null
                          , null
                      );
               newContainer.add(newItem);
              syncContainer.add(newContainer);
            }
          }
        } else if (actionArray[i].getActionType() ==ActionDB.DELETE_ACTION)  { // DELETE
          String serverID = this.getServerUIDForAOUID(actionArray[i].getActionObjectUID());
          if (serverID != null) {
            newCommand = new RequestCommand(Element.DELETE);
            newContainer = new ContainerMessage(newCommand);
            newCommand.setMetaInformation(new String(GeneralCoder.encodeBase64(actionArray[i].getActionObjectUID().getBytes())));
            newItem = new DataItem(
                        Element.ITEM
                        , new String(GeneralCoder.encodeBase64(("AOUID:"+serverID).getBytes()))
                        , ""
                        , null
                        , null
                    );
            newContainer.add(newItem);
            syncContainer.add(newContainer);
          }
        } // REMOVE action nothing done.
      }
      message.add(syncContainer);
    } catch (Throwable ex)  {
      this.setSynchroStatus(Status.STATUS_FAILED);
      LogManager.traceError(0, ex);
    }

  }

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener#receiveMapCommand(org.openmobileis.synchro.openmsp.protocol.ContainerMessage, long)
   */
  public void receiveMapCommand(ContainerMessage mapContainer) throws OpenMSPException {
    mapContainer.resetCursor();
    try {
      while (mapContainer.hasMoreMessage()) {;
        Item map = (Item)mapContainer.nextMessage().getElement();
        String PDAUID = new String(GeneralCoder.decodeBase64(map.getTarget().getBytes()));
        String serverUID = new String(GeneralCoder.decodeBase64(map.getSource().getBytes()));
        this.notifyMAPofUID(PDAUID, serverUID);
      }
    } catch (java.io.IOException ex)  {
      this.setSynchroStatus(Status.STATUS_FAILED);
      LogManager.trace(new SynchroException(ex));
    }
  }

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener#receiveResultCommand(org.openmobileis.synchro.openmsp.protocol.ContainerMessage, long)
   */
  public void receiveResultCommand(ContainerMessage resultContainer, ContainerMessage initialCommand) throws OpenMSPException {
    Result resultCommand = (Result)resultContainer.getElement(); //result
    try {
      long newSessionID = Long.parseLong(resultCommand.getMetaInformation());
      while (resultContainer.hasMoreMessage())  {
        DataItem item = (DataItem) resultContainer.nextMessage().getElement();
        String meta = item.getMetaInformation();
        String file = item.getData();
        if (meta.equals("ImportDataFile"))  {
          String importFileName = this.installPath+file;
          this.getImportQueryManager().importDataFile(importFileName);
        } else if (meta.equals("DatabaseFile"))  {
          String importFileName = this.installPath+file;
          this.importCompletSynchroDatabaseFile(importFileName);
        }
      }
      
      NumSyncManagerDB.getManager().saveSyncNumberForService(newSessionID, this.getSyncName());
    } catch (Throwable ex)  {
      throw new OpenMSPException(ex);
    }
  }

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener#receiveStatusCommand(org.openmobileis.synchro.openmsp.protocol.Status, long)
   */
  public void receiveStatusCommand(Status statusCommande, ContainerMessage initialContenaire) throws OpenMSPException {
    try {
//      int cmlRef = statusCommande.getCmdRef();
      int status = statusCommande.getStatus();
      if (initialContenaire != null)  {
        AbstractCommand command = (AbstractCommand)initialContenaire.getElement();
        this.manageReturnStatusLog(status, command.getElementType(), initialContenaire);
        if (status == Status.STATUS_OK)  {
          if ((command instanceof RequestCommand) && (command.getElementType() != Element.GET) && (command.getElementType() != Element.SYNC)) {
            String pdaID = new String(GeneralCoder.decodeBase64(((RequestCommand)command).getMetaInformation().getBytes()));
            if (pdaID != null)  {
              int actionType = ActionDB.UPDATE_ACTION;
              if (command.getElementType() == Element.DELETE) {
                actionType = ActionDB.DELETE_ACTION;
                this.removeServerUIDForAOUID(pdaID);
             }
             this.notifySyncDataDone(pdaID, actionType);
             SyncActionDBManager.getManager().deleteActionDB(this.getSyncName(), pdaID);
            }
          }
        } else if (status == Status.STATUS_NO_CONTENT) {
        	// do nothing first get return status. Not use.
        } else {
          //set sync in error
          this.notifySynchroFailure();
        }
      } 
    }catch (Exception ex) {
      this.setSynchroStatus(Status.STATUS_FAILED);
      LogManager.trace(new SynchroException(ex));      
    }
  }

  /**
   * notify when receiving a status commnand the status code, the CMLElement type of the send command and the container
   * containing the command.
   */
  protected void manageReturnStatusLog(int statusCode, int cmlType, ContainerMessage container) {
    this.setSynchroStatus(statusCode);
    if (statusCode != Status.STATUS_OK) {
      String message = null;
      short journalEntryType = 0;
      if (cmlType == Element.ADD)  {
        journalEntryType = JournalEntry.ERROR_ADD;
        message = DefaultOpenMSPSyncListener.ADD_ERROR_MESSAGE;
      } else if (cmlType == Element.REPLACE)  {
        journalEntryType = JournalEntry.ERROR_REPLACE;
        message = DefaultOpenMSPSyncListener.UPDATE_ERROR_MESSAGE;
      } else if (cmlType == Element.DELETE)  {
        journalEntryType = JournalEntry.ERROR_DELETE;
        message = DefaultOpenMSPSyncListener.DELETE_ERROR_MESSAGE;
      } else if (cmlType == Element.ITEM)  {
        journalEntryType = JournalEntry.ERROR_SYNC;
        message = DefaultOpenMSPSyncListener.ITEM_ERROR_MESSAGE;
      } else if (cmlType == Element.SYNC)  {
        journalEntryType = JournalEntry.ERROR_SYNC_MOBILE;
        if (statusCode == Status.STATUS_UNAUTHORIZED) {
          message = DefaultOpenMSPSyncListener.SYNC_UNAUTHORIZED_MESSAGE;
        } else  {
          message = DefaultOpenMSPSyncListener.SYNC_ERROR_MESSAGE;
        }
      }
      if (message != null)  {
        JournalManager.getManager().saveJournalEntry(

           this.getSyncName()

          , journalEntryType

          , message

          , statusCode

        );
      }
    }

  }

  protected void receiveSyncItem( Credential cred, Item item , long newSyncNumber) {
  }

  /**
   * return the data to add to the synchro command ADD OR REPLACE.
   * The format must be the same has the one use on dispatcher size.
   * the default implementation call getDBDataForAdd() getDBDataForReplace() repending on the action and encode the result
   */
  protected byte[] getSyncDataForAtomicObjectUID(String pdaUID, int action) throws OpenMSPException {
    byte[] aoData = null;
    char[] charData = null;
    if (action == ActionDB.ADD_ACTION)  {
      String[] ret  = this.getDBDataForAdd(pdaUID);
      if (ret != null)  {
       charData = DBImportFileCoder.getCoder().encodeData(DBImportFileCoder.ADDAction, ret);
      } else  {
        //remove action because not use
        SyncActionDBManager.getManager().deleteActionDB(this.getSyncName(), pdaUID);
      }
    } else if (action == ActionDB.UPDATE_ACTION) {
      String[] ret  = this.getDBDataForReplace(pdaUID);
      if (ret != null)  {
        charData = DBImportFileCoder.getCoder().encodeData(DBImportFileCoder.REPLACEAction, ret);
      } else  {
        //remove action because not use
        SyncActionDBManager.getManager().deleteActionDB(this.getSyncName(), pdaUID);
      }
    }
    if (charData != null) {
      ByteArrayOutputStream out = new ByteArrayOutputStream();
      OutputStreamWriter writer = new OutputStreamWriter(out);
      try {
        writer.write(charData);
        writer.flush();
        writer.close();
        aoData = out.toByteArray();
      } catch (Exception ex)  {
        LogManager.trace(new SynchroException(ex));
      }
    }
    return aoData;
  }
  
  protected String getSendCommandMetaData()  throws OpenMSPException  {
    return null;
  }

  /**
   * return the server uid for the specified pdaUID. Atomiv Object uid can be different on the server side
   * and on the pda size. This method return the server uid use on server side to reference the pda uid specified.
   * return null if not found.
   */
  protected String getServerUIDForAOUID(String pdaUID)  {
    return pdaUID;
  }
  
  /**
   * Call when a client data modification (ActionDB) has been update on the server side.
   * The actionType is defined by the class ActionDB.
   * @param pdaID : Id of the object updated on the server side.
   * @param actionType ActionDB modification type
   * @throws OpenMSPException
   */
  public void notifySyncDataDone(String pdaID, int actionType) throws OpenMSPException {
  }

  protected void notifyMAPofUID(String pdauid, String serveruid) throws OpenMSPException  {
    
  }
  
  /**
   *remove the pda uid associated server uid.
   * this method is call when a delete commande from the PDA is has been done to the server.
   * delete AO server id must be deleted to the pda too.
   */
  protected void removeServerUIDForAOUID(String pdaUID) throws OpenMSPException {
  }
  
  /**
   * Call during complete synchronisation to dump the database with server data store in
   * the specified file name.
   * @param importFile : file containing server dump data file.
   */
  protected void importCompletSynchroDatabaseFile(String importFile) throws DatabaseException {
  }

  /**
   * return the data conforme to the disptacher create AbstractQueryManager method for specified pdaUID .
   * Return null if no data found
   */
  protected abstract String[] getDBDataForAdd(String pdaUID) throws OpenMSPException;
  /**
   * return the data conforme to the disptacher update AbstractQueryManager method for specified pdaUID .
   * Return null if no data found
   */
  protected abstract String[] getDBDataForReplace(String pdaUID) throws OpenMSPException;

  /**
   * return the target of the SYNC command use to synchronize listener data.
   */
  protected abstract String getSynchroTarget();

  /**
   * return the credential to use for data synchronization.
   * If no need return null.
   */
  protected abstract Credential getSyncroCredential();

  /**
   * return the ImportQueryManager use to make synchro data import.
   */
  protected abstract ImportQueryManager getImportQueryManager();

}
