/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */

package org.openmobileis.synchro.openmsp.client;

import org.openmobileis.common.util.exception.ServiceException;
import org.openmobileis.synchro.client.SynchroDescriptor;
import org.openmobileis.synchro.openmsp.OpenMSPException;
import org.openmobileis.synchro.openmsp.protocol.*;
import org.openmobileis.synchro.security.auth.Credential;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Base classe that all processus involved in the synchronisation must implement.
 * Synchro listener are managed by the SynchroManager.
 * The sendData method is call for sending synchronisation data and receiveData is call if there is an answer to
 * send data. 
 * Synchronisation end when no data is added to the send data container
 * For each listener, the synchronisation protocol manage an unique id that identity the synchronisation session;
 * see OpenMobile Synchronisation Protocol.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public interface OpenMSPSyncListener {
  public static final int SYNCHRO_OK=Status.STATUS_OK;
  public static final int SYNCHRO_FAILLED=Status.STATUS_FAILED;

 /**
 * Event sent when the sync manager sends a openML message to the server
 * The specified message is the message use to construct the synchro message
 * This method fill the message with listener commande container.
 */
 public void sendData(Message message) throws OpenMSPException ;

/**
* Return the name used to identiy the of the listener in the synchro processus.
* This name correspond to the source in the OpenMS protocol.
*/
 public String getSyncName();

/**
* The listener is called if the syncML command processing failed
*/
public void notifySynchroFailure();

/**
* Return the synchro status of the listener. The return value is defined by the OpenMSP Status class.
* @see org.openmobileis.synchro.openmsp.protocol.Status
*/
public int getSynchroStatus();

/**
* return true if the synchronisation is OK for the listener.
*/
public boolean isSyncOK();

/**
* Fire this event once a synchro is starting
*/
public void startSync(Credential cred, SynchroDescriptor synchrodescriptor) throws OpenMSPException ;

/**
* Fire this event once the current synchro is ended
*/
public void endSync() throws OpenMSPException ;

/**
 * Event sent when the sync manager receives Sync command openML messages from the server
 * @param ContainerMessage container : container containing the message
 * @param long newSessionID : the session id use for this synchro for this listener message.
 */
public void receiveSyncCommand( ContainerMessage syncContainer, long newSyncNumber)  throws OpenMSPException;

/**
 * Event sent when the sync manager receives MAP command openML messages from the server
 * @param ContainerMessage container : container containing the message
 * @param long newSessionID : the session id use for this synchro for this listener message.
 */
public void receiveMapCommand( ContainerMessage mapContainer)   throws OpenMSPException;

/**
 * Event sent when the sync manager receives RESULT openML messages from the server
 * @param Result resultCommande : resultCommande containing the result of a command
 * @param long newSessionID : the session id use for this synchro for this listener message.
 * @param ContainerMessage : command sent that generate the status. 
 */
public void receiveResultCommand(ContainerMessage resultContainer, ContainerMessage initialCommand)  throws OpenMSPException;

/**
 * Event sent when the sync manager receives STATUS openML messages from the server
 * @param Status statusCommande : resultCommande containing the Status of a command
 * @param long newSessionID : the session id use for this synchro for this listener message.
 * @param ContainerMessage : command sent that generate the status. 
 */
public void receiveStatusCommand( Status statusCommande, ContainerMessage initialCommand)  throws OpenMSPException;

/**
 * Call when the server answer container contains a GET command.
 * Extends this method to process GET command.
 * @param ContainerMessage container : container containing the GET message
 * @param long newSyncNumber : the synchro number use for this synchro for this listener message.
* @param Credential cred : the credential is any added in the GET command.
 */
public void receiveGetCommand( Credential cred, ContainerMessage replaceContainer, long newSyncNumber)  throws OpenMSPException;

/**
 * Call during listener registering. The init occurs before the synchronisation.
 * If an exception is throws the listener is not registered.
 * @throws ServiceException
 */
public void initListener() throws OpenMSPException;

}
