/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */
package org.openmobileis.synchro.openmsp.client.core;

import java.util.Hashtable;

import org.openmobileis.common.util.codec.GeneralCoder;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.common.util.log.LogServices;
import org.openmobileis.synchro.client.SynchroDescriptor;
import org.openmobileis.synchro.openmsp.OpenMSPException;
import org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener;
import org.openmobileis.synchro.openmsp.protocol.AbstractCommand;
import org.openmobileis.synchro.openmsp.protocol.Command;
import org.openmobileis.synchro.openmsp.protocol.ContainerMessage;
import org.openmobileis.synchro.openmsp.protocol.Element;
import org.openmobileis.synchro.openmsp.protocol.Header;
import org.openmobileis.synchro.openmsp.protocol.Message;
import org.openmobileis.synchro.openmsp.protocol.MessageFactory;
import org.openmobileis.synchro.openmsp.protocol.RequestCommand;
import org.openmobileis.synchro.openmsp.protocol.Result;
import org.openmobileis.synchro.openmsp.protocol.Status;
import org.openmobileis.synchro.security.auth.Credential;
import org.openmobileis.synchro.security.auth.CredentialCodec;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: create OpenMSP message for synchronisation. Manage send an recieve message
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class OpenMSPSyncMessageFactory {
  public static final String OpenMSPTARGET = "OpenMSPService";
  public static final String OpenMSPSOURCE = "OpenMSPSource";
  
  private static final String sessionIDServiceName = "OpenMSPSynchroManager";

  private Credential credential;
  private long sessionID = 0;

  /**
   * container the association between synchro message Cmd Commande source and Cml ids
   */
  private Hashtable synchroCmdIDServiceMapTable;

  /**
   * container the association between synchro message Cmd Commande and Cml ids
   */
  private Hashtable synchroCmdIDCommandMapTable;

  /**
   * 
   */
  public OpenMSPSyncMessageFactory() {
    super();
    sessionID = NumSyncManagerDB.getManager().getSyncNumberForService(sessionIDServiceName);
    synchroCmdIDServiceMapTable = new Hashtable();
    synchroCmdIDCommandMapTable = new Hashtable();
  }
  
  public void beginSynchro(Credential cred, SynchroDescriptor descriptor) throws OpenMSPException {
    this.credential = cred;
  }
  
  public void endSynchro() throws OpenMSPException {
    this.credential = null;
    synchroCmdIDCommandMapTable.clear();
    synchroCmdIDServiceMapTable.clear();
  }
  
  public Message getOpenMSPMessage(Array listenerList)  throws OpenMSPException {
   Header header = new Header(this.getSessionID(), OpenMSPSyncMessageFactory.OpenMSPSOURCE, OpenMSPSyncMessageFactory.OpenMSPTARGET);
   // add credential
   try {
     String currentCredential = this.credential.getPrincipal()+":"+this.credential.getPassword();
     header.setCredential("<Type>OpenMSP:auth-basic</Type><Format>b64</Format>", new String(GeneralCoder.encodeBase64(currentCredential.getBytes())));
   } catch (Throwable ex)  {
     LogManager.traceError(LogServices.WEBSERVICE, "Error during B64 encoding of synchro credential. Synchro abort.");
     throw new OpenMSPException(ex);
   }
   Message message = MessageFactory.getFactory().createMessage(header);
    int lsize = listenerList.size();
   for (int i=0; i<lsize; i++) {
     try {
       ((OpenMSPSyncListener)listenerList.get(i)).sendData(message);
     } catch (Throwable ex)  {
       LogManager.traceError(LogServices.WEBSERVICE, ex);
     }
   }
   if (!message.isEmpty()) {
     this.saveCommandMessageIDs(message);
   }
   return message;
   
  }

  /*
   * Dispath the server answer OpenMSP message to the appropriate syncListener.
   * To manager status properly, we create a hashtable containing for each cmdId of the sent message
   * and the associated syncName.
   */
  public void receiveOpenMSPMessage(Array synclistenerList, Message message)  throws OpenMSPException {
    long newSessionID =message.getHeader().getSessionID();
    this.setSessionID(newSessionID);
    message.resetCursor();
    while (message.hasMoreMessage()) {
      ContainerMessage container = message.nextMessage();
      Element element = container.getElement();
      container.resetCursor(); //reset cursor before processing. Avoid problems if the cursor has been modified before by other processing (use in unit test).
      try {
         if ( (element.getElementType() == Element.SYNC) ) {
          String target = ((Command)element).getTarget();
          OpenMSPSyncListener listener = (OpenMSPSyncListener)this.getListenerFromListWithName(synclistenerList, target);              
          if (listener != null) {
            listener.receiveSyncCommand(container, newSessionID);
          }
         } else  if (element.getElementType() == Element.MAP) {
           String target = ((Command)element).getTarget();
           OpenMSPSyncListener listener = (OpenMSPSyncListener)this.getListenerFromListWithName(synclistenerList, target);              
           if (listener != null) {
             listener.receiveMapCommand(container);
           }
         } else  if (element.getElementType() == Element.GET) {
           String target = ((Command)element).getTarget();
           OpenMSPSyncListener listener = (OpenMSPSyncListener)this.getListenerFromListWithName(synclistenerList, target);              
           if (listener != null) {
             Credential cred = CredentialCodec.getOpenMSPCredential(((Command)element).getCrendential());
             listener.receiveGetCommand( cred, container, newSessionID);
           }
        } else if (element.getElementType() == Element.RESULT) {
            Result result = (Result)container.getElement();
            String syncName = (String)synchroCmdIDServiceMapTable.get(new Integer(result.getCmdRef()));
            if (syncName != null) {
              OpenMSPSyncListener listener = (OpenMSPSyncListener)this.getListenerFromListWithName(synclistenerList,syncName);              
              if (listener != null) {
                ContainerMessage initialContainer = (ContainerMessage)synchroCmdIDCommandMapTable.get(new Integer(result.getCmdRef()));
                listener.receiveResultCommand(container, initialContainer);
              }
            }
        } else if (element.getElementType() == Element.STATUS) {
            Status status = (Status)container.getElement();
            String syncName = (String)synchroCmdIDServiceMapTable.get(new Integer(status.getCmdRef()));
            if (syncName != null) {
              OpenMSPSyncListener listener = (OpenMSPSyncListener)this.getListenerFromListWithName(synclistenerList,syncName);              
              if (listener != null) {
                ContainerMessage initialContainer = (ContainerMessage)synchroCmdIDCommandMapTable.get(new Integer(status.getCmdRef()));
                listener.receiveStatusCommand(status, initialContainer);
              }
            }
        } else  // unknown command
            LogManager.traceError(LogServices.WEBSERVICE, "Unknown command : " + element.getElementType());
      } catch (Throwable ex)  {
        LogManager.traceError(LogServices.WEBSERVICE, ex);
        try	{
	    	  OpenMSPSyncListener listener = null;
	    	  if (element instanceof Command)	{
		        String target = ((Command)element).getTarget();
		        listener = (OpenMSPSyncListener)this.getListenerFromListWithName(synclistenerList, target);
	    	  } else if (element instanceof Result)	{
	          Result result = (Result)container.getElement();
	          String syncName = (String)synchroCmdIDServiceMapTable.get(new Integer(result.getCmdRef()));
	          if (syncName != null) {
	            listener = (OpenMSPSyncListener)this.getListenerFromListWithName(synclistenerList,syncName);              
	          }
	        } else if (element.getElementType() == Element.STATUS) {
	          Status status = (Status)container.getElement();
	          String syncName = (String)synchroCmdIDServiceMapTable.get(new Integer(status.getCmdRef()));
	          if (syncName != null) {
	            listener = (OpenMSPSyncListener)this.getListenerFromListWithName(synclistenerList,syncName);              
	          }
	    	  }
	        if (listener != null) {
	          listener.notifySynchroFailure();
	        }
        } catch (Throwable ex2)	{
          LogManager.traceError(LogServices.WEBSERVICE, "Error during Listener synchro error management. Can't notify to listener synchro error.");
          LogManager.traceError(LogServices.WEBSERVICE, ex2);
        }
      }
    } //end while nextMessage()
  }
  
  private OpenMSPSyncListener getListenerFromListWithName(Array list, String name)  {
    for (int i=0; i<list.size(); i++) {
      OpenMSPSyncListener listener = (OpenMSPSyncListener) list.get(i);
      if (listener.getSyncName().equals(name))  {
        return listener;
      }
    }
    return null;
  }
  

  private long getSessionID() {
    return sessionID;
  }

  private void setSessionID (long sessionid) {
    NumSyncManagerDB.getManager().saveSyncNumberForService(sessionid, sessionIDServiceName);
    sessionID = sessionid;
  }

  private void saveCommandMessageIDs(Message message)  {
    synchroCmdIDServiceMapTable.clear();
    synchroCmdIDCommandMapTable.clear();
    message.resetCursor();
    while (message.hasMoreMessage())  {
      ContainerMessage container = message.nextMessage();
      saveOpenMSPSourcesAndIDs(container, null);
      saveOpenMSPCommandAndIds(container);
    }
  }

  private void saveOpenMSPSourcesAndIDs(ContainerMessage container, String currentSourceName) {
    container.resetCursor();
    Element element = container.getElement();
    if (element instanceof Command) {
      Command command = (Command)element;
      if (command.getSource() != null)  {
        currentSourceName = command.getSource();
      }
      synchroCmdIDServiceMapTable.put(new Integer(command.getCmdId()), currentSourceName);
    } else if (element instanceof RequestCommand) {
      RequestCommand command = (RequestCommand)element;
      if (currentSourceName != null)  {
        synchroCmdIDServiceMapTable.put(new Integer(command.getCmdId()), currentSourceName);
      } else  {
        LogManager.traceError(LogServices.SYNCHROSERVICE, "CmlID with no source in message. Can't make reference in return message :"+command.getCmdId());
      }
    }

    while (container.hasMoreMessage())  {
      saveOpenMSPSourcesAndIDs(container.nextMessage(), currentSourceName);
    }
  }

  private void saveOpenMSPCommandAndIds(ContainerMessage container) {
    Element element = container.getElement();
    if (element instanceof AbstractCommand) {
      synchroCmdIDCommandMapTable.put(new Integer(((AbstractCommand)element).getCmdId()), container);
      if (element.getElementType() == Element.SYNC)  {
        container.resetCursor();
        while (container.hasMoreMessage())  {
          saveOpenMSPCommandAndIds(container.nextMessage());
        }
      }
    }
  }

}
