/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */
package org.openmobileis.synchro.openmsp.server.synctarget;

import java.util.HashMap;

import org.openmobileis.common.context.ApplicationContextManager;
import org.openmobileis.common.context.UserTerminal;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.synchro.openmsp.server.synctarget.proxy.GenericProxySynchroTarget;
import org.openmobileis.synchro.openmsp.server.synctarget.proxy.ProxySyncroTarget;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: 
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class SynchroTargerManager {
  private static SynchroTargerManager manager;
  private HashMap targetMap;
  private HashMap proxyList;
  private ProxySyncroTarget defaultProxy;
  
  class TargetVersion {
    UserTerminal terminal;
    String version;
    OpenMSPSynchroTargetListener target; 
  }
  
  
  class TargetContainer {
    Array TargetList;
    String targetName;
    
    public boolean equals(Object obj) {
      if ((obj != null)
        && (obj instanceof TargetContainer)
        && ((TargetContainer)obj).targetName.equals(this.targetName)
      ) {
        return true;
      }
      return false;
    }

    public int hashCode() {
      return this.targetName.hashCode();
    }
  }
  
  /**
   * 
   */
  private SynchroTargerManager() {
    super();
    targetMap = new HashMap(20);
    proxyList = new HashMap(10);
  }
  
  public static SynchroTargerManager getManager() {
    if (manager == null)  {
      synchronized (SynchroTargerManager.class) {
        if (manager == null)  {
          manager = new SynchroTargerManager();
          ApplicationContextManager.getManager().addManager(manager);
        }
      }
    }
    return manager;
  }
  
  public void registerProxyTargetForTerminal(ProxySyncroTarget proxy, UserTerminal terminal)  {
    if(terminal == null){
    	defaultProxy = proxy;
    }
	  proxyList.put(terminal, proxy);
  }
  
  /**
   * Register a new Synchro target listener.
   * Version and terminal are optional.
   * Several target can be registered depending on their version and/or terminal.
   * Depending on the synchro terminal and synchro request, the right listener will be call.
   * If no listener is found for a version and/or terminal the nearest if return.
   * @param listener : target listener to be registered
   * @param version : version of the target listener (optional)
   * @param terminal : terminal that can use this listener (optional).
   */
  public void addSynchroTargetListener(OpenMSPSynchroTargetListener listener, String version, UserTerminal terminal)  {
    TargetContainer container = (TargetContainer)targetMap.get(listener.getTargetName());
    if (container == null)  {
      container = new TargetContainer();
      container.TargetList = new Array();
      container.targetName = listener.getTargetName();
      targetMap.put(listener.getTargetName(), container);
    }
    TargetVersion target = new TargetVersion();
    target.target = listener;
    target.version = version;
    target.terminal = terminal;
    container.TargetList.add(target);
  }
  
  /**
   * return the SynchroTargetListener with specified name, version and terminal.
   * Version and terminal are optional.
   * If no listener is found return null;
   * @param listenerName : name of the listener to return.
   * @param version : version of the target listener (optional)
   * @param terminal : terminal that can use this listener (optional).
   * @return
   */
  public OpenMSPSynchroTargetListener getSynchroTargetListener(String listenerName, String version , UserTerminal terminal) {
    TargetContainer container = (TargetContainer)targetMap.get(listenerName);
    if ((container == null) || (container.TargetList.size() == 0))  {
      return null;
    }  
    OpenMSPSynchroTargetListener foundlistener = ((TargetVersion)container.TargetList.get(0)).target;
    int size = container.TargetList.size();
    for (int i=0; i<size; i++)  {
      TargetVersion currentTarget = (TargetVersion)container.TargetList.get(i);
      if ((version != null) && (terminal != null))  {
        if ((currentTarget.version != null) && (currentTarget.terminal != null)) {
          if ((currentTarget.version.equals(version)) && (currentTarget.terminal.equals(terminal))){
            foundlistener = currentTarget.target;
            break;
          }
        }else   if ((currentTarget.version != null) && (currentTarget.version.equals(version))) {
          foundlistener = currentTarget.target;
        }else   if ((currentTarget.terminal != null) && (currentTarget.terminal.equals(terminal))) {
          foundlistener = currentTarget.target;
        }
      } else  if (version != null)  {
        if ((currentTarget.version != null) && (currentTarget.version.equals(version))) {
          foundlistener = currentTarget.target;
        }
      } else  if (terminal != null)  {
        if ((currentTarget.terminal != null) && (currentTarget.terminal.equals(terminal))) {
          foundlistener = currentTarget.target;
        }
      } else  {
        if ((currentTarget.terminal == null) && (currentTarget.version == null)) {
          foundlistener = currentTarget.target;
        }
      }
    }
    return foundlistener;
  }
  
  public ProxySyncroTarget getProxySynchroTargetForTerminal(UserTerminal terminal)  {
    ProxySyncroTarget proxy = (ProxySyncroTarget)proxyList.get(terminal);
    if (proxy == null)  {
      proxy  = defaultProxy;
    }
    return proxy;
  }

}
