/**
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 * 
 */
package org.openmobileis.test.database;

import java.io.File;

import org.odbms.ObjectSet;
import org.odbms.Query;
import org.openmobileis.common.context.SessionContext;
import org.openmobileis.common.context.SessionContextManager;
import org.openmobileis.common.user.UserManager;
import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.database.fastobjectdb.FODBCollectionDescriptor;
import org.openmobileis.database.fastobjectdb.FODBIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FODBStringIndexDescriptor;
import org.openmobileis.database.fastobjectdb.FastObjectDB;
import org.openmobileis.database.fastobjectdb.FastObjectDBManager;
import org.openmobileis.database.fastobjectdb.synchro.client.SynchroFastObjectDB;
import org.openmobileis.database.fastobjectdb.synchro.client.SynchroFastObjectDBManager;
import org.openmobileis.database.fastobjectdb.synchro.server.FODBOpenMSPSynchroTargetWrapper;
import org.openmobileis.synchro.algo.replication.utils.DefaultSynchroAtomicObject;
import org.openmobileis.synchro.algo.syncnumber.SyncNumberManager;
import org.openmobileis.synchro.algo.syncnumber.impl.EmptySyncNumberManagerDelegate;
import org.openmobileis.synchro.client.SynchroDescriptor;
import org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener;
import org.openmobileis.synchro.openmsp.client.OpenMSPSynchroManager;
import org.openmobileis.synchro.openmsp.client.conduit.HttpOpenMSPSynchroConduit;
import org.openmobileis.synchro.security.auth.Credential;
import org.openmobileis.test.mock.common.MockUserManagerFactory;
import org.openmobileis.test.mock.synchro.SimpleMockSynchroManager;

import junit.framework.Assert;
import junit.framework.TestCase;

/*
 * Title:        OpenMobileIS project source <BR>
 * Description:
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */

public class TestFODBSynchro extends TestCase {
  
  private TestSynchroFODBReturnListener returnListener;
  private String synchroInstallPath;

  /**
   * @param name
   */
  public TestFODBSynchro(String name) {
    super(name);

    returnListener = new TestSynchroFODBReturnListener();
    synchroInstallPath = System.getProperty("user.dir") + "/synchro";

    //setup application context
    LogManager.registerLogManager(null);

    PropertiesManager.getManager().addProperty("org.openmobileis.synchro.client.openmsp.numsyncpropsfile", System.getProperty("user.dir") + "/WEB-INF/conf/numsync.txt");
 
    //init db properties.
    PropertiesManager.getManager().addProperty("fastobjectdb.database.path", System.getProperty("user.dir") + "/database");
    PropertiesManager.getManager().addProperty("fastobjectdb.database.name", "testdb");
    

    //init context data
    Credential cred = new Credential("test", "test");
    UserManager.getManager().registerUserManagerFactoryForGroup("defgroup", new MockUserManagerFactory("defgroup", "defname", cred));

    //manage session
    SessionContext context = SessionContextManager.getManager().createSessionContext("testsession");
    SessionContextManager.getManager().joinSessionContext("testsession");
    context.setUserGroup("defgroup");
    context.setUserId("defname");
    
    //init synchro
    PropertiesManager.getManager().addProperty("org.openmobileis.synchro.direct.installpath", synchroInstallPath);   
    SyncNumberManager.getManager().registerDelegate(new EmptySyncNumberManagerDelegate());
}

  /* (non-Javadoc)
   * @see junit.framework.TestCase#setUp()
   */
  protected void setUp() throws Exception {
    super.setUp();
    //remove all previously created files
    //db files
    String path = System.getProperty("user.dir") + "/database/testdb/dummy.col";
    File file = new File(path);
    file.delete();
    path = System.getProperty("user.dir") + "/database/testdb/FWKACTIONDB.col";
    file = new File(path);
    file.delete();
    path = System.getProperty("user.dir") + "/database/testdb/OMJournalEntry.col";
    file = new File(path);
    file.delete();
    path = System.getProperty("user.dir") + "/database/testdb/testdb.db";
    file = new File(path);
    file.delete();
    
    path = System.getProperty("user.dir") + "/WEB-INF/conf/numsync.txt";
    file = new File(path);
    file.delete();
    
    path = System.getProperty("user.dir") + "/WEB-INF/conf/misc.properties";
    file = new File(path);
    file.delete();
    
    path = System.getProperty("user.dir") + "/synchro/importdb/dummy";
    file = new File(path);
    file.delete();
       
    //create collection
    try {
      FastObjectDB db = SynchroFastObjectDBManager.getCurrentFODB();
      // test if the collection has already been created.
      //if not create it. Created collection are automatiquely 
      // loaded when FODB is initialized.
      if (!db.isCollectionExist("dummy")) {
        //SimpleOpenMISInit intialize the synchronized version of FODB.
        // all created collection are synchronized by default.
        // Synchro listener are automatiquely registered when the collection is created or loaded.
        // use the FODBCollectionDescriptor to define a collection that is not synchronized.
        FODBCollectionDescriptor coldes = new FODBCollectionDescriptor("dummy", DummyObject.class);
        db.createCollection(coldes);
        //add index to query the collection by id
        FODBStringIndexDescriptor IDDescriptor = new FODBStringIndexDescriptor("ID", FODBIndexDescriptor.UNIQUE, "id", 12, 26);
        IDDescriptor.setCaseSensitive(true); //id is case sensitive. 
        db.addIndex("dummy", IDDescriptor);
      }
      // always register notification listener because it's not persistent.
      ((SynchroFastObjectDB) db).registerSynchroFODBReturnListener("dummy", returnListener);

    } catch (Exception ex) {
      LogManager.traceError(0, ex);
      Assert.fail("Exception occurs during setUp :" + ex.getMessage());
    }
  }

  /* (non-Javadoc)
   * @see junit.framework.TestCase#tearDown()
   */
  protected void tearDown() throws Exception {
    super.tearDown();
  }
  
  public void testDoSynchro()    {
    
    //init synchro
    SynchroDescriptor descriptor = new SynchroDescriptor();
    descriptor.setSynchroConduit(new HttpOpenMSPSynchroConduit(synchroInstallPath, ""));
    SimpleMockSynchroManager synchromock = new SimpleMockSynchroManager(synchroInstallPath);
    
    //get synchro terminal listener 
    OpenMSPSyncListener source = OpenMSPSynchroManager.getManager().getListenerByName("dummy");
    TestDummyFODBSyncTarget target = new TestDummyFODBSyncTarget();
    
    //do the synchro
    synchromock.registerTerminalSynchroListener(source);
    synchromock.registerServerSynchroTarget(new FODBOpenMSPSynchroTargetWrapper(target));
    Credential cred = new Credential("test", "test");
    try {
      // do a empty synchro No meta data, do data on both side
      synchromock.synchronize("defgroup", cred, descriptor);
      //validate that the synchro goes everywhere
      Assert.assertTrue(returnListener.notifySynchroBeginCall);
      Assert.assertTrue(returnListener.notifySynchroEndCall);
      Assert.assertNull(returnListener.testMetadata);
      Assert.assertNull(returnListener.notifyFODBObjectSynchroDoneId);
      Assert.assertNull(returnListener.notifyDBDeleteId);
      Assert.assertFalse(returnListener.notifyBeginDBUpdateCall);
      Assert.assertNull(target.synchroMetaData);
      Assert.assertNull(target.updateCollectionObject);
      Assert.assertNull(target.deleteCollectionObjectId);
      Assert.assertTrue(target.connectCall);
      Assert.assertTrue(target.disconnectCall);
      Assert.assertFalse(target.updateSynchroDBCall); //call only for complete.
      returnListener.notifySynchroBeginCall = false;
      target.connectCall = false;
      
      // do a meta data synchro. no collection object
      returnListener.testMetadata = "test metadata";
      synchromock.synchronize("defgroup", cred, descriptor);
      Assert.assertEquals("test metadata", target.synchroMetaData);
      Assert.assertEquals("test metadata", returnListener.testMetadata);
      
      //validate that the synchro goes everywhere
      Assert.assertTrue(returnListener.notifySynchroBeginCall);
      Assert.assertTrue(returnListener.notifySynchroEndCall);
      Assert.assertNull(returnListener.notifyFODBObjectSynchroDoneId);
      Assert.assertNull(returnListener.notifyDBDeleteId);
      Assert.assertFalse(returnListener.notifyBeginDBUpdateCall);
      Assert.assertNull(target.updateCollectionObject);
      Assert.assertNull(target.deleteCollectionObjectId);
      Assert.assertTrue(target.connectCall);
      Assert.assertTrue(target.disconnectCall);
      Assert.assertFalse(target.updateSynchroDBCall);
      returnListener.notifySynchroBeginCall = false;
      target.connectCall = false;
      
      //synchro one update to server.
      returnListener.testMetadata = null;
      
      DummyObject obj = new DummyObject("first");
      obj.value = "first object";
      FastObjectDBManager.getCurrentFODB().add("dummy", obj);
      synchromock.synchronize("defgroup", cred, descriptor);
      
      //validate the object has been synchronized
      Assert.assertEquals("first", returnListener.notifyFODBObjectSynchroDoneId);
      Assert.assertTrue(obj.equals(target.updateCollectionObject));
      
      //validate that the synchro goes everywhere
      Assert.assertTrue(returnListener.notifySynchroBeginCall);
      Assert.assertTrue(returnListener.notifySynchroEndCall);
      Assert.assertNull(returnListener.notifyDBDeleteId);
      Assert.assertFalse(returnListener.notifyBeginDBUpdateCall);
      Assert.assertNull(returnListener.testMetadata);
      Assert.assertNull(target.synchroMetaData);
      Assert.assertNull(target.deleteCollectionObjectId);
      Assert.assertTrue(target.connectCall);
      Assert.assertTrue(target.disconnectCall);
      Assert.assertFalse(target.updateSynchroDBCall);
      returnListener.notifySynchroBeginCall = false;
      target.connectCall = false;
      
      
      //do an insert with meta data
      DummyObject obj2 = new DummyObject("second");
      obj2.value = "second object";
      FastObjectDBManager.getCurrentFODB().add("dummy", obj2);
      returnListener.testMetadata = "test metadata";
      synchromock.synchronize("defgroup", cred, descriptor);
      Assert.assertEquals("test metadata", target.synchroMetaData);
      Assert.assertEquals("test metadata", returnListener.testMetadata);
      
      //validate the object has been synchronized
      Assert.assertTrue(obj2.equals(target.updateCollectionObject));
      Assert.assertEquals("second", returnListener.notifyFODBObjectSynchroDoneId);
      
      //validate that the synchro goes everywhere
      Assert.assertTrue(returnListener.notifySynchroBeginCall);
      Assert.assertTrue(returnListener.notifySynchroEndCall);
      Assert.assertNull(returnListener.notifyDBDeleteId);
      Assert.assertFalse(returnListener.notifyBeginDBUpdateCall);
      Assert.assertNull(target.deleteCollectionObjectId);
      Assert.assertTrue(target.connectCall);
      Assert.assertTrue(target.disconnectCall);
      Assert.assertFalse(target.updateSynchroDBCall);
      returnListener.notifySynchroBeginCall = false;
      target.connectCall = false;
      
      
      //test  synchro server update CREATE.
      target.updateCollectionObject = null;
      returnListener.notifyFODBObjectSynchroDoneId = null;
      returnListener.testMetadata = "test metadata";
      target.currentAO = new DefaultSynchroAtomicObject("third", "");
      synchromock.synchronize("defgroup", cred, descriptor);
      Assert.assertEquals("test metadata", target.synchroMetaData);
      Assert.assertEquals("test metadata", returnListener.testMetadata);
      
      //validate the object has been synchronized
      Assert.assertTrue(returnListener.notifyBeginDBUpdateCall);
      Query q = FastObjectDBManager.getCurrentFODB().query();
      q.constrain(DummyObject.class);
      Query subq = q.descend("id");
      subq.constrain("third").equal();
      ObjectSet set = q.execute();
      obj = (DummyObject)set.next();
      Assert.assertNotNull(obj);
      Assert.assertEquals("getCollectionObjectWithId", obj.value);
      
      //validate that the synchro goes everywhere
      Assert.assertTrue(returnListener.notifySynchroBeginCall);
      Assert.assertTrue(returnListener.notifySynchroEndCall);
      Assert.assertNull(returnListener.notifyDBDeleteId);
      Assert.assertNull(target.deleteCollectionObjectId);
      Assert.assertTrue(target.connectCall);
      Assert.assertTrue(target.disconnectCall);
      Assert.assertFalse(target.updateSynchroDBCall);
      Assert.assertNull(target.updateCollectionObject);
      Assert.assertNull(returnListener.notifyFODBObjectSynchroDoneId);
      returnListener.notifySynchroBeginCall = false;
      target.connectCall = false;
      
      //test  synchro server update UPDATE and terminal update.
      target.updateCollectionObject = null;
      returnListener.notifyFODBObjectSynchroDoneId = null;
      returnListener.testMetadata = "test metadata";
      target.currentAO = new DefaultSynchroAtomicObject("first", ""); //update first.
      
      FastObjectDBManager.getCurrentFODB().replace("dummy", obj2);
      
      synchromock.synchronize("defgroup", cred, descriptor);
      Assert.assertEquals("test metadata", target.synchroMetaData);
      Assert.assertEquals("test metadata", returnListener.testMetadata);
      Assert.assertTrue(obj2.equals(target.updateCollectionObject));
      Assert.assertEquals("second", returnListener.notifyFODBObjectSynchroDoneId);
      
      //validate the object has been synchronized
      Assert.assertTrue(returnListener.notifyBeginDBUpdateCall);
      q = FastObjectDBManager.getCurrentFODB().query();
      q.constrain(DummyObject.class);
      subq = q.descend("id");
      subq.constrain("first").equal();
      set = q.execute();
      obj = (DummyObject)set.next();
      Assert.assertNotNull(obj);
      Assert.assertEquals("getCollectionObjectWithId", obj.value);
      
      //validate that the synchro goes everywhere
      Assert.assertTrue(returnListener.notifySynchroBeginCall);
      Assert.assertTrue(returnListener.notifySynchroEndCall);
      Assert.assertNull(returnListener.notifyDBDeleteId);
      Assert.assertNull(target.deleteCollectionObjectId);
      Assert.assertTrue(target.connectCall);
      Assert.assertTrue(target.disconnectCall);
      Assert.assertFalse(target.updateSynchroDBCall);
      returnListener.notifySynchroBeginCall = false;
      target.connectCall = false;
 
      //test  complete synchro.
      target.updateCollectionObject = null;
      returnListener.notifyFODBObjectSynchroDoneId = null;
      returnListener.testMetadata = "test metadata";
      target.currentAO = new DefaultSynchroAtomicObject("first", ""); //server update first server to terminal. Not done because complete forced
      target.updateMaxNbRow = 0; //force complete.
      obj2 = new DummyObject("sync_comp");
      obj2.value = "second object";
      FastObjectDBManager.getCurrentFODB().add("dummy", obj2); //ob2 synchronized terminal to server
      
      synchromock.synchronize("defgroup", cred, descriptor);
      Assert.assertEquals("test metadata", target.synchroMetaData);
      Assert.assertEquals("test metadata", returnListener.testMetadata);
      Assert.assertTrue(obj2.equals(target.updateCollectionObject));
      Assert.assertEquals(obj2.id, returnListener.notifyFODBObjectSynchroDoneId); //terminal to server done
      Assert.assertTrue(target.updateSynchroDBCall); //complete done
      Assert.assertFalse(returnListener.notifyBeginDBUpdateCall); //no update synchro complete
      
      //validate the object has been synchronized (only COMPLETE not updated obj2
      q = FastObjectDBManager.getCurrentFODB().query();
      q.constrain(DummyObject.class);
      subq = q.descend("id");
      subq.constrain("COMPLETE").equal();
      set = q.execute();
      obj = (DummyObject)set.next();
      Assert.assertNotNull(obj);
      Assert.assertEquals("complete synchro", obj.value);
      q = FastObjectDBManager.getCurrentFODB().query();
      q.constrain(DummyObject.class);
      subq = q.descend("id");
      subq.constrain(obj2.id).equal();
      set = q.execute();
      Assert.assertFalse(set.hasNext());
      
      //validate that the synchro goes everywhere
      Assert.assertTrue(returnListener.notifySynchroBeginCall);
      Assert.assertTrue(returnListener.notifySynchroEndCall);
      Assert.assertNull(returnListener.notifyDBDeleteId);
      Assert.assertNull(target.deleteCollectionObjectId);
      Assert.assertTrue(target.connectCall);
      Assert.assertTrue(target.disconnectCall);
      returnListener.notifySynchroBeginCall = false;
      target.connectCall = false;
    
    } catch (Exception ex) {
      LogManager.traceError(0, ex);
      Assert.fail("Exception occurs during doSynchro :" + ex.getMessage());
    }
   
  }

}
